/******************************************************************************
*
* Copyright (C) 2006-2017  Brian Berlin
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
*
******************************************************************************
*
* FILENAME: sdlguiinput.cpp
*
* DESCRIPTION:  My own version of Guichan's SDLInput to handle joysticks
*
* AUTHORS:  bberlin
*
* CHANGE LOG
*
*  VER     DATE     CHANGED BY   DESCRIPTION OF CHANGE
* ------  --------  ----------   ------------------------------------
* 0.5.0   09/20/06  bberlin      Creation
******************************************************************************/
#include "sdlguiinput.hpp"
#include "guichan/exception.hpp"

extern "C" {
    #include "../interface/sdl_if.h"
}

namespace gcn
{
    SDLGuiInput::SDLGuiInput()
    {

    }

    bool SDLGuiInput::isJoystickQueueEmpty()
    {
        return mJoystickInputQueue.empty();
    }

    JoystickInput SDLGuiInput::dequeueJoystickInput()
    {
        JoystickInput joystickInput;

        if (mJoystickInputQueue.empty())
        {
            throw GCN_EXCEPTION("The queue is empty.");
        }

        joystickInput = mJoystickInputQueue.front();
        mJoystickInputQueue.pop();

        return joystickInput;
    }

    void SDLGuiInput::pushInput(SDL_Event event)
    {
        KeyInput keyInput;
        MouseInput mouseInput;
        JoystickInput joystickInput;
        int tmp_joy_index = 0;

        switch (event.type)
        {
          case SDL_JOYBALLMOTION:
              tmp_joy_index = pc_get_joystick_index_from_id(event.jball.which);
              joystickInput.setDevice(event.jball.which);
              joystickInput.setDeviceType(JoystickInput::JOYSTICK);
              joystickInput.setType(JoystickInput::BALL);
              joystickInput.setPart(event.jball.ball);
              joystickInput.setX(event.jball.xrel);
              joystickInput.setY(event.jball.yrel);
              mJoystickInputQueue.push(joystickInput);
              break;
          case SDL_JOYHATMOTION:
              tmp_joy_index = pc_get_joystick_index_from_id(event.jhat.which);
              joystickInput.setDevice(tmp_joy_index);
              joystickInput.setDeviceType(JoystickInput::JOYSTICK);
              joystickInput.setType(JoystickInput::HAT);
              joystickInput.setPart(event.jhat.hat);
              joystickInput.setValue(event.jhat.value);
              mJoystickInputQueue.push(joystickInput);
              break;
          case SDL_JOYAXISMOTION:
              tmp_joy_index = pc_get_joystick_index_from_id(event.jaxis.which);
              joystickInput.setDevice(tmp_joy_index);
              joystickInput.setDeviceType(JoystickInput::JOYSTICK);
              joystickInput.setType(JoystickInput::AXIS);
              joystickInput.setPart(event.jaxis.axis);
              joystickInput.setValue(event.jaxis.value);
              mJoystickInputQueue.push(joystickInput);
              break;
          case SDL_JOYBUTTONDOWN:
              tmp_joy_index = pc_get_joystick_index_from_id(event.jbutton.which);
              joystickInput.setDevice(tmp_joy_index);
              joystickInput.setDeviceType(JoystickInput::JOYSTICK);
              joystickInput.setType(JoystickInput::BUTTON);
              joystickInput.setPart(event.jbutton.button);
              joystickInput.setValue(JoystickInput::PRESS);
              mJoystickInputQueue.push(joystickInput);
              break;
          case SDL_JOYBUTTONUP:
              tmp_joy_index = pc_get_joystick_index_from_id(event.jbutton.which);
              joystickInput.setDevice(tmp_joy_index);
              joystickInput.setDeviceType(JoystickInput::JOYSTICK);
              joystickInput.setType(JoystickInput::BUTTON);
              joystickInput.setPart(event.jbutton.button);
              joystickInput.setValue(JoystickInput::RELEASE);
              mJoystickInputQueue.push(joystickInput);
              break;
          case SDL_CONTROLLERBUTTONUP:
              tmp_joy_index = pc_get_joystick_index_from_id(event.cbutton.which);
              joystickInput.setDevice(tmp_joy_index);
              joystickInput.setDeviceType(JoystickInput::CONTROLLER);
              joystickInput.setType(JoystickInput::BUTTON);
              joystickInput.setPart(event.cbutton.button);
              joystickInput.setValue(JoystickInput::RELEASE);
              mJoystickInputQueue.push(joystickInput);
              break;
          case SDL_CONTROLLERBUTTONDOWN:
              tmp_joy_index = pc_get_joystick_index_from_id(event.cbutton.which);
              joystickInput.setDevice(tmp_joy_index);
              joystickInput.setDeviceType(JoystickInput::CONTROLLER);
              joystickInput.setType(JoystickInput::BUTTON);
              joystickInput.setPart(event.cbutton.button);
              joystickInput.setValue(JoystickInput::PRESS);
              mJoystickInputQueue.push(joystickInput);
              break;
          case SDL_CONTROLLERAXISMOTION:
              tmp_joy_index = pc_get_joystick_index_from_id(event.caxis.which);
              joystickInput.setDevice(tmp_joy_index);
              joystickInput.setDeviceType(JoystickInput::CONTROLLER);
              joystickInput.setType(JoystickInput::AXIS);
              joystickInput.setPart(event.caxis.axis);
              joystickInput.setValue(event.caxis.value);
              mJoystickInputQueue.push(joystickInput);
              break;

        } // end switch

        SDLInput::pushInput(event);
    }
}
