/******************************************************************************
*
* Copyright (C) 2017  Brian Berlin
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
*
******************************************************************************
*
* FILENAME: shadercompile.cpp
*
* DESCRIPTION: This shows a window while compiling the shader
*
* AUTHORS:  bberlin
*
* CHANGE LOG
*
*  VER     DATE     CHANGED BY   DESCRIPTION OF CHANGE
* ------  --------  ----------   ------------------------------------
* 0.8.0   09/11/17  bberlin      Creation
******************************************************************************/
extern "C" {
    #include "../interface/shader.h"
}
#include "shadercompile.hpp"

#include <sys/stat.h>
#include <zlib.h>

extern "C" {
    #include "../interface/sdl_if.h"
    #include "../interface/logger.h"
    #include "../interface/util.h"
    #include "../interface/video.h"
    #include "../core/gtia.h"
    #include "../core/memory.h"
    #include "../core/console.h"
    #include "../core/pia.h"
    #include "../kat5200.h"


    extern t_atari_video g_video;
    extern t_sdl_if g_pc_if;
    extern int default_font_length;
    extern unsigned char default_font[];
}

ShaderCompile::ShaderCompile()
{
    t_config *p_config = config_get_ptr ();
    FILE *fp;
    gzFile zip_fp;
    struct stat file_status;
    unsigned char *new_buffer;
    int i;
    char string[1100];
    char *tmp;
    int width, height;

    /*
     * Setup the initial screen and input stuff
     */
    mScreen = (SDL_Window *) pc_set_window (g_video.width, g_video.height);
    mRenderer = (SDL_Renderer *) pc_get_renderer();

    /* Make sure that we reset SDL OpenGL state when needed */
    if ( g_video.shader_use ) {
        SDL_GL_UnbindTexture(g_pc_if.screen_texture);
        glDisable(GL_BLEND);
        SDL_SetRenderDrawBlendMode(mRenderer, SDL_BLENDMODE_NONE);
        SDL_RenderDrawPoint(mRenderer, -1, -1);
    }

    width = g_video.width < 640 ? g_video.width : 640;
    height = g_video.height < 480 ? g_video.height : 480;

    SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "linear");
    SDL_RenderSetLogicalSize(mRenderer, width, height);
    SDL_SetModState(KMOD_NONE);

    mImageLoader = new gcn::SDLImageLoader();
    gcn::Image::setImageLoader(mImageLoader); 
    mGraphics = new gcn::SDLGraphics();
    mGraphics->setTarget(mRenderer, width, height);
    mInput = new gcn::SDLGuiInput();
    SDL_SetWindowGrab(mScreen, SDL_FALSE);
    SDL_FlushEvents(SDL_KEYDOWN, SDL_MULTIGESTURE);
    
    mTop = new gcn::Container();    
    mTop->setDimension(gcn::Rectangle(0, 0, width, height));

    mGui = new gcn::KatGui();
    mGui->setGraphics(mGraphics);
    mGui->setInput(mInput); 
    mGui->setTop(mTop);
    mGui->addGlobalKeyListener ( this );
    mGui->addGlobalJoystickListener ( this );
    pc_detect_and_open_joysticks ();

    /*
     * Check for Font file, if not there create
     */
    strcpy ( string, p_config->font_file );
    util_set_file_to_program_dir ( string );
    if ( stat ( string, &file_status ) ) {
        sprintf ( string, "ShaderCompile: Font File \"%s\" ... does not exist!", p_config->font_file );
        logger_log_message ( LOG_ERROR, string, "" );
        strcpy ( string, "ShaderCompile: Attempting to use internal default..." );
        logger_log_message ( LOG_INFO, string, "" );
        
        /*
         * Use the internal default font
         */
        strcpy ( string, "tmp_font.bmp.gz" );
        util_set_file_to_program_dir ( string );

        fp = fopen ( string, "wb" );
        if ( fp ) {
            fwrite ( default_font, 1, default_font_length, fp );
            fclose ( fp );
        }

        new_buffer = (unsigned char *)malloc(55000);

        zip_fp = gzopen ( string, "rb" );
        if ( zip_fp ) {
            gzread ( zip_fp, new_buffer, 54582 );
        }
        gzclose (zip_fp);

        tmp = strstr ( string, ".gz" );
        *tmp = '\0';
            
        fp = fopen ( string, "wb" );
        if ( fp ) {
            for ( i = 0; i < 54582; ++i )
                fputc ( new_buffer[i], fp );
            fclose ( fp );
        }
        free ( new_buffer );
        mFont = new gcn::ImageFont(string, " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`'*#=[]\"<>{}^~|_@$\\");
    }
    else {
        mFont = new gcn::ImageFont(string, " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,!?-+/():;%&`'*#=[]\"<>{}^~|_@$\\");
    }
    gcn::Widget::setGlobalFont(mFont); 
    mGraphics->setFont(mFont);

    /*
     * Now create all windows and widgets
     */
    mTop->add(&mCompileDec);
    mTop->add(&mCompileTitle);

    mCompileDec.setDimension ( getRectangle(2,2,75,3.5) );
    mCompileTitle.setDimension ( getRectangle(4.5,3.5,75,1) );

    updateProgressMsg ( (char *)("Loading Shader. Please Wait...") );
}

ShaderCompile::~ShaderCompile()
{
    /*
     * Delete top level stuff
     */
    delete mFont;
    delete mTop;
    delete mGui;
    
    delete mInput;
    delete mGraphics;
    delete mImageLoader;
}

gcn::Rectangle ShaderCompile::getRectangle ( double x, double y, double w, double h )
{
    gcn::Rectangle NewRect;

    NewRect.x = (int)(x * (double)mFont->getWidth("A"));
    NewRect.y = (int)(y * (double)mFont->getHeight());
    NewRect.width = (int)(w * (double)mFont->getWidth("A"));
    NewRect.height = (int)(h * (double)mFont->getHeight());

    return NewRect;
}

gcn::Rectangle ShaderCompile::getPosition ( double x, double y )
{
    gcn::Rectangle NewRect;

    NewRect.x = (int)(x * (double)mFont->getWidth("A"));
    NewRect.y = (int)(y * (double)mFont->getHeight());

    return NewRect;
}

void ShaderCompile::updateProgressMsg( char *msg )
{
    mCompileTitle.setCaption(msg);
    mCompileTitle.adjustSize();
}

void ShaderCompile::exit( )
{
    mRunning = false;
}

void ShaderCompile::run()
{
    SDL_Event event;
    bool mJoystickCheck = false;
    int loop = 0;
    int start;
    int stop;

    mRunning = true;

    start = pc_get_ticks();
    while ( mRunning ) 
    {
        while ( SDL_PollEvent ( &event ) ) 
        {
            if (event.type == SDL_QUIT)
            {
                this->exit();
            }
            else if (event.type ==  SDL_WINDOWEVENT) {
                if (event.window.event == SDL_WINDOWEVENT_RESIZED) {
                    if (!(SDL_GetWindowFlags(g_pc_if.screen) & SDL_WINDOW_FULLSCREEN_DESKTOP)) {
                        g_video.width = event.window.data1;
                        g_video.height = event.window.data2;
                        video_save_profile(g_video.name, &g_video);
                    }
                }
                if (event.window.event == SDL_WINDOWEVENT_MOVED) {
                    if (!(SDL_GetWindowFlags(g_pc_if.screen) & SDL_WINDOW_FULLSCREEN_DESKTOP)) {
                        g_video.x = event.window.data1;
                        g_video.y = event.window.data2;
                        video_save_profile(g_video.name, &g_video);
                    }
                }
            }
        }

        SDL_RenderClear(mRenderer);

        mGui->logic();
        mGui->draw();

        SDL_RenderPresent(mRenderer);

        stop = pc_get_ticks ();
        while ( (stop-start) < 17 ) {
            stop = pc_get_ticks ();
        }
        start = pc_get_ticks ();

        ++loop;
        if ( loop == 2 ) 
            mRunning = false;
    }
}

