/******************************************************************************
*
* Copyright (C) 2006-2017  Brian Berlin
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
*
******************************************************************************
*
* FILENAME: input.c
*
* DESCRIPTION:  This contains input handling
*
* AUTHORS:  bberlin
*
* CHANGE LOG
*
*  VER     DATE     CHANGED BY   DESCRIPTION OF CHANGE
* ------  --------  ----------   ------------------------------------
* 0.5.0   01/10/06  bberlin      Creation, break out from config
* 0.6.2   06/24/09  bberlin      input_setup_keypad change stop to 16, fixes
*                                keypad #1 not working at all
* 0.7.1   02/17/17  bberlin      Simulate analog addressed wrong in input
*                                setup stick
* 0.8.0   04/08/17  bberlin      Fix setup_800keypad to use 16 keys instead
*                                of 64.  Add function to check for input
*                                conficts
* 0.8.0   04/12/17  bberlin      Fix input_setup_paddles_button to not crash
*                                with mod key keyboard key
* 0.8.0   11/13/17  bberlin      Add combo inputs
* 0.8.0   11/21/17  bberlin      input_keyboard_callback 54->55 a key loaded
* 0.8.0   12/03/17  bberlin      Fix frame functions to respect pot_low
*                                  Simulate Analog fix
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "input.h"
#include "kconfig.h"
#include "states.h"
#include "media.h"
#include "video.h"
#include "sound.h"
#include "db_if.h"
#include "ui.h"
#include "util.h"
#include "keycodes.h"
#include "../kat5200.h"
#include "../core/pokey.h"
#include "../core/pia.h"
#include "../core/debug.h"
#include "../core/memory.h"
#include "../gui/gui.h"

typedef struct {
    int axis_idle[2];
    int axis_value[2];
    int pot_low;
    int pot_high;
    int sa_high_threshold;
    int sa_low_threshold;
    int pressed[MAX_DIR];
    int digital_pressed[MAX_DIR];
    unsigned char pot_values[0x10000];
} t_pc_stick_interface;

static t_pc_stick_interface g_stick_interface[MAX_CONTROLLER];

t_atari_input g_input;
t_pc_input g_pc_input;

/*
 * Profile functions
 */
char * input_get_key_string ( int key, char *name );

/*
 * Setup functions
 */
void input_combo_disengage_all ( void );
void input_set_controller_defaults ( void );
int input_set_controls ( void );
int input_preload_controls ( void );
int input_setup_frame_functions ( int player );
int input_setup_stick_values ( int player );
int input_setup_stick ( int player, t_atari_stick *stick );
int input_setup_trackball ( int player, t_atari_stick *stick );
int input_setup_keypad ( int player, t_atari_common *ctrl );
int input_setup_digital_stick ( int player, t_atari_stick *stick );
int input_setup_800keypad ( int player );
int input_setup_buttons ( int player );
int input_setup_paddle_button ( int player );
int input_setup_switch ( void );
int input_setup_combo_actions ( void );
void input_free_combo_function_list ( void );

/*
 * Frame functions
 *   Called by our PC event loop every frame
 */
int input_center_x_axis_frame (int player, int dir);
int input_center_y_axis_frame (int player, int dir);
int input_simulate_analog_x_axis_frame (int player, int tmp);
int input_simulate_analog_y_axis_frame (int player, int tmp);

/*
 * Device Interface functions
 *   Called by our PC event loop to set Atari Registers
 */
int input_do_nothing (int key, int dir, int temp);
int input_frame_do_nothing (int key, int dir);
int input_key_down_stick (int player, int dir, int key);
int input_key_down_stick_sa (int player, int dir, int key);
int input_key_up_stick (int player, int dir, int key);
int input_key_up_stick_sa (int player, int dir, int key);
int input_key_down_digital_stick (int player, int dir, int key);
int input_key_up_digital_stick (int player, int dir, int key);
int input_joystick_pots_sa ( int player, int dir, int reading );
int input_joystick_pots ( int player, int dir, int reading );
int input_keypad_down ( int player, int dir, int kbcode );
int input_keypad_up ( int player, int dir, int kbcode );
int input_axis_digital_stick ( int player, int dir, int reading );
int input_modkey_down ( int player, int dir, int kbcode );
int input_modkey_up ( int player, int dir, int kbcode );
int input_joyaxis_modkey ( int player, int reading, int kbcode );
int input_switch_down ( int player, int dir, int mask );
int input_switch_up ( int player, int dir, int mask );
int input_joyaxis_switch ( int player, int reading, int mask );
int input_joyaxis_keypad ( int player, int reading, int kbcode );
int input_800keypad_down ( int player, int dir, int kbcode );
int input_800keypad_up ( int player, int dir, int kbcode );
int input_joyaxis_800keypad ( int player, int reading, int kbcode );
int input_top_button_down ( int player, int dir, int temp );
int input_top_button_up ( int player, int dir, int temp );
int input_joyaxis_top_button ( int player, int dir, int reading );
int input_joyaxis_bottom_button ( int player, int dir, int reading );
int input_stick_h_axis ( int player, int reading );
int input_stick_v_axis ( int player, int reading );
int input_mouse_axis (int player, int dir, int value);
int input_mouse_vel_axis (int player, int dir, int value);
int input_trackball_x_axis (int player, int dir, int value);
int input_trackball_y_axis (int player, int dir, int value);
int input_joyhat_combo_engage (int device_num, int direction, int part_num);
int input_joyhat_combo_disengage (int device_num, int direction, int part_num);
int input_joyaxis_combo (int device_num, int part_num, int value);
int input_keyboard_combo_engage (int device_num, int part_type, int part_num);
int input_keyboard_combo_disengage (int device_num, int part_type, int part_num);
/*
 * User Interface functions
 *   Called by our PC event loop to perform user actions
 */
int input_ui_debug (int key, int temp, int temp1);
int input_ui_exit (int key, int temp, int temp1);
int input_ui_back (int key, int temp, int temp1);
int input_ui_fullscreen (int key, int temp, int temp1);
int input_ui_load_state (int player, int temp, int temp1);
int input_ui_save_state (int player, int temp, int temp1);

/******************************************************************************
**  Function   :  input_set_defaults                                            
**                                                                    
**  Objective  :  This function sets up controller and input defaults
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise failure
** 
******************************************************************************/ 
int input_set_defaults ( void ) {

    strcpy ( g_input.name, "Internal Defaults" );
    strcpy ( g_input.description, "Internal Defaults" );
    pc_set_controller_defaults ();
    pc_set_keyboard_defaults ();

    input_set_controls();

    return 0;

} /* end input_set_defaults */

/******************************************************************************
**  Function   :  input_set_controls                                            
**                                                                    
**  Objective  :  This function loads up the pc interface with the control
**                functions based on the input configuration
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_set_controls ( void ) {

    int i;
    int max_controller = MAX_CONTROLLER;

    input_free_combo_function_list ();
    input_preload_controls ();
    ui_setup_keys ();

    if ( g_input.machine_type == MACHINE_TYPE_XL )
        max_controller = 2;

    for ( i = 0; i < max_controller; ++i ) {
        input_setup_frame_functions ( i );
        input_setup_stick_values ( i );

        if ( g_input.machine_type == MACHINE_TYPE_5200 ) {
            if ( g_input.players[i].control_type != CTRLR_TYPE_NONE ) {
                if ( g_input.players[i].control_type == CTRLR_TYPE_TRACKBALL )
                    input_setup_trackball ( i, &g_input.players[i].stick );
                else
                    input_setup_stick ( i, &g_input.players[i].stick );
                input_setup_keypad ( i, g_input.players[i].keypad );
            }
        }
        else {
            switch ( g_input.players[i].control_type ) {
                case CTRLR_TYPE_JOYSTICK: 
                    input_setup_digital_stick ( i, &g_input.players[i].stick );
                    break;
                case CTRLR_TYPE_PADDLES: 
                    input_setup_stick ( i, &g_input.players[i].paddles );
                    break;
                case CTRLR_TYPE_KEYPAD: 
                    input_setup_800keypad ( i );
                    break;
                default:
                    break;
            }
        }
        if ( g_input.players[i].control_type != CTRLR_TYPE_NONE ) {
            if ( g_input.players[i].control_type == CTRLR_TYPE_PADDLES )
                input_setup_paddle_button ( i );
            else
                input_setup_buttons ( i );
        }
    }

    if ( g_input.machine_type != MACHINE_TYPE_5200 ) {
        input_setup_keypad ( 0, g_input.keyboard );
        input_setup_switch ( );
    }

    input_setup_combo_actions ();
    input_combo_disengage_all ();

    return 0;

} /* input_set_controls */

t_atari_input_list * input_create_list_input ( t_atari_input_list *p_list, 
                                      t_atari_common *p_input, char *title ) {

    t_atari_input_list *tmp;
    if (p_list == NULL) {
        p_list = malloc(sizeof(t_atari_input_list));
        tmp = p_list;
    }
    else {
        p_list->next = malloc(sizeof(t_atari_input_list));
        tmp = p_list->next;
    }

    strcpy ( tmp->name, title );
    tmp->ctrl = p_input;

    return tmp;

} /* input_create_list_input */

void input_add_combo_function ( t_atari_common *ctrl,
       int (**d_addr)(int,int,int), int (*d)(int,int,int),
       int *param1addr, int *param2addr, int *param3addr ) {

    t_pc_combo_function *tmp = g_pc_input.combo_list;
    t_pc_combo_function *prev = g_pc_input.combo_list;

    if (g_pc_input.combo_list == NULL) {
        g_pc_input.combo_list = malloc(sizeof(t_pc_combo_function));
        tmp = g_pc_input.combo_list;
        tmp->combo_off_function = input_do_nothing;
        tmp->next = NULL;
    }
    else {
        while ( tmp && tmp->function_addr != d_addr ) {
            prev = tmp;
            tmp = tmp->next;
        }
        if ( !tmp ) {
            prev->next = malloc(sizeof(t_atari_input_list));
            prev->next->combo_off_function = input_do_nothing;
            prev->next->next = NULL;
            tmp = prev->next;
        }
    }

    tmp->ctrl = ctrl;
    tmp->function_addr = d_addr;
    tmp->combo_on_function = d;
    tmp->param_addr[0] = param1addr;
    tmp->param_addr[1] = param2addr;
    tmp->param_addr[2] = param3addr;
    tmp->param_on[0] = *tmp->param_addr[0];
    tmp->param_on[1] = *tmp->param_addr[1];
    tmp->param_on[2] = *tmp->param_addr[2];

} /* input_add_combo_function */

int input_set_combo_off_function ( int (**e_addr)(int,int,int), 
       int param1, int param2, int param3 ) {

    t_pc_combo_function *tmp = g_pc_input.combo_list;
    t_pc_combo_function *prev = g_pc_input.combo_list;

    if (g_pc_input.combo_list == NULL) {
        g_pc_input.combo_list = malloc(sizeof(t_pc_combo_function));
        tmp = g_pc_input.combo_list;
        tmp->next = NULL;
    }
    else {
        while ( tmp && tmp->function_addr != e_addr ) {
            prev = tmp;
            tmp = tmp->next;
        }
        if ( !tmp ) {
            prev->next = malloc(sizeof(t_atari_input_list));
            prev->next->next = NULL;
            tmp = prev->next;
        }
    }

    tmp->function_addr = e_addr;
    tmp->combo_off_function = *e_addr;
    tmp->param_off[0] = param1;
    tmp->param_off[1] = param2;
    tmp->param_off[2] = param3;

    return 0;

} /* input_set_combo_off_function */

int input_update_combo_off_function ( int (**e_addr)(int,int,int), 
       int param1, int param2, int param3 ) {

    int status = 0;
    t_pc_combo_function *tmp = g_pc_input.combo_list;
    while ( tmp ) {
        if ( tmp->function_addr == e_addr ) {
            tmp->combo_off_function = *e_addr;
            tmp->param_off[0] = param1;
            tmp->param_off[1] = param2;
            tmp->param_off[2] = param3;
            status = 1; 
        }
        tmp = tmp->next;
    }
    return status;

} /* input_update_combo_off_function */

void input_combo_disengage_all ( void ) {

    int i;
    t_pc_combo_function *tmp = g_pc_input.combo_list;
    while ( tmp ) {
        *tmp->function_addr = tmp->combo_off_function;
        for ( i = 0; i < 3; ++i )
            *(tmp->param_addr[i]) = tmp->param_off[i];
        tmp = tmp->next;
    }
}

void input_free_combo_function_list ( void ) {

    t_pc_combo_function *tmp, *tmp2;

    if ( g_pc_input.combo_list == NULL )
        return;

    tmp = g_pc_input.combo_list;

    while ( tmp ) {
        tmp2 = tmp;
        tmp = tmp->next;
        free(tmp2);
    }

    g_pc_input.combo_list = NULL;

} /* input_free_combo_function_list */

/******************************************************************************
**  Function   :  input_check_conflicts                                            
**                                                                    
**  Objective  :  This function loads up the pc interface with the control
**                functions based on the input configuration
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
t_atari_input_list * input_check_conflicts ( t_atari_input *p_input ) {

    int i, j;
    int max_controller = MAX_CONTROLLER;
    char tmp_name[256];
    t_atari_input_list *list;
    t_atari_input_list *current;
    int key[] = {KEY_0>>1,KEY_1>>1,KEY_2>>1,KEY_3>>1,KEY_4>>1,
                 KEY_5>>1,KEY_6>>1,KEY_7>>1,KEY_8>>1,KEY_9>>1};

    if ( p_input->machine_type == MACHINE_TYPE_XL )
        max_controller = 2;

    list = malloc(sizeof(t_atari_input_list));
    current = list;

    /*
     * Create a list to search for
     */

    /* Add controllers to the list */
    for ( i = 0; i < max_controller; ++i ) {
        switch ( p_input->players[i].control_type ) {
            case CTRLR_TYPE_JOYSTICK: 
            case CTRLR_TYPE_TRACKBALL: 
                sprintf ( tmp_name, "Player %d Left", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].stick.direction[DIR_LEFT]), 
                        tmp_name );
                sprintf ( tmp_name, "Player %d Right", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].stick.direction[DIR_RIGHT]), 
                        tmp_name );
                sprintf ( tmp_name, "Player %d Up", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].stick.direction[DIR_UP]), 
                        tmp_name );
                sprintf ( tmp_name, "Player %d Down", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].stick.direction[DIR_DOWN]), 
                        tmp_name );

                if ( p_input->machine_type == MACHINE_TYPE_5200 ||
                       p_input->players[i].control_type == CTRLR_TYPE_KEYPAD ) {
                    for ( j = 0; j < 10; ++j ) {
                        sprintf ( tmp_name, "Player %d Keypad %d", i+1, j);
                        current = input_create_list_input ( current, 
                                &(p_input->players[i].keypad[key[j]]), 
                                tmp_name );
                    }
                    sprintf ( tmp_name, "Player %d Keypad Pound", i+1);
                    current = input_create_list_input ( current, 
                            &(p_input->players[i].keypad[KEY_POUND>>1]), 
                            tmp_name );
                    sprintf ( tmp_name, "Player %d Keypad Star", i+1);
                    current = input_create_list_input ( current, 
                            &(p_input->players[i].keypad[KEY_STAR>>1]), 
                            tmp_name );
                    sprintf ( tmp_name, "Player %d Start", i+1);
                    current = input_create_list_input ( current, 
                            &(p_input->players[i].keypad[KEY_START>>1]), 
                            tmp_name );
                    sprintf ( tmp_name, "Player %d Pause", i+1);
                    current = input_create_list_input ( current, 
                            &(p_input->players[i].keypad[KEY_PAUSE>>1]), 
                            tmp_name );
                    sprintf ( tmp_name, "Player %d Reset", i+1);
                    current = input_create_list_input ( current, 
                            &(p_input->players[i].keypad[KEY_RESET>>1]), 
                            tmp_name );
                    sprintf ( tmp_name, "Player %d Top Button", i+1);
                    current = input_create_list_input ( current, 
                            &(p_input->players[i].top_button), 
                            tmp_name );
                }

                break;
            case CTRLR_TYPE_PADDLES: 
                sprintf ( tmp_name, "Player %d Paddle 1 Left", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].paddles.direction[DIR_LEFT]), 
                        tmp_name );
                sprintf ( tmp_name, "Player %d Paddle 1 Right", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].stick.direction[DIR_RIGHT]), 
                        tmp_name );
                sprintf ( tmp_name, "Player %d Paddle 2 Left", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].stick.direction[DIR_UP]), 
                        tmp_name );
                sprintf ( tmp_name, "Player %d Paddle 2 Right", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].stick.direction[DIR_DOWN]), 
                        tmp_name );
                break;
            case CTRLR_TYPE_KEYPAD: /* 800 keypad */ 
                for ( j = 0; j < 10; ++j ) {
                    sprintf ( tmp_name, "Player %d Keypad %d", i+1, j);
                    current = input_create_list_input ( current, 
                            &(p_input->players[i].keypad[key[j]]), 
                            tmp_name );
                }
                sprintf ( tmp_name, "Player %d Keypad Pound", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].keypad[KEY_POUND>>1]), 
                        tmp_name );
                sprintf ( tmp_name, "Player %d Keypad Star", i+1);
                current = input_create_list_input ( current, 
                        &(p_input->players[i].keypad[KEY_STAR>>1]), 
                        tmp_name );
                break;
            default:
                break;
        }

        if ( p_input->players[i].control_type != CTRLR_TYPE_KEYPAD &&
             p_input->players[i].control_type != CTRLR_TYPE_NONE ) {

            sprintf ( tmp_name, "Player %d Bottom Button", i+1);
            current = input_create_list_input ( current, 
                    &(p_input->players[i].bottom_button), 
                    tmp_name );
        }

    } /* end for each player */

    /* Add keyboard and special keys to the list */
    if ( p_input->machine_type != MACHINE_TYPE_5200 ) {
        for ( j = 0; j < MAX_KEYS; ++j ) {
            input_get_key_string ( j, tmp_name );
            if (strlen(tmp_name)) {
                current = input_create_list_input ( current, 
                        &(p_input->keyboard[j]), 
                        tmp_name );
            }
        }
        strcpy ( tmp_name, "Control Key");
        current = input_create_list_input ( current, 
                &(p_input->ctrl_key), 
                tmp_name );
        strcpy ( tmp_name, "Shift Key");
        current = input_create_list_input ( current, 
                &(p_input->shift_key), 
                tmp_name );
        strcpy ( tmp_name, "Break Key");
        current = input_create_list_input ( current, 
                &(p_input->break_key), 
                tmp_name );
        strcpy ( tmp_name, "Start Key");
        current = input_create_list_input ( current, 
                &(p_input->start_key), 
                tmp_name );
        strcpy ( tmp_name, "Select Key");
        current = input_create_list_input ( current, 
                &(p_input->select_key), 
                tmp_name );
        strcpy ( tmp_name, "Option Key");
        current = input_create_list_input ( current, 
                &(p_input->option_key), 
                tmp_name );
        strcpy ( tmp_name, "Reset Key");
        current = input_create_list_input ( current, 
                &(p_input->reset_key), 
                tmp_name );

    } /* end if not a 5200 */

    /*
     * Now go through the list
     *   We will create another list of the pairs of conflicts
     *   We just use the string member
     */
    t_atari_input_list *string_list = malloc(sizeof(t_atari_input_list));
    t_atari_input_list *tmp_string_list = string_list;
    tmp_string_list->name[0] = '\0';

    current->next = NULL;
    current = list->next;
    free (list);
    while (current) {
        list = current->next;
        while (list) {
            if ( list->ctrl && current->ctrl &&
                 list->ctrl->device[0] == current->ctrl->device[0] &&
                 list->ctrl->device_num[0] == current->ctrl->device_num[0] &&
                 list->ctrl->part_type[0] == current->ctrl->part_type[0] &&
                 list->ctrl->part_num[0] == current->ctrl->part_num[0] &&
                 list->ctrl->direction[0] == current->ctrl->direction[0] &&
                 (!list->ctrl->combo ||
                  (list->ctrl->device[1] == current->ctrl->device[1] &&
                  list->ctrl->device_num[1] == current->ctrl->device_num[1] &&
                  list->ctrl->part_type[1] == current->ctrl->part_type[1] &&
                  list->ctrl->part_num[1] == current->ctrl->part_num[1] &&
                  list->ctrl->direction[1] == current->ctrl->direction[1])
                 )
                 ) {

                sprintf ( tmp_string_list->name,  "%s : %s", current->name, 
                                                            list->name );
                tmp_string_list->next = malloc(sizeof(t_atari_input_list));
                tmp_string_list = tmp_string_list->next;
                tmp_string_list->name[0] = '\0';
                tmp_string_list->next = NULL;
            }
            list = list->next;
        }
        list = current->next;
        free (current);
        current = list;
    }

    if ( !strlen(string_list->name) ) {
        free(string_list);
        string_list = NULL;
    }

    return string_list;

} /* input_check_conflicts */

/************************************************************************
 *
 *  Profile functions
 *
 ************************************************************************/

/******************************************************************************
**  Function   :  input_set_profile                                            
**                                                                    
**  Objective  :  This function loads up an input profile from file and puts it
**                in our global input struct
**                                                   
**  Parameters :  profile_name - filename containing input to load
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_set_profile ( char *profile_name ) {
        
    if ( *profile_name == '\0' ) {
        if ( config_get_ptr()->machine_type == MACHINE_TYPE_5200 )
            input_load_profile ( config_get_ptr()->default_input_profile, &g_input );
        else
            input_load_profile ( config_get_ptr()->default_800input_profile, &g_input );
    }
    else
        input_load_profile ( profile_name, &g_input );
    input_set_controls ();

    return 0;

} /* end input_set_profile */

/******************************************************************************
**  Function   :  input_keyboard_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  keyboard  - pointer to keyboard array
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_keyboard_callback ( void *keyboard, int argc, char **argv, char **azColName ) {

    int i;
    int row_value;
    t_atari_common *p_keyboard = keyboard;
    char statement[50];

    if ( argc < 54 )
        return -1;

    for ( i = 1; i < 55; ++i ) {
        if ( argv[i] ) {
            row_value = 0;
            sprintf ( statement, "SELECT * FROM KeyboardCodes WHERE ROWID='%d'", i );
            db_if_exec_sql ( statement, db_if_id_callback, &row_value );
            input_find_common ( atoi(argv[i]), &p_keyboard[row_value] ); 
        }
    }

    return 0;

} /* end input_keyboard_callback */

/******************************************************************************
**  Function   :  input_keypad_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  keypad    - pointer to keypad array
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_keypad_callback ( void *keypad, int argc, char **argv, char **azColName ) {

    int i;
    t_atari_common *p_keypad = keypad;
    int key_values[] = {KEY_START>>1,KEY_PAUSE>>1,KEY_RESET>>1,KEY_STAR>>1,
                        KEY_POUND>>1,KEY_0>>1,KEY_1>>1,KEY_2>>1,KEY_3>>1,KEY_4>>1,
                        KEY_5>>1,KEY_6>>1,KEY_7>>1,KEY_8>>1,KEY_9>>1};

    if ( argc < 16 )
        return -1;

    for ( i = 1; i < 16; ++i ) {
        if ( argv[i] )
            input_find_common ( atoi(argv[i]), &p_keypad[key_values[i-1]] ); 
    }

    return 0;

} /* end input_keypad_callback */

/******************************************************************************
**  Function   :  input_player_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  player    - pointer to player struct
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_player_callback ( void *player, int argc, char **argv, char **azColName ) {

    int i, keypad=0;
    t_atari_player *p_player = player;
    char statement[50];

    if ( argc < 16 )
        return -1;

    /*
     * THIS DEPENDS ON MATCHING THE DATABASE!
     */
    i = 0;
    if ( argv[++i] ) p_player->control_type = atoi ( argv[i] );
    if ( argv[++i] ) p_player->stick.deadzone = atoi ( argv[i] );
    if ( argv[++i] ) p_player->stick.sensitivity = atoi ( argv[i] );
    if ( argv[++i] ) p_player->stick.simulate_analog = atoi ( argv[i] );
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->stick.direction[DIR_LEFT] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->stick.direction[DIR_RIGHT] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->stick.direction[DIR_UP] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->stick.direction[DIR_DOWN] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->paddles.direction[DIR_LEFT] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->paddles.direction[DIR_RIGHT] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->paddles.direction[DIR_UP] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->paddles.direction[DIR_DOWN] ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->bottom_button ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_player->top_button ); 
    if ( argv[++i] ) keypad = atoi ( argv[i] );

    sprintf ( statement, "SELECT * FROM KeyPad WHERE KeyPadID='%d'", keypad );
    db_if_exec_sql ( statement, input_keypad_callback, p_player->keypad );

    return 0;

} /* end input_player_callback */

/******************************************************************************
**  Function   :  input_load_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  input     - pointer to input struct
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_load_callback ( void *input, int argc, char **argv, char **azColName ) {

    int i;
    int keyboard=0;
    int player[4] = {0};
    char statement[50];
    t_atari_input *p_input = input;

    if ( argc < 15 )
        return -1;

    /*
     * THIS DEPENDS ON MATCHING THE DATABASE!
     */
    i = -1;
    if ( argv[++i] ) strcpy ( p_input->name, argv[i] );
    if ( argv[++i] ) strcpy ( p_input->description, argv[i] );
    if ( argv[++i] ) p_input->machine_type = atoi ( argv[i] );
    if ( argv[++i] ) p_input->mouse_speed = atoi ( argv[i] );
    if ( argv[++i] ) player[0] = atoi ( argv[i] );
    if ( argv[++i] ) player[1] = atoi ( argv[i] );
    if ( argv[++i] ) player[2] = atoi ( argv[i] );
    if ( argv[++i] ) player[3] = atoi ( argv[i] );
    if ( argv[++i] ) keyboard = atoi ( argv[i] );
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_input->start_key ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_input->select_key ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_input->option_key ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_input->ctrl_key ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_input->shift_key ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_input->break_key ); 
    if ( argv[++i] ) input_find_common ( atoi(argv[i]), &p_input->reset_key ); 

    for ( i = 0; i < MAX_CONTROLLER; ++i ) {
        sprintf ( statement, "SELECT * FROM Player WHERE PlayerID='%d'", player[i] );
        db_if_exec_sql ( statement, input_player_callback, &p_input->players[i] );
    }

    sprintf ( statement, "SELECT * FROM Keyboard WHERE KeyboardID='%d'", keyboard );
    db_if_exec_sql ( statement, input_keyboard_callback, p_input->keyboard );

    return 0;

} /* end input_load_callback */

/******************************************************************************
**  Function   :  input_load_profile                                            
**                                                                    
**  Objective  :  This function loads up an input profile from the database
**                                                   
**  Parameters :  name     - name of the input profile to load
**                p_input  - input structure to load with info
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_load_profile ( const char *name, t_atari_input *p_input ) {
    
    int status;
    char statement[257];
    char *p_tmp;

    /*
     * Create the SQL query statement
     */
    db_if_exec_sql ( "BEGIN", 0, 0 );

    sprintf ( statement, "SELECT * FROM Input WHERE Name='%s'", 
               p_tmp=util_replace_and_copy(name, "'", "''" ) );

    free ( p_tmp );

    status = db_if_exec_sql ( statement, input_load_callback, p_input );

    db_if_exec_sql ( "END", 0, 0 );

    return status;

} /* end input_load_profile */

/******************************************************************************
**  Function   :  input_write_keypad                                            
**                                                                    
**  Objective  :  This function writes player parameters to the profile
**                                                   
**  Parameters :  statement - SQL to check for existence of this entry
**                keypad    - struct containing values to write
**                id        - returns the id of this entry in the table
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_write_keypad ( char *statement, t_atari_common *keypad, int *id ) {

    int keyid,i;
    char values[256];
    char tstatement[256];
    char key_strings[][20] = {"KEY_START","KEY_PAUSE","KEY_RESET", "KEY_STAR",
                              "KEY_POUND","KEY_0","KEY_1","KEY_2","KEY_3","KEY_4",
                              "KEY_5","KEY_6","KEY_7","KEY_8","KEY_9"};
    int key_values[] = {KEY_START>>1,KEY_PAUSE>>1,KEY_RESET>>1,KEY_STAR>>1,
                        KEY_POUND>>1,KEY_0>>1,KEY_1>>1,KEY_2>>1,KEY_3>>1,KEY_4>>1,
                        KEY_5>>1,KEY_6>>1,KEY_7>>1,KEY_8>>1,KEY_9>>1};
    *id = 0;

    db_if_exec_sql ( statement, db_if_id_callback, id );

    for ( i = 0; i < 15; ++i ) {
        sprintf ( tstatement, "SELECT %s FROM Keypad WHERE KeypadID = %d", key_strings[i], *id );
        input_write_common ( tstatement, &keypad[key_values[i]], &keyid );
        if ( i == 0 )
            sprintf ( values, "%d", keyid );
        else
            sprintf ( values, "%s, %d", values, keyid );
    }

    db_if_insert_or_update ( statement, values, "Keypad", id );

    return 0;

} /* end input_write_keypad */

/******************************************************************************
**  Function   :  input_write_player                                            
**                                                                    
**  Objective  :  This function writes player parameters to the profile
**                                                   
**  Parameters :  statement - SQL to check for existence of this entry
**                player    - struct containing values to write
**                id        - returns the id of this entry in the table
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_write_player ( e_machine_type m_type, char *statement, 
                                        t_atari_player *player, int *id ) {

    char values[512];
    char tstatement[512];
    int keyid;

    *id = 0;

    db_if_exec_sql ( statement, db_if_id_callback, id );

    sprintf ( values, "%d, %d, %d, %d", player->control_type, 
                player->stick.deadzone, player->stick.sensitivity, player->stick.simulate_analog );

    sprintf ( tstatement, "SELECT StickLeft FROM Player WHERE PlayerID= %d", *id );
    input_write_common ( tstatement, &player->stick.direction[DIR_LEFT], &keyid );
    sprintf ( values, "%s, %d", values, keyid );

    sprintf ( tstatement, "SELECT StickRight FROM Player WHERE PlayerID= %d", *id );
    input_write_common ( tstatement, &player->stick.direction[DIR_RIGHT], &keyid );
    sprintf ( values, "%s, %d", values, keyid );

    sprintf ( tstatement, "SELECT StickUp FROM Player WHERE PlayerID= %d", *id );
    input_write_common ( tstatement, &player->stick.direction[DIR_UP], &keyid );
    sprintf ( values, "%s, %d", values, keyid );

    sprintf ( tstatement, "SELECT StickDown FROM Player WHERE PlayerID= %d", *id );
    input_write_common ( tstatement, &player->stick.direction[DIR_DOWN], &keyid );
    sprintf ( values, "%s, %d", values, keyid );

    if ( m_type != MACHINE_TYPE_5200 ) {
        sprintf ( tstatement, "SELECT Paddle0CounterClockwise FROM Player WHERE PlayerID= %d", *id );
        input_write_common ( tstatement, &player->paddles.direction[DIR_LEFT], &keyid );
        sprintf ( values, "%s, %d", values, keyid );

        sprintf ( tstatement, "SELECT Paddle0Clockwise FROM Player WHERE PlayerID= %d", *id );
        input_write_common ( tstatement, &player->paddles.direction[DIR_RIGHT], &keyid );
        sprintf ( values, "%s, %d", values, keyid );

        sprintf ( tstatement, "SELECT Paddle1CounterClockwise FROM Player WHERE PlayerID= %d", *id );
        input_write_common ( tstatement, &player->paddles.direction[DIR_UP], &keyid );
        sprintf ( values, "%s, %d", values, keyid );

        sprintf ( tstatement, "SELECT Paddle1Clockwise FROM Player WHERE PlayerID= %d", *id );
        input_write_common ( tstatement, &player->paddles.direction[DIR_DOWN], &keyid );
        sprintf ( values, "%s, %d", values, keyid );
    }
    else {
        strcat ( values, ", NULL, NULL, NULL, NULL" );
    }

    sprintf ( tstatement, "SELECT BottomButton FROM Player WHERE PlayerID= %d", *id );
    input_write_common ( tstatement, &player->bottom_button, &keyid );
    sprintf ( values, "%s, %d", values, keyid );

    if ( m_type == MACHINE_TYPE_5200 ) {
        sprintf ( tstatement, "SELECT TopButton FROM Player WHERE PlayerID= %d", *id );
        input_write_common ( tstatement, &player->top_button, &keyid );
        sprintf ( values, "%s, %d", values, keyid );
    }
    else {
        strcat ( values, ", NULL" );
    }

    sprintf ( tstatement, "SELECT KeypadID FROM Player WHERE PlayerID= %d", *id );
    input_write_keypad ( tstatement, player->keypad, &keyid );
    sprintf ( values, "%s, %d", values, keyid );

    db_if_insert_or_update ( statement, values, "Player", id );

    return 0;

} /* end input_write_player */

int input_string_callback ( void *input, int argc, char **argv, char **azColName ) {

    int i;
    char *string = input;

    if ( argc < 2 )
        return -1;

    i = 0;
    if ( argv[++i] ) strcpy ( string, argv[i] );

    return 0;

} /* end input_string_callback */

/******************************************************************************
**  Function   :  input_write_keyboard                                            
**                                                                    
**  Objective  :  This function writes player parameters to the profile
**                                                   
**  Parameters :  statement - SQL to check for existence of this entry
**                keyboard  - struct containing values to write
**                id        - returns the id of this entry in the table
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_write_keyboard ( char *statement, t_atari_common *keyboard, int *id ) {

    int keyid, row_value;
    int i;
    char string[50];
    char values[1024];
    char tstatement[512];

    *id = 0;

    db_if_exec_sql ( statement, db_if_id_callback, id );

    for ( i = 0; i < 54; ++i ) {

        sprintf ( tstatement, "SELECT * FROM KeyboardCodes WHERE ROWID= %d", i+1 );
        db_if_exec_sql ( tstatement, db_if_id_callback, &row_value );
        sprintf ( tstatement, "SELECT * FROM KeyboardCodes WHERE ROWID= %d", i+1 );
        db_if_exec_sql ( tstatement, input_string_callback, string );

        sprintf ( tstatement, "SELECT %s FROM Keyboard WHERE KeyboardID= %d", string, *id );
        input_write_common ( tstatement, &keyboard[row_value], &keyid );
        if ( i == 0 )
            sprintf ( values, "%d", keyid );
        else
            sprintf ( values, "%s, %d", values, keyid );
    }

    db_if_insert_or_update ( statement, values, "Keyboard", id );

    return 0;

} /* end input_write_keyboard */

/******************************************************************************
**  Function   :  input_save_profile                                            
**                                                                    
**  Objective  :  This function saves the input configuration to database
**                                                   
**  Parameters :  name    - name of input profile to save
**                p_input - input structure containing info to save
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_save_profile ( const char *name, t_atari_input *p_input ) {

    int i,id;
    int status = 0;
    char statement[257];
    char tstatement[257];
    char *p_tmp[2];

    db_if_exec_sql ( "BEGIN", 0, 0 );

    /*
     * Put together our update or insert string.
     *   At the same time, update or insert players and keyboard
     */
    sprintf ( statement, "INSERT OR REPLACE INTO Input VALUES ( '%s', '%s', '%d', '%d'",
                p_tmp[0]=util_replace_and_copy(name, "'", "''" ), 
                p_tmp[1]=util_replace_and_copy(p_input->description, "'", "''" ), 
                p_input->machine_type, p_input->mouse_speed ); 

    free ( p_tmp[0] );
    free ( p_tmp[1] );

    for ( i = 0; i < 4; ++i ) {
        sprintf ( tstatement, "SELECT Player%d FROM Input WHERE Name='%s'", i, name );
        input_write_player ( p_input->machine_type, tstatement, &p_input->players[i], &id );
        sprintf ( statement, "%s, %d", statement, id );
    }

    if ( p_input->machine_type != MACHINE_TYPE_5200 ) {
        sprintf ( tstatement, "SELECT KeyboardID FROM Input WHERE Name='%s'", name );
        input_write_keyboard ( tstatement, p_input->keyboard, &id );
        sprintf ( statement, "%s, %d", statement, id );

        sprintf ( tstatement, "SELECT StartKey FROM Input WHERE Name='%s'", name );
        input_write_common ( tstatement, &p_input->start_key, &id );
        sprintf ( statement, "%s, %d", statement, id );

        sprintf ( tstatement, "SELECT SelectKey FROM Input WHERE Name='%s'", name );
        input_write_common ( tstatement, &p_input->select_key, &id );
        sprintf ( statement, "%s, %d", statement, id );
    
        sprintf ( tstatement, "SELECT OptionKey FROM Input WHERE Name='%s'", name );
        input_write_common ( tstatement, &p_input->option_key, &id );
        sprintf ( statement, "%s, %d", statement, id );
    
        sprintf ( tstatement, "SELECT CtrlKey FROM Input WHERE Name='%s'", name );
        input_write_common ( tstatement, &p_input->ctrl_key, &id );
        sprintf ( statement, "%s, %d", statement, id );

        sprintf ( tstatement, "SELECT ShiftKey FROM Input WHERE Name='%s'", name );
        input_write_common ( tstatement, &p_input->shift_key, &id );
        sprintf ( statement, "%s, %d", statement, id );

        sprintf ( tstatement, "SELECT BreakKey FROM Input WHERE Name='%s'", name );
        input_write_common ( tstatement, &p_input->break_key, &id );
        sprintf ( statement, "%s, %d", statement, id );

        sprintf ( tstatement, "SELECT ResetKey FROM Input WHERE Name='%s'", name );
        input_write_common ( tstatement, &p_input->reset_key, &id );
        sprintf ( statement, "%s, %d )", statement, id );
    }
    else {
        strcat ( statement, ", NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL )" );
    }

    status = db_if_exec_sql ( statement, 0, 0 );

    db_if_exec_sql ( "END", 0, 0 );

    return status;

} /* end input_save_profile */

/******************************************************************************
**  Function   :  input_convert_sdl_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  ctrl      - pointer to common control struct
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_convert_sdl_callback ( void *ctrl, int argc, char **argv, char **azColName ) {

    char statement[100];

    if ( argc < 6 )
        return -1;

    sprintf ( statement, "UPDATE InputList SET DeviceNum=%d, PartNum=%d WHERE InputID=%s", 
               key_mod_convert(atoi(argv[2])), key_convert(atoi(argv[4])), argv[0] );
    db_if_exec_sql ( statement, NULL, NULL );

    return 0;

} /* end input_find_common_callback */

/******************************************************************************
**  Function   :  input_convert_sdl_keycodes
**                                                                    
**  Objective  :  This function finds keyboard key entries and then updates
**                  to SDL2 in the callback
**                                                   
**  Parameters :  none
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_convert_sdl_keycodes () {

    char statement[50];

    strcpy ( statement, "SELECT * FROM InputList WHERE Device=0 AND Part=4" );
    db_if_exec_sql ( statement, input_convert_sdl_callback, NULL );

    return 0;

} /* end input_convert_sdl_keycodes */

/******************************************************************************
**  Function   :  input_find_common_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  ctrl      - pointer to common control struct
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_find_common_callback ( void *ctrl, int argc, char **argv, char **azColName ) {

    t_atari_common *p_ctrl = ctrl;
    int i;

    if ( argc < 12 )
        return -1;

    /*
     * THIS DEPENDS ON MATCHING THE ORDER IN THE DATABASE!
     */
    i = 0;
    if ( argv[++i] ) p_ctrl->device[0] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->device_num[0] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->part_type[0] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->part_num[0] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->direction[0] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->combo = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->device[1] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->device_num[1] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->part_type[1] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->part_num[1] = atoi ( argv[i] );
    if ( argv[++i] ) p_ctrl->direction[1] = atoi ( argv[i] );

    return 0;

} /* end input_find_common_callback */

/******************************************************************************
**  Function   :  input_find_common                                            
**                                                                    
**  Objective  :  This function finds common contol parameters in the profile
**                                                   
**  Parameters :  id   - id of control to find in database table
**                ctrl - struct to store found values
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_find_common ( int id, t_atari_common *ctrl ) {

    char statement[50];

    sprintf ( statement, "SELECT * FROM InputList WHERE InputID=%d", id );
    db_if_exec_sql ( statement, input_find_common_callback, ctrl );

    return 0;

} /* end input_find_common */

/******************************************************************************
**  Function   :  input_write_common                                            
**                                                                    
**  Objective  :  This function writes common contol parameters to the profile
**                                                   
**  Parameters :  statement - SQL to check for existence of this entry
**                ctrl      - struct containing values to write
**                id        - returns the id of this entry in the table
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int input_write_common ( char *statement, t_atari_common *ctrl, int *id ) {

    char values[256];

    *id = 0;

    sprintf ( values, "%d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d", 
                ctrl->device[0], ctrl->device_num[0], 
                ctrl->part_type[0], ctrl->part_num[0], ctrl->direction[0],
                ctrl->combo, ctrl->device[1], ctrl->device_num[1],
                ctrl->part_type[1], ctrl->part_num[1], ctrl->direction[1]
            );

    db_if_insert_or_update ( statement, values, "InputList", id );

    return 0;

} /* end input_write_common */

/******************************************************************************
**  Function   :  input_get_key_string                                            
**                                                                    
**  Objective  :  This function gets the name of the key from the value
**                                                   
**  Parameters :  key  - input key number
**                name - name of key to return
**                 
**  return     :  keyname
******************************************************************************/ 
char * input_get_key_string ( int key, char *name ) {

    switch ( key ) {

        case A800_KEY_L:      return strcpy(name,"key_l"); break;
        case A800_KEY_J:      return strcpy(name,"key_j"); break;
        case A800_KEY_SEMI:   return strcpy(name,"key_semicolon"); break;
        case A800_KEY_K:      return strcpy(name,"key_k"); break;
        case A800_KEY_PLUS:   return strcpy(name,"key_plus"); break;
        case A800_KEY_STAR:   return strcpy(name,"key_star"); break;
        case A800_KEY_O:      return strcpy(name,"key_o"); break;
        case A800_KEY_P:      return strcpy(name,"key_p"); break;
        case A800_KEY_U:      return strcpy(name,"key_u"); break;
        case A800_KEY_RETURN: return strcpy(name,"key_return"); break;
        case A800_KEY_I:      return strcpy(name,"key_i"); break;
        case A800_KEY_MINUS:  return strcpy(name,"key_-"); break;
        case A800_KEY_EQUALS: return strcpy(name,"key_equals"); break;
        case A800_KEY_V:      return strcpy(name,"key_v"); break;
        case A800_KEY_HELP:   return strcpy(name,"key_help"); break;
        case A800_KEY_C:      return strcpy(name,"key_c"); break;
        case A800_KEY_B:      return strcpy(name,"key_b"); break;
        case A800_KEY_X:      return strcpy(name,"key_x"); break;
        case A800_KEY_Z:      return strcpy(name,"key_z"); break;
        case A800_KEY_4:      return strcpy(name,"key_4"); break;
        case A800_KEY_3:      return strcpy(name,"key_3"); break;
        case A800_KEY_6:      return strcpy(name,"key_6"); break;
        case A800_KEY_ESC:    return strcpy(name,"key_esc"); break;
        case A800_KEY_5:      return strcpy(name,"key_5"); break;
        case A800_KEY_2:      return strcpy(name,"key_2"); break;
        case A800_KEY_1:      return strcpy(name,"key_1"); break;
        case A800_KEY_COMMA:  return strcpy(name,"key_comma"); break;
        case A800_KEY_SPACE:  return strcpy(name,"key_space"); break;
        case A800_KEY_PERIOD: return strcpy(name,"key_."); break;
        case A800_KEY_N:      return strcpy(name,"key_n"); break;
        case A800_KEY_M:      return strcpy(name,"key_m"); break;
        case A800_KEY_SLASH:  return strcpy(name,"key_slash"); break;
        case A800_KEY_FUJI:   return strcpy(name,""); break;
        case A800_KEY_R:      return strcpy(name,"key_r"); break;
        case A800_KEY_E:      return strcpy(name,"key_e"); break;
        case A800_KEY_Y:      return strcpy(name,"key_y"); break;
        case A800_KEY_TAB:    return strcpy(name,"key_tab"); break;
        case A800_KEY_T:      return strcpy(name,"key_t"); break;
        case A800_KEY_W:      return strcpy(name,"key_w"); break;
        case A800_KEY_Q:      return strcpy(name,"key_q"); break;
        case A800_KEY_9:      return strcpy(name,"key_9"); break;
        case A800_KEY_0:      return strcpy(name,"key_0"); break;
        case A800_KEY_7:      return strcpy(name,"key_7"); break;
        case A800_KEY_BKSP:   return strcpy(name,"key_bksp"); break;
        case A800_KEY_8:      return strcpy(name,"key_8"); break;
        case A800_KEY_LESST:  return strcpy(name,"key_lessthan"); break;
        case A800_KEY_MORET:  return strcpy(name,"key_greaterthan"); break;
        case A800_KEY_F:      return strcpy(name,"key_f"); break;
        case A800_KEY_H:      return strcpy(name,"key_h"); break;
        case A800_KEY_D:      return strcpy(name,"key_d"); break;
        case A800_KEY_CAPS:   return strcpy(name,"key_caps"); break;
        case A800_KEY_G:      return strcpy(name,"key_g"); break;
        case A800_KEY_S:      return strcpy(name,"key_s"); break;
        case A800_KEY_A:      return strcpy(name,"key_a"); break;
        default: return strcpy(name,""); break;

     } /* end switch key */

} /* end input_get_key_string */

/************************************************************************
 *
 *  Assign PC inputs to Atari Inputs and setup input functions
 *
 ************************************************************************/
/******************************************************************************
**  Function   :  input_preload_controls
**                            
**  Objective  :  This function preloads all control functions with do nothing
**                frame
**
**  Parameters :  NONE
**                                                
**  return     :  0
**      
******************************************************************************/
int input_preload_controls ( void ) {
    
    int i, j, k;

    for ( i = 0; i < MAX_PC_KEY; ++i ) {
        g_pc_input.keyboard[i].key_up_function = input_do_nothing;
        g_pc_input.keyboard[i].key_down_function = input_do_nothing;
        g_pc_input.keyboard[i].player = 0;
        g_pc_input.keyboard[i].atari_key = 0;
        g_pc_input.keyboard[i].direction = 0;
    }
    for ( i = 0; i < UI_KEY_MAX; ++i ) {
        g_pc_input.ui_keyboard[i].key_up_function = input_do_nothing;
        g_pc_input.ui_keyboard[i].key_down_function = input_do_nothing;
        g_pc_input.ui_keyboard[i].modifier = 0;
        g_pc_input.ui_keyboard[i].key = 0;
    }

    for ( i = 0; i < MAX_PC_DEVICE; ++i ) {
        for ( j = 0; j < MAX_PC_PART; ++j ) {

            for ( k = 0; k < MAX_DIR; ++k ) {
                g_pc_input.pov_hat[i][j].hat_pressed_function[k] = input_do_nothing;
                g_pc_input.pov_hat[i][j].hat_not_pressed_function[k] = input_do_nothing;
                g_pc_input.pov_hat[i][j].player[k] = 0;
            }

            g_pc_input.analog_axis[i][j].axis_pos_function = input_do_nothing;
            g_pc_input.analog_axis[i][j].axis_neg_function = input_do_nothing;
            g_pc_input.analog_axis[i][j].player = 0;

            g_pc_input.trackball[i][j].x_axis.axis_pos_function = input_do_nothing;
            g_pc_input.trackball[i][j].x_axis.axis_neg_function = input_do_nothing;
            g_pc_input.trackball[i][j].y_axis.axis_pos_function = input_do_nothing;
            g_pc_input.trackball[i][j].y_axis.axis_neg_function = input_do_nothing;
            g_pc_input.trackball[i][j].x_axis.player = 0;
            g_pc_input.trackball[i][j].y_axis.player = 0;

            g_pc_input.button[i][j].button_up_function = input_do_nothing;
            g_pc_input.button[i][j].button_down_function = input_do_nothing;
            g_pc_input.button[i][j].player = 0;
            g_pc_input.button[i][j].atari_key = 0;
            g_pc_input.button[i][j].direction = 0;
        }
    }

    g_pc_input.mouse.x_axis.axis_pos_function = input_do_nothing;
    g_pc_input.mouse.x_axis.axis_neg_function = input_do_nothing;
    g_pc_input.mouse.y_axis.axis_pos_function = input_do_nothing;
    g_pc_input.mouse.y_axis.axis_neg_function = input_do_nothing;
    g_pc_input.mouse.x_axis.player = 0;
    g_pc_input.mouse.y_axis.player = 0;
    g_pc_input.mouse.speed = 0;
    g_pc_input.mouse.x_axis.direction_pos = 0;
    g_pc_input.mouse.y_axis.direction_pos = 0;
    g_pc_input.mouse.x_axis.direction_neg = 0;
    g_pc_input.mouse.y_axis.direction_neg = 0;

    for ( i = 0; i < 4; ++i ) {
        g_pc_input.mouse_button[i].button_up_function = input_do_nothing;
        g_pc_input.mouse_button[i].button_down_function = input_do_nothing;
        g_pc_input.mouse_button[i].player = 0;
        g_pc_input.mouse_button[i].atari_key = 0;
        g_pc_input.mouse_button[i].direction = 0;
    }

    return 0;

} /* end input_preload_controls */

/*
 * Combo Macros
 */

#define KEYBOARD_CHECK_COMBO_OFF( ctrl_combo ) \
            if ( ctrl_combo ) {\
                input_set_combo_off_function ( \
                        &g_pc_input.keyboard[tmp_part_num].key_up_function,\
                        g_pc_input.keyboard[tmp_part_num].player,\
                        g_pc_input.keyboard[tmp_part_num].direction,\
                        g_pc_input.keyboard[tmp_part_num].atari_key);\
                input_set_combo_off_function ( \
                        &g_pc_input.keyboard[tmp_part_num].key_down_function, \
                        g_pc_input.keyboard[tmp_part_num].player,\
                        g_pc_input.keyboard[tmp_part_num].direction,\
                        g_pc_input.keyboard[tmp_part_num].atari_key);\
            }

#define KEYBOARD_CHECK_COMBO_ON( ctrl_combo, ctrl_addr ) \
            /* Store Combo */\
            if ( ctrl_combo ) {\
                input_add_combo_function ( ctrl_addr, \
                        &g_pc_input.keyboard[tmp_part_num].key_up_function, \
                        g_pc_input.keyboard[tmp_part_num].key_up_function, \
                        &g_pc_input.keyboard[tmp_part_num].player,\
                        &g_pc_input.keyboard[tmp_part_num].direction,\
                        &g_pc_input.keyboard[tmp_part_num].atari_key);\
                input_add_combo_function ( ctrl_addr, \
                        &g_pc_input.keyboard[tmp_part_num].key_down_function, \
                        g_pc_input.keyboard[tmp_part_num].key_down_function,\
                        &g_pc_input.keyboard[tmp_part_num].player,\
                        &g_pc_input.keyboard[tmp_part_num].direction,\
                        &g_pc_input.keyboard[tmp_part_num].atari_key);\
            }\
            \
            /* Else Store this as combo off */\
            else {\
                input_update_combo_off_function ( \
                        &g_pc_input.keyboard[tmp_part_num].key_up_function, \
                        g_pc_input.keyboard[tmp_part_num].player,\
                        g_pc_input.keyboard[tmp_part_num].direction,\
                        g_pc_input.keyboard[tmp_part_num].atari_key);\
                input_update_combo_off_function ( \
                        &g_pc_input.keyboard[tmp_part_num].key_down_function,\
                        g_pc_input.keyboard[tmp_part_num].player,\
                        g_pc_input.keyboard[tmp_part_num].direction,\
                        g_pc_input.keyboard[tmp_part_num].atari_key);\
            }

#define JOYAXIS_CHECK_COMBO_OFF( ctrl_combo ) \
            /* If combo store current function as off */\
            if ( ctrl_combo ) {\
                input_set_combo_off_function ( tmp_func_addr,\
                 g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player, \
                 g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_pos,\
                 g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_neg );\
            }

#define JOYAXIS_CHECK_COMBO_ON( ctrl_combo, ctrl_addr) \
                /* Store Combo */ \
                if ( ctrl_combo ) { \
                    input_add_combo_function ( ctrl_addr,  \
                      tmp_func_addr, *tmp_func_addr, \
                      &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player, \
                      &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_pos,\
                      &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_neg );\
                } \
             \
                /* Else Store this as combo off */ \
                else { \
                    input_update_combo_off_function ( tmp_func_addr, \
                      g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player, \
                      g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_pos,\
                      g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_neg );\
                } \

#define JOYHAT_CHECK_COMBO_OFF( ctrl_combo ) \
                /* If combo store current function as off */\
                if ( ctrl_combo ) {\
                    input_set_combo_off_function ( \
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir], \
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir]);\
                    input_set_combo_off_function ( \
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir], \
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir]);\
                }

#define JOYHAT_CHECK_COMBO_ON( ctrl_combo, ctrl_addr ) \
                /* Store Combo */ \
                if ( ctrl_combo ) { \
                    input_add_combo_function ( ctrl_addr, \
                            &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir],\
                            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir],\
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir],\
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir],\
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir]);\
                    input_add_combo_function ( ctrl_addr, \
                            &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir],\
                            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir],\
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir],\
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir],\
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir]);\
                }\
                /* Else Store this as combo off */\
                else {\
                    input_update_combo_off_function ( \
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir], \
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir]);\
                    input_update_combo_off_function ( \
                      &g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir], \
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir],\
                      g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir]);\
                }

#define JOYBALL_CHECK_COMBO_OFF( ctrl_combo ) \
                /* If combo store current function as off */\
                if ( ctrl_combo ) \
                    input_set_combo_off_function ( tmp_func_addr, \
                        tmp_axis->player,\
                        tmp_axis->direction_pos,\
                        tmp_axis->direction_neg );

#define JOYBALL_CHECK_COMBO_ON( ctrl_combo, ctrl_addr ) \
                /* Store Combo */\
                if ( ctrl_combo ) {\
                    input_add_combo_function ( ctrl_addr,\
                        tmp_func_addr, *tmp_func_addr, \
                        &(tmp_axis->player),\
                        &(tmp_axis->direction_pos),\
                        &(tmp_axis->direction_neg) );\
                }\
                /* Else Store this as combo off */ \
                else { \
                    input_update_combo_off_function ( tmp_func_addr, \
                        tmp_axis->player,\
                        tmp_axis->direction_pos,\
                        tmp_axis->direction_neg );\
                }

#define JOYBUTTON_CHECK_COMBO_OFF( ctrl_combo ) \
                /* If combo store current function as off */\
                if ( ctrl_combo ) { \
                    input_set_combo_off_function (  \
                       &g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function, \
                       g_pc_input.button[tmp_device_num][tmp_part_num].player,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].direction,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].atari_key);\
                    input_set_combo_off_function (  \
                       &g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function, \
                       g_pc_input.button[tmp_device_num][tmp_part_num].player,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].direction,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].atari_key);\
                }

#define JOYBUTTON_CHECK_COMBO_ON( ctrl_combo, ctrl_addr ) \
                /* Store Combo */\
                if ( ctrl_combo ) {\
                    input_add_combo_function ( ctrl_addr, \
                       &g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function, \
                       g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function ,\
                       &g_pc_input.button[tmp_device_num][tmp_part_num].player,\
                       &g_pc_input.button[tmp_device_num][tmp_part_num].direction,\
                       &g_pc_input.button[tmp_device_num][tmp_part_num].atari_key);\
                    input_add_combo_function ( ctrl_addr, \
                       &g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function, \
                       g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function,\
                       &g_pc_input.button[tmp_device_num][tmp_part_num].player,\
                       &g_pc_input.button[tmp_device_num][tmp_part_num].direction,\
                       &g_pc_input.button[tmp_device_num][tmp_part_num].atari_key);\
                }\
                /* Else Store this as combo off */\
                else {\
                    input_update_combo_off_function (  \
                       &g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function, \
                       g_pc_input.button[tmp_device_num][tmp_part_num].player,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].direction,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].atari_key);\
                    input_update_combo_off_function (  \
                       &g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function, \
                       g_pc_input.button[tmp_device_num][tmp_part_num].player,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].direction,\
                       g_pc_input.button[tmp_device_num][tmp_part_num].atari_key);\
                }\

#define MOUSEBUTTON_CHECK_COMBO_OFF( ctrl_combo ) \
                /* If combo store current function as off */\
                if ( ctrl_combo ) { \
                    input_set_combo_off_function (  \
                       &g_pc_input.mouse_button[tmp_part_num].button_up_function, \
                       g_pc_input.mouse_button[tmp_part_num].player,\
                       g_pc_input.mouse_button[tmp_part_num].direction,\
                       g_pc_input.mouse_button[tmp_part_num].atari_key);\
                    input_set_combo_off_function (  \
                       &g_pc_input.mouse_button[tmp_part_num].button_down_function, \
                       g_pc_input.mouse_button[tmp_part_num].player,\
                       g_pc_input.mouse_button[tmp_part_num].direction,\
                       g_pc_input.mouse_button[tmp_part_num].atari_key);\
                }

#define MOUSEBUTTON_CHECK_COMBO_ON( ctrl_combo, ctrl_addr ) \
                /* Store Combo */\
                if ( ctrl_combo ) {\
                    input_add_combo_function ( ctrl_addr, \
                       &g_pc_input.mouse_button[tmp_part_num].button_up_function, \
                       g_pc_input.mouse_button[tmp_part_num].button_up_function ,\
                       &g_pc_input.mouse_button[tmp_part_num].player,\
                       &g_pc_input.mouse_button[tmp_part_num].direction,\
                       &g_pc_input.mouse_button[tmp_part_num].atari_key);\
                    input_add_combo_function ( ctrl_addr, \
                       &g_pc_input.mouse_button[tmp_part_num].button_down_function, \
                       g_pc_input.mouse_button[tmp_part_num].button_down_function,\
                       &g_pc_input.mouse_button[tmp_part_num].player,\
                       &g_pc_input.mouse_button[tmp_part_num].direction,\
                       &g_pc_input.mouse_button[tmp_part_num].atari_key);\
                }\
                /* Else Store this as combo off */\
                else {\
                    input_update_combo_off_function (  \
                       &g_pc_input.mouse_button[tmp_part_num].button_up_function, \
                       g_pc_input.mouse_button[tmp_part_num].player,\
                       g_pc_input.mouse_button[tmp_part_num].direction,\
                       g_pc_input.mouse_button[tmp_part_num].atari_key);\
                    input_update_combo_off_function (  \
                       &g_pc_input.mouse_button[tmp_part_num].button_down_function, \
                       g_pc_input.mouse_button[tmp_part_num].player,\
                       g_pc_input.mouse_button[tmp_part_num].direction,\
                       g_pc_input.mouse_button[tmp_part_num].atari_key);\
                }\

/******************************************************************************
**  Function   :  input_setup_stick
**                            
**  Objective  :  This function sets up controller options for the emulated
**                stick.  For 800 paddles, direction is reversed
**
**  Parameters :  player - which controller to setup for
**                stick  - player array (5200 stick or 800 paddle)
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_stick ( int player, t_atari_stick *stick ) {

    int j, tmp_dir, tmp_device_num, tmp_part_num;
    int i = player;
    t_atari_common *ctrl = stick->direction;
    t_pc_analog_axis *tmp_axis;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    /*
     * Check each direction of Atari stick
     */
    for ( j = DIR_LEFT; j < MAX_DIR; ++j ) {

        /*
         * Check for Keyboard
         */
        if ( ctrl[j].device[0] == DEV_KEYBOARD ) {
            tmp_part_num =  pc_adjust_key_value ( ctrl[j].part_num[0] );
            if ( g_input.machine_type == MACHINE_TYPE_5200 )
                g_pc_input.keyboard[tmp_part_num].direction = j;
            else
                g_pc_input.keyboard[tmp_part_num].direction = (j ^ 1);

            KEYBOARD_CHECK_COMBO_OFF( ctrl[j].combo )

            if ( stick->simulate_analog ) {
                    g_pc_input.keyboard[tmp_part_num].key_up_function = input_key_up_stick_sa;
                    g_pc_input.keyboard[tmp_part_num].key_down_function = input_key_down_stick_sa;
            }
            else {
                    g_pc_input.keyboard[tmp_part_num].key_up_function = input_key_up_stick;
                    g_pc_input.keyboard[tmp_part_num].key_down_function = input_key_down_stick;
            }
            g_pc_input.keyboard[tmp_part_num].player = i;
            g_pc_input.keyboard[tmp_part_num].atari_key = 0;
            
            KEYBOARD_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

        } /* end if keyboard */

        /*
         * Check for Joystick
         */
        tmp_device_num = ctrl[j].device_num[0]; 
        tmp_part_num = ctrl[j].part_num[0]; 
        tmp_dir = ctrl[j].direction[0]; 
        if ( ctrl[j].device[0] == DEV_JOYSTICK || ctrl[j].device[0] == DEV_CONTROLLER) {

            if ( ctrl[j].direction[0] == DIR_PLUS )
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
            else
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;

            if ( ctrl[j].part_type[0] == PART_TYPE_AXIS ) {

                JOYAXIS_CHECK_COMBO_OFF( ctrl[j].combo )

                if (stick->simulate_analog) {
                    *tmp_func_addr = input_joystick_pots_sa;
                }
                else {
                    *tmp_func_addr = input_joystick_pots;
                }

                g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;
                if ( ctrl[j].direction[0] == DIR_PLUS ) {
                    if ( g_input.machine_type == MACHINE_TYPE_5200 )
                        g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_pos = j;
                    else
                        g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_pos = (j ^ 1);
                }
                else {
                    if ( g_input.machine_type == MACHINE_TYPE_5200 )
                        g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_neg = j;
                    else
                        g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_neg = (j ^ 1);
                }

                JOYAXIS_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }

            if ( ctrl[j].part_type[0] == PART_TYPE_BALL ) {

                if ( g_input.machine_type == MACHINE_TYPE_5200 )
                    tmp_dir = j;
                else
                    tmp_dir = (j ^ 1);

                if ( !ctrl[j].part_num[0] )
                    tmp_axis = &g_pc_input.trackball[tmp_device_num][tmp_part_num].x_axis;
                else
                    tmp_axis = &g_pc_input.trackball[tmp_device_num][tmp_part_num].y_axis;

                if ( ctrl[j].direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &(tmp_axis->axis_pos_function);
                    tmp_direction = &(tmp_axis->direction_pos);
                }
                else {
                    tmp_func_addr = &(tmp_axis->axis_neg_function);
                    tmp_direction = &(tmp_axis->direction_neg);
                }

                JOYBALL_CHECK_COMBO_OFF( ctrl[j].combo ) 

                *tmp_func_addr = input_mouse_axis;
                *tmp_direction = tmp_dir;
            
                tmp_axis->player = i;
                g_pc_input.trackball[tmp_device_num][tmp_part_num].speed = g_input.mouse_speed;
                JOYBALL_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] ) 
            }

            if ( ctrl[j].part_type[0] == PART_TYPE_HAT ) {

                JOYHAT_CHECK_COMBO_OFF( ctrl[j].combo )
               
                if ( stick->simulate_analog ) {
                    g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_key_up_stick_sa;
                    g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_key_down_stick_sa;
                }
                else {
                    g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_key_up_stick;
                    g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_key_down_stick;
                }
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;
                if ( g_input.machine_type == MACHINE_TYPE_5200 )
                    g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir] = j;
                else
                    g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir] = (j ^ 1);

                JOYHAT_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

            } /* end if hat on joystick */

            if ( ctrl[j].part_type[0] == PART_TYPE_BUTTON ) {

                JOYBUTTON_CHECK_COMBO_OFF( ctrl[j].combo )

                if (stick->simulate_analog) {
                    g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_key_up_stick_sa;
                    g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_key_down_stick_sa;
                }
                else {
                    g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_key_up_stick;
                    g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_key_down_stick;
                }
                g_pc_input.button[tmp_device_num][tmp_part_num].player = i;
                if ( g_input.machine_type == MACHINE_TYPE_5200 )
                    g_pc_input.button[tmp_device_num][tmp_part_num].direction = j;
                else
                    g_pc_input.button[tmp_device_num][tmp_part_num].direction = (j ^ 1);

                JOYBUTTON_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

            } /* end if button on joystick */

        } /* end if joystick */

        if ( ctrl[j].device[0] == DEV_MOUSE ) {

            if ( g_input.machine_type == MACHINE_TYPE_5200 )
                tmp_dir = j;
            else
                tmp_dir = (j ^ 1);

            if ( !ctrl[j].part_num[0] )
                tmp_axis = &g_pc_input.mouse.x_axis;
            else
                tmp_axis = &g_pc_input.mouse.y_axis;

            if ( ctrl[j].direction[0] == DIR_PLUS ) {
                tmp_func_addr = &(tmp_axis->axis_pos_function);
                tmp_direction = &(tmp_axis->direction_pos);
            }
            else {
                tmp_func_addr = &(tmp_axis->axis_neg_function);
                tmp_direction = &(tmp_axis->direction_neg);
            }

            JOYBALL_CHECK_COMBO_OFF( ctrl[j].combo ) 

            *tmp_func_addr = input_mouse_axis;
            *tmp_direction = tmp_dir;

            tmp_axis->player = i;
            g_pc_input.mouse.speed = g_input.mouse_speed;

            JOYBALL_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] ) 

        } /* end if mouse */

    } /* end for stick direction */

    return 0;

} /* input_setup_stick */

/******************************************************************************
**  Function   :  input_setup_digital_stick
**                            
**  Objective  :  This function sets up controller options for the emulated
**                stick (for the 800 digital stick)
**
**  Parameters :  player - which controller to setup for
**                stick  - player array (5200 stick or 800 paddle)
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_digital_stick ( int player, t_atari_stick *stick ) {

    int j, tmp_dir, tmp_device_num, tmp_part_num;
    int i = player;
    t_atari_common *ctrl = stick->direction;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    /*
     * Check each direction of Atari stick
     */
    for ( j = DIR_LEFT; j < MAX_DIR; ++j ) {

        /*
         * Check for Keyboard
         */
        if ( ctrl[j].device[0] == DEV_KEYBOARD ) {
            tmp_part_num =  pc_adjust_key_value ( ctrl[j].part_num[0] );

            KEYBOARD_CHECK_COMBO_OFF( ctrl[j].combo )

            g_pc_input.keyboard[tmp_part_num].direction = j;
            g_pc_input.keyboard[tmp_part_num].key_up_function = input_key_up_digital_stick;
            g_pc_input.keyboard[tmp_part_num].key_down_function = input_key_down_digital_stick;
            g_pc_input.keyboard[tmp_part_num].player = i;
            g_pc_input.keyboard[tmp_part_num].atari_key = 0;

            KEYBOARD_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

        } /* end if keyboard */

        /*
         * Check for Joystick
         */
        tmp_device_num = ctrl[j].device_num[0]; 
        tmp_part_num = ctrl[j].part_num[0]; 
        tmp_dir = ctrl[j].direction[0]; 
        if ( ctrl[j].device[0] == DEV_JOYSTICK || ctrl[j].device[0] == DEV_CONTROLLER) {

            if ( ctrl[j].part_type[0] == PART_TYPE_AXIS ) {
                if ( ctrl[j].direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_pos;
                }
                else {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_neg;
                }

                JOYAXIS_CHECK_COMBO_OFF( ctrl[j].combo )

                *tmp_func_addr = input_axis_digital_stick;
                *tmp_direction = j;

                g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;

                JOYAXIS_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

            } /* end if axis on joystick */

            if ( ctrl[j].part_type[0] == PART_TYPE_HAT ) {

                JOYHAT_CHECK_COMBO_OFF( ctrl[j].combo )

                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_key_up_digital_stick;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_key_down_digital_stick;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir] = j;

                JOYHAT_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

            } /* end if hat on joystick */

            if ( ctrl[j].part_type[0] == PART_TYPE_BUTTON ) {

                JOYBUTTON_CHECK_COMBO_OFF( ctrl[j].combo )

                g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_key_up_digital_stick;
                g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_key_down_digital_stick;
                g_pc_input.button[tmp_device_num][tmp_part_num].player = i;
                g_pc_input.button[tmp_device_num][tmp_part_num].direction = j;

                JOYBUTTON_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            
            } /* end if button on joystick */

        } /* end if joystick */

    } /* end for stick direction */

    return 0;

} /* input_setup_digital_stick */

/******************************************************************************
**  Function   :  input_setup_trackball
**                            
**  Objective  :  This function sets up controller options for the emulated
**                trackball.
**
**  Parameters :  player - which controller to setup for
**                stick  - player array (5200 trackball)
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_trackball ( int player, t_atari_stick *stick ) {

    int j, tmp_dir, tmp_device_num, tmp_part_num;
    int i = player;
    t_atari_common *ctrl = stick->direction;
    t_pc_analog_axis *tmp_axis;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    /*
     * Check each direction of Atari stick
     */
    for ( j = DIR_LEFT; j < MAX_DIR; ++j ) {

        /*
         * Check for Keyboard
         */
        if ( ctrl[j].device[0] == DEV_KEYBOARD ) {

            tmp_part_num =  pc_adjust_key_value ( ctrl[j].part_num[0] );

            KEYBOARD_CHECK_COMBO_OFF( ctrl[j].combo )

            g_pc_input.keyboard[ctrl[j].part_num[0]].direction = j;
            g_pc_input.keyboard[tmp_part_num].key_up_function = input_key_up_stick;
            g_pc_input.keyboard[tmp_part_num].key_down_function = input_key_down_stick;
            g_pc_input.keyboard[tmp_part_num].player = i;
            g_pc_input.keyboard[tmp_part_num].atari_key = 0;

            KEYBOARD_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

        } /* end if keyboard */

        /*
         * Check for Joystick
         */
        tmp_device_num = ctrl[j].device_num[0]; 
        tmp_part_num = ctrl[j].part_num[0]; 
        tmp_dir = ctrl[j].direction[0]; 
        if ( ctrl[j].device[0] == DEV_JOYSTICK || ctrl[j].device[0] == DEV_CONTROLLER) {

            if ( ctrl[j].part_type[0] == PART_TYPE_AXIS ) {

                if ( ctrl[j].direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_pos;
                }
                else {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_neg;
                }

                JOYAXIS_CHECK_COMBO_OFF( ctrl[j].combo )

                *tmp_func_addr = input_joystick_pots;
                *tmp_direction = j;

                g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;

                JOYAXIS_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }

            if ( ctrl[j].part_type[0] == PART_TYPE_BALL ) {
                if ( g_input.machine_type == MACHINE_TYPE_5200 )
                    tmp_dir = j;
                else
                    tmp_dir = (j ^ 1);

                if ( !ctrl[j].part_num[0] )
                    tmp_axis = &g_pc_input.trackball[tmp_device_num][tmp_part_num].x_axis;
                else
                    tmp_axis = &g_pc_input.trackball[tmp_device_num][tmp_part_num].y_axis;

                if ( ctrl[j].direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &(tmp_axis->axis_pos_function);
                    tmp_direction = &(tmp_axis->direction_pos);
                }
                else {
                    tmp_func_addr = &(tmp_axis->axis_neg_function);
                    tmp_direction = &(tmp_axis->direction_neg);
                }

                JOYBALL_CHECK_COMBO_OFF( ctrl[j].combo ) 

                *tmp_func_addr = input_mouse_vel_axis;
                *tmp_direction = tmp_dir;

                tmp_axis->player = i;

                JOYBALL_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] ) 
            }

            if ( ctrl[j].part_type[0] == PART_TYPE_HAT ) {

                JOYHAT_CHECK_COMBO_OFF( ctrl[j].combo )

                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_key_up_stick;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_key_down_stick;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir] = j;

                JOYHAT_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )

            } /* end if hat on joystick */

            if ( ctrl[j].part_type[0] == PART_TYPE_BUTTON ) {

                JOYBUTTON_CHECK_COMBO_OFF( ctrl[j].combo )
                
                g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_key_up_stick;
                g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_key_down_stick;
                g_pc_input.button[tmp_device_num][tmp_part_num].player = i;
                g_pc_input.button[tmp_device_num][tmp_part_num].direction = j;

                JOYBUTTON_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
                            
            } /* end if button on joystick */

        } /* end if joystick */

        if ( ctrl[j].device[0] == DEV_MOUSE ) {

            if ( g_input.machine_type == MACHINE_TYPE_5200 )
                tmp_dir = j;
            else
                tmp_dir = (j ^ 1);

            if ( !ctrl[j].part_num[0] )
                tmp_axis = &g_pc_input.mouse.x_axis;
            else
                tmp_axis = &g_pc_input.mouse.y_axis;

            if ( ctrl[j].direction[0] == DIR_PLUS ) {
                tmp_func_addr = &(tmp_axis->axis_pos_function);
                tmp_direction = &(tmp_axis->direction_pos);
            }
            else {
                tmp_func_addr = &(tmp_axis->axis_neg_function);
                tmp_direction = &(tmp_axis->direction_neg);
            }

            JOYBALL_CHECK_COMBO_OFF( ctrl[j].combo ) 

            *tmp_func_addr = input_mouse_vel_axis;
            *tmp_direction = tmp_dir;

            tmp_axis->player = i;
            g_pc_input.mouse.speed = g_input.mouse_speed;

            JOYBALL_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] ) 

        } /* end if mouse */

    } /* end for stick direction */

    return 0;

} /* input_setup_trackball */

/******************************************************************************
**  Function   :  input_setup_keypad
**                            
**  Objective  :  This function sets up the emulated controller keypad
**                stick
**
**  Parameters :  player - which controller to setup for
**                ctrl   - ctrl struct array (either player keypad for 5200 or
**                           keyboard for 800)
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_keypad ( int player, t_atari_common *ctrl ) {

    int j, tmp_part_num, tmp_device_num, tmp_dir;
    int i = player;
    int stop;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    if ( g_input.machine_type == MACHINE_TYPE_5200 )
        stop = 16;
    else
        stop = 0x40;

    for ( j = 0; j < stop; j++ ) {

        tmp_part_num = ctrl[j].part_num[0];
        tmp_device_num = ctrl[j].device_num[0];
        tmp_dir = ctrl[j].direction[0];

        if ( ctrl[j].device[0] == DEV_KEYBOARD ) {

            tmp_part_num =  pc_adjust_key_value ( tmp_part_num );

            KEYBOARD_CHECK_COMBO_OFF( ctrl[j].combo )

            g_pc_input.keyboard[tmp_part_num].key_up_function = input_keypad_up;
            g_pc_input.keyboard[tmp_part_num].key_down_function = input_keypad_down;
            g_pc_input.keyboard[tmp_part_num].atari_key = j;
            g_pc_input.keyboard[tmp_part_num].player = i;

            if ( (g_input.machine_type !=  MACHINE_TYPE_5200) && 
                 (pc_check_for_lr_keys(ctrl[j].part_num[0])) ) {
                tmp_part_num = pc_check_for_lr_keys(ctrl[j].part_num[0]);
                g_pc_input.keyboard[tmp_part_num].key_up_function = input_keypad_up;
                g_pc_input.keyboard[tmp_part_num].key_down_function = input_keypad_down;
                g_pc_input.keyboard[tmp_part_num].atari_key = j;
                g_pc_input.keyboard[tmp_part_num].player = i;
            }

            KEYBOARD_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
        }

        if ( ctrl[j].device[0] == DEV_JOYSTICK || ctrl[j].device[0] == DEV_CONTROLLER) {
            if ( ctrl[j].part_type[0] == PART_TYPE_HAT ) {

                JOYHAT_CHECK_COMBO_OFF( ctrl[j].combo )

                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_keypad_up;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_keypad_down;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir] = j;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;

                JOYHAT_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }
            else if ( ctrl[j].part_type[0] == PART_TYPE_AXIS ) {

                if ( ctrl[j].direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_pos;
                }
                else {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_neg;
                }

                JOYAXIS_CHECK_COMBO_OFF( ctrl[j].combo )

                *tmp_func_addr = input_joyaxis_keypad;
                *tmp_direction = j;

                g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;

                JOYAXIS_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }
            else {

                JOYBUTTON_CHECK_COMBO_OFF( ctrl[j].combo )

                g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_keypad_up;
                g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_keypad_down;
                g_pc_input.button[tmp_device_num][tmp_part_num].atari_key = j;
                g_pc_input.button[tmp_device_num][tmp_part_num].player = i;

                JOYBUTTON_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }
        }

        if ( ctrl[j].device[0] == DEV_MOUSE ) {

            MOUSEBUTTON_CHECK_COMBO_OFF( ctrl[j].combo )

            g_pc_input.mouse_button[tmp_part_num].button_up_function = input_keypad_up;
            g_pc_input.mouse_button[tmp_part_num].button_down_function = input_keypad_down;
            g_pc_input.mouse_button[tmp_part_num].atari_key = j;
            g_pc_input.mouse_button[tmp_part_num].player = i;

            MOUSEBUTTON_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
        }
    }

    return 0;

} /* input_setup_keypad */

/******************************************************************************
**  Function   :  input_setup_800keypad
**                            
**  Objective  :  This function sets up the emulated controller 800 keypad
**
**  Parameters :  player - which controller to setup for
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_800keypad ( int player ) {

    int j, tmp_part_num, tmp_device_num, tmp_dir;
    int i = player;
    t_atari_common *ctrl = g_input.players[player].keypad;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    for ( j = 0; j < 16; j++ ) {

        tmp_part_num = ctrl[j].part_num[0];
        tmp_device_num = ctrl[j].device_num[0];
        tmp_dir = ctrl[j].direction[0];

        if ( ctrl[j].device[0] == DEV_KEYBOARD ) {
            tmp_part_num =  pc_adjust_key_value ( tmp_part_num );

            KEYBOARD_CHECK_COMBO_OFF( ctrl[j].combo )

            g_pc_input.keyboard[tmp_part_num].key_up_function = input_800keypad_up;
            g_pc_input.keyboard[tmp_part_num].key_down_function = input_800keypad_down;
            g_pc_input.keyboard[tmp_part_num].atari_key = j;
            g_pc_input.keyboard[tmp_part_num].player = i;

            KEYBOARD_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
        }

        if ( ctrl[j].device[0] == DEV_JOYSTICK || ctrl[j].device[0] == DEV_CONTROLLER) {
            if ( ctrl[j].part_type[0] == PART_TYPE_HAT ) {

                JOYHAT_CHECK_COMBO_OFF( ctrl[j].combo )

                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_800keypad_up;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_800keypad_down;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir] = j;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;

                JOYHAT_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }
            else if ( ctrl[j].part_type[0] == PART_TYPE_AXIS ) {

                if ( ctrl[j].direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_pos;
                }
                else {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_neg;
                }

                JOYAXIS_CHECK_COMBO_OFF( ctrl[j].combo )

                *tmp_func_addr = input_joyaxis_800keypad;
                *tmp_direction = j;

                g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;

                JOYAXIS_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }
            else {

                JOYBUTTON_CHECK_COMBO_OFF( ctrl[j].combo )

                g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_800keypad_up;
                g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_800keypad_down;
                g_pc_input.button[tmp_device_num][tmp_part_num].atari_key = j;
                g_pc_input.button[tmp_device_num][tmp_part_num].player = i;

                JOYBUTTON_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
            }
        }

        if ( ctrl[j].device[0] == DEV_MOUSE ) {

            MOUSEBUTTON_CHECK_COMBO_OFF( ctrl[j].combo )

            g_pc_input.mouse_button[tmp_part_num].button_up_function = input_800keypad_up;
            g_pc_input.mouse_button[tmp_part_num].button_down_function = input_800keypad_down;
            g_pc_input.mouse_button[tmp_part_num].atari_key = j;
            g_pc_input.mouse_button[tmp_part_num].player = i;

            MOUSEBUTTON_CHECK_COMBO_ON( ctrl[j].combo, &ctrl[j] )
        }
    }

    return 0;

} /* input_setup_800keypad */

/******************************************************************************
**  Function   :  input_setup_modkeys
**                            
**  Objective  :  This function sets up the emulated modifier keys on the 800
**                keyboard (shift, control, break - well, ok break is not a
**                modifier but it fits here!)
**
**  Parameters :  NONE 
**                
**  return     :  0
**      
******************************************************************************/
int input_setup_modkeys ( void ) {

    int j, tmp_part_num, tmp_device_num, tmp_dir;
    t_atari_common *ctrl;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    for ( j = 0; j < 4; j++ ) {

        switch ( j ) {
            case 0: ctrl = &g_input.shift_key; break;
            case 1: ctrl = &g_input.ctrl_key; break;
            case 2: ctrl = &g_input.break_key; break;
            case 3: ctrl = &g_input.reset_key; break;
            default: ctrl = &g_input.shift_key; break;
        }

        tmp_part_num = ctrl->part_num[0];
        tmp_device_num = ctrl->device_num[0];
        tmp_dir = ctrl->direction[0];

        if ( ctrl->device[0] == DEV_KEYBOARD ) {
            tmp_part_num =  pc_adjust_key_value ( tmp_part_num );

            KEYBOARD_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.keyboard[tmp_part_num].key_up_function = input_modkey_up;
            g_pc_input.keyboard[tmp_part_num].key_down_function = input_modkey_down;
            g_pc_input.keyboard[tmp_part_num].atari_key = j;
            g_pc_input.keyboard[tmp_part_num].player = 0;

            KEYBOARD_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }

        if ( ctrl->device[0] == DEV_JOYSTICK || ctrl->device[0] == DEV_CONTROLLER) {
            if ( ctrl->part_type[0] == PART_TYPE_HAT ) {

                JOYHAT_CHECK_COMBO_OFF( ctrl->combo )

                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_modkey_up;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_modkey_down;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir] = j;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = 0;

                JOYHAT_CHECK_COMBO_ON( ctrl->combo, ctrl )
            }
            else if ( ctrl->part_type[0] == PART_TYPE_AXIS ) {

                if ( ctrl->direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_pos;
                }
                else {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_neg;
                }

                JOYAXIS_CHECK_COMBO_OFF( ctrl->combo )

                *tmp_func_addr = input_joyaxis_modkey;
                *tmp_direction = j;

                g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = 0;

                JOYAXIS_CHECK_COMBO_ON( ctrl->combo, ctrl )
            }
            else {

                JOYBUTTON_CHECK_COMBO_OFF( ctrl->combo )

                g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_modkey_up;
                g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_modkey_down;
                g_pc_input.button[tmp_device_num][tmp_part_num].atari_key = j;
                g_pc_input.button[tmp_device_num][tmp_part_num].player = 0;

                JOYBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
            }
        }

        if ( ctrl->device[0] == DEV_MOUSE ) {

            MOUSEBUTTON_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.mouse_button[tmp_part_num].button_up_function = input_modkey_up;
            g_pc_input.mouse_button[tmp_part_num].button_down_function = input_modkey_down;
            g_pc_input.mouse_button[tmp_part_num].atari_key = j;
            g_pc_input.mouse_button[tmp_part_num].player = 0;

            MOUSEBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
    }

    return 0;

} /* input_setup_modkeys */

/******************************************************************************
**  Function   :  input_setup_switch
**                            
**  Objective  :  This function sets up the emulated switches on the 800
**                keyboard (start, select, option)
**
**  Parameters :  NONE 
**                
**  return     :  0
**      
******************************************************************************/
int input_setup_switch ( void ) {

    int j, tmp_part_num, tmp_device_num, tmp_dir;
    t_atari_common *ctrl;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    for ( j = 0; j < 3; j++ ) {

        switch ( j ) {
            case 0: ctrl = &g_input.start_key; break;
            case 1: ctrl = &g_input.select_key; break;
            case 2: ctrl = &g_input.option_key; break;
            default: ctrl = &g_input.start_key; break;
        }

        tmp_part_num = ctrl->part_num[0];
        tmp_device_num = ctrl->device_num[0];
        tmp_dir = ctrl->direction[0];

        if ( ctrl->device[0] == DEV_KEYBOARD ) {
            tmp_part_num =  pc_adjust_key_value ( tmp_part_num );

            KEYBOARD_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.keyboard[tmp_part_num].key_up_function = input_switch_up;
            g_pc_input.keyboard[tmp_part_num].key_down_function = input_switch_down;
            g_pc_input.keyboard[tmp_part_num].atari_key = 1<<j;
            g_pc_input.keyboard[tmp_part_num].player = 0;

            KEYBOARD_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }

        if ( ctrl->device[0] == DEV_JOYSTICK || ctrl->device[0] == DEV_CONTROLLER) {
            if ( ctrl->part_type[0] == PART_TYPE_HAT ) {

                JOYHAT_CHECK_COMBO_OFF( ctrl->combo )

                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_switch_up;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_switch_down;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].atari_key[tmp_dir] = 1<<j;
                g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = 0;

                JOYHAT_CHECK_COMBO_ON( ctrl->combo, ctrl )
            }
            else if ( ctrl->part_type[0] == PART_TYPE_AXIS ) {

                if ( ctrl->direction[0] == DIR_PLUS ) {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_pos;
                }
                else {
                    tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
                    tmp_direction = &g_pc_input.analog_axis[tmp_device_num][ctrl[j].part_num[0]].direction_neg;
                }

                JOYAXIS_CHECK_COMBO_OFF( ctrl->combo )

                *tmp_func_addr = input_joyaxis_switch;
                *tmp_direction = 1<<j;

                g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = 0;

                JOYAXIS_CHECK_COMBO_ON( ctrl->combo, ctrl )
            }
            else {

                JOYBUTTON_CHECK_COMBO_OFF( ctrl->combo )

                g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_switch_up;
                g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_switch_down;
                g_pc_input.button[tmp_device_num][tmp_part_num].atari_key = 1<<j;
                g_pc_input.button[tmp_device_num][tmp_part_num].player = 0;

                JOYBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
            }
        }

        if ( ctrl->device[0] == DEV_MOUSE ) {

            MOUSEBUTTON_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.mouse_button[tmp_part_num].button_up_function = input_switch_up;
            g_pc_input.mouse_button[tmp_part_num].button_down_function = input_switch_down;
            g_pc_input.mouse_button[tmp_part_num].atari_key = 1<<j;
            g_pc_input.mouse_button[tmp_part_num].player = 0;

            MOUSEBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
    }

    return 0;

} /* input_setup_switch */

/******************************************************************************
**  Function   :  input_setup_buttons
**                            
**  Objective  :  This function sets up the emulated controller buttons
**
**  Parameters :  player - which controller to setup for
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_buttons ( int player ) {

    int tmp_part_num, tmp_device_num, tmp_dir;
    int i = player;
    t_atari_common *ctrl;
    int ( **tmp_func_addr ) ( int,int,int );

    /***************************************
     * Top Button
     **************************************/
    ctrl = &g_input.players[i].top_button;
    tmp_device_num = ctrl->device_num[0];
    tmp_part_num = ctrl->part_num[0];
    tmp_dir = ctrl->direction[0];

    /*
     * Check for Keyboard
     */
    if (ctrl->device[0] == DEV_KEYBOARD ) {
        tmp_part_num =  pc_adjust_key_value ( tmp_part_num );

        KEYBOARD_CHECK_COMBO_OFF( ctrl->combo )

        g_pc_input.keyboard[tmp_part_num].key_up_function = input_top_button_up;
        g_pc_input.keyboard[tmp_part_num].key_down_function = input_top_button_down;
        g_pc_input.keyboard[tmp_part_num].player = i;

        KEYBOARD_CHECK_COMBO_ON( ctrl->combo, ctrl )
    }

    /*
     * Check for Joystick
     */
    if ( ctrl->device[0] == DEV_JOYSTICK || ctrl->device[0] == DEV_CONTROLLER) {

        if ( ctrl->part_type[0] == PART_TYPE_HAT ) {

            JOYHAT_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_top_button_up;
            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_top_button_down;
            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;

            JOYHAT_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
        else if ( ctrl->part_type[0] == PART_TYPE_AXIS ) {

            if ( tmp_dir == DIR_PLUS ) {
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
            }
            else {
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
            }

            JOYAXIS_CHECK_COMBO_OFF( ctrl->combo )

            *tmp_func_addr = input_joyaxis_top_button;

            g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;

            JOYAXIS_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
        else {

            JOYBUTTON_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_top_button_up;
            g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_top_button_down;
            g_pc_input.button[tmp_device_num][tmp_part_num].player = i;

            JOYBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
    }

    /*
     * Check for Mouse
     */
    if ( ctrl->device[0] == DEV_MOUSE ) {

        MOUSEBUTTON_CHECK_COMBO_OFF( ctrl->combo )

        g_pc_input.mouse_button[tmp_part_num].button_up_function = input_top_button_up;
        g_pc_input.mouse_button[tmp_part_num].button_down_function = input_top_button_down;
        g_pc_input.mouse_button[tmp_part_num].player = i;

        MOUSEBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
    }

    /***************************************
     * Bottom Button
     **************************************/
    ctrl = &g_input.players[i].bottom_button;
    tmp_device_num = ctrl->device_num[0];
    tmp_part_num = ctrl->part_num[0];
    tmp_dir = ctrl->direction[0];

    /*
     * Check for Keyboard
     */
    if ( ctrl->device[0] == DEV_KEYBOARD ) {
        tmp_part_num =  pc_adjust_key_value ( tmp_part_num );

        KEYBOARD_CHECK_COMBO_OFF( ctrl->combo )

        g_pc_input.keyboard[tmp_part_num].key_up_function = gtia_bottom_button_up;
        g_pc_input.keyboard[tmp_part_num].key_down_function = gtia_bottom_button_down;
        g_pc_input.keyboard[tmp_part_num].player = i;

        KEYBOARD_CHECK_COMBO_ON( ctrl->combo, ctrl )
    }
    if ( ctrl->device[0] == DEV_JOYSTICK || ctrl->device[0] == DEV_CONTROLLER) {
        if ( ctrl->part_type[0] == PART_TYPE_HAT ) {

            JOYHAT_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = gtia_bottom_button_up;
            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = gtia_bottom_button_down;
            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;

            JOYHAT_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
        else if ( ctrl->part_type[0] == PART_TYPE_AXIS ) {

            if ( tmp_dir == DIR_PLUS ) {
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
            }
            else {
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
            }

            JOYAXIS_CHECK_COMBO_OFF( ctrl->combo )

            *tmp_func_addr = input_joyaxis_bottom_button;

            g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;

            JOYAXIS_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
        else {
            JOYBUTTON_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = gtia_bottom_button_up;
            g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = gtia_bottom_button_down;
            g_pc_input.button[tmp_device_num][tmp_part_num].player = i;

            JOYBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
        }
    }

    /*
     * Check for Mouse
     */
    if ( ctrl->device[0] == DEV_MOUSE ) {

        MOUSEBUTTON_CHECK_COMBO_OFF( ctrl->combo )

        g_pc_input.mouse_button[tmp_part_num].button_up_function = gtia_bottom_button_up;
        g_pc_input.mouse_button[tmp_part_num].button_down_function = gtia_bottom_button_down;
        g_pc_input.mouse_button[tmp_part_num].player = i;

        MOUSEBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
    }

    return 0;

} /* input_setup_buttons */

/******************************************************************************
**  Function   :  input_setup_paddle_button
**                            
**  Objective  :  This function sets up controller options for the emulated
**                paddle button (for the 800)
**
**  Parameters :  player - which controller to setup for
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_paddle_button ( int player ) {

    int j, tmp_dir, tmp_device_num, tmp_part_num;
    int i;
    t_atari_common *ctrl = &g_input.players[player].bottom_button;
    int ( **tmp_func_addr ) ( int,int,int );
    int *tmp_direction;

    /*
     * Similiar to reading the stick via PIA.
     *   P0 is the same as P0 left
     *   P1 is the same as P0 right
     *   P2 is the same as P1 left
     *   P3 is the same as P1 right
     */
    switch ( player ) {
        case 0: i = 0; j = DIR_LEFT; break;
        case 1: i = 0; j = DIR_RIGHT; break;
        case 2: i = 1; j = DIR_LEFT; break;
        case 3: i = 1; j = DIR_RIGHT; break;
        default: i = 0; j = DIR_LEFT; break;
    }

    /*
     * Check for Keyboard
     */
    if ( ctrl->device[0] == DEV_KEYBOARD ) {
        tmp_part_num =  pc_adjust_key_value ( ctrl->part_num[0] );

        KEYBOARD_CHECK_COMBO_OFF( ctrl->combo )

        g_pc_input.keyboard[tmp_part_num].direction = j;
        g_pc_input.keyboard[tmp_part_num].key_up_function = input_key_up_digital_stick;
        g_pc_input.keyboard[tmp_part_num].key_down_function = input_key_down_digital_stick;
        g_pc_input.keyboard[tmp_part_num].player = i;
        g_pc_input.keyboard[tmp_part_num].atari_key = 0;

        KEYBOARD_CHECK_COMBO_ON( ctrl->combo, ctrl )

    } /* end if keyboard */

    /*
     * Check for Joystick
     */
    tmp_device_num = ctrl->device_num[0]; 
    tmp_part_num = ctrl->part_num[0]; 
    tmp_dir = ctrl->direction[0]; 
    if ( ctrl->device[0] == DEV_JOYSTICK || ctrl->device[0] == DEV_CONTROLLER) {

        if ( ctrl->part_type[0] == PART_TYPE_AXIS ) {

            if ( tmp_dir == DIR_PLUS ) {
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_pos_function;
                tmp_direction = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_pos;
            }
            else {
                tmp_func_addr = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].axis_neg_function;
                tmp_direction = &g_pc_input.analog_axis[tmp_device_num][tmp_part_num].direction_neg;
            }

            JOYAXIS_CHECK_COMBO_OFF( ctrl->combo )

            *tmp_func_addr = input_axis_digital_stick;
            *tmp_direction = j;

            g_pc_input.analog_axis[tmp_device_num][tmp_part_num].player = i;

            JOYAXIS_CHECK_COMBO_ON( ctrl->combo, ctrl )

        } /* end if axis on joystick */

        if ( ctrl->part_type[0] == PART_TYPE_HAT ) {

            JOYHAT_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_not_pressed_function[tmp_dir] = input_key_up_digital_stick;
            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].hat_pressed_function[tmp_dir] = input_key_down_digital_stick;
            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].player[tmp_dir] = i;
            g_pc_input.pov_hat[tmp_device_num][tmp_part_num].direction[tmp_dir] = j;

            JOYHAT_CHECK_COMBO_ON( ctrl->combo, ctrl )

        } /* end if hat on joystick */

        if ( ctrl->part_type[0] == PART_TYPE_BUTTON ) {

            JOYBUTTON_CHECK_COMBO_OFF( ctrl->combo )

            g_pc_input.button[tmp_device_num][tmp_part_num].button_up_function = input_key_up_digital_stick;
            g_pc_input.button[tmp_device_num][tmp_part_num].button_down_function = input_key_down_digital_stick;
            g_pc_input.button[tmp_device_num][tmp_part_num].player = i;
            g_pc_input.button[tmp_device_num][tmp_part_num].direction = j;

            JOYBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )

        } /* end if button on joystick */

    } /* end if joystick */

    /*
     * Check for Mouse
     */
    if ( ctrl->device[0] == DEV_MOUSE ) {

        MOUSEBUTTON_CHECK_COMBO_OFF( ctrl->combo )

        g_pc_input.mouse_button[tmp_part_num].button_up_function = input_key_up_digital_stick;
        g_pc_input.mouse_button[tmp_part_num].button_down_function = input_key_down_digital_stick;
        g_pc_input.mouse_button[tmp_part_num].player = i;

        MOUSEBUTTON_CHECK_COMBO_ON( ctrl->combo, ctrl )
    }

    return 0;

} /* input_setup_paddle_button */

/******************************************************************************
**  Function   :  input_setup_combo_actions
**                            
**  Objective  :  This function assigns the combo check function to the
**                assigned PC input
**
**  Parameters :  NONE
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_combo_actions ( void ) {

    int t_dir, t_dn, t_pn;
    t_pc_combo_function *tmp = g_pc_input.combo_list;

    while ( tmp ) {

        t_dir = tmp->ctrl->direction[1];
        t_dn = tmp->ctrl->device_num[1];
        t_pn = tmp->ctrl->part_num[1];
        *tmp->function_addr = tmp->combo_off_function;

        if ( tmp->ctrl->device[1] == DEV_KEYBOARD ) {

            t_pn =  pc_adjust_key_value ( tmp->ctrl->part_num[1] );
            g_pc_input.keyboard[t_pn].key_up_function = input_keyboard_combo_disengage;
            g_pc_input.keyboard[t_pn].key_down_function = input_keyboard_combo_engage;
            g_pc_input.keyboard[t_pn].player = t_dn;
            g_pc_input.keyboard[t_pn].direction = t_dir;
            g_pc_input.keyboard[t_pn].atari_key = tmp->ctrl->part_num[1];
        }

        else if ( tmp->ctrl->device[1] == DEV_JOYSTICK || tmp->ctrl->device[1] == DEV_CONTROLLER) {

            if ( tmp->ctrl->part_type[1] == PART_TYPE_AXIS ) {

                if ( t_dir == DIR_PLUS ) {
                    g_pc_input.analog_axis[t_dn][t_pn].axis_pos_function = input_joyaxis_combo;
                    g_pc_input.analog_axis[t_dn][t_pn].direction_pos = t_pn;
                }
                else {
                    g_pc_input.analog_axis[t_dn][t_pn].axis_neg_function = input_joyaxis_combo;
                    g_pc_input.analog_axis[t_dn][t_pn].direction_neg = t_pn;
                }

                g_pc_input.analog_axis[t_dn][t_pn].player = t_dn;
            }
            else if ( tmp->ctrl->part_type[1] == PART_TYPE_HAT ) {

                g_pc_input.pov_hat[t_dn][t_pn].hat_not_pressed_function[t_dir] = input_joyhat_combo_disengage;
                g_pc_input.pov_hat[t_dn][t_pn].hat_pressed_function[t_dir] = input_joyhat_combo_engage;
                g_pc_input.pov_hat[t_dn][t_pn].player[t_dir] = t_dn;
                g_pc_input.pov_hat[t_dn][t_pn].direction[t_dir] = t_dir;
                g_pc_input.pov_hat[t_dn][t_pn].atari_key[t_dir] = t_pn;
            }
            else if ( tmp->ctrl->part_type[1] == PART_TYPE_BUTTON ) {

                g_pc_input.button[t_dn][t_pn].button_up_function = input_keyboard_combo_disengage;
                g_pc_input.button[t_dn][t_pn].button_down_function = input_keyboard_combo_engage;
                g_pc_input.button[t_dn][t_pn].player = t_dn;
                g_pc_input.button[t_dn][t_pn].direction = t_dir;
                g_pc_input.button[t_dn][t_pn].atari_key = t_pn;
            }
        }

        else if ( tmp->ctrl->device[1] == DEV_MOUSE ) {

            g_pc_input.mouse_button[t_pn].button_up_function = input_keyboard_combo_disengage;
            g_pc_input.mouse_button[t_pn].button_down_function = input_keyboard_combo_engage;
            g_pc_input.mouse_button[t_pn].player = t_dn;
            g_pc_input.mouse_button[t_pn].direction = t_dir;
            g_pc_input.mouse_button[t_pn].player = t_pn;
        }

        tmp = tmp->next;
    }
    return 0;

} /* input_setup_combo_actions */

/******************************************************************************
**  Function   :  input_setup_frame_functions
**                            
**  Objective  :  This function sets up control functions to be called every 
**                frame
**
**  Parameters :  player - which controller to setup for
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_frame_functions ( int player ) {

    t_atari_common *ctrl = g_input.players[player].stick.direction;

    g_pc_input.frame_x_function[player] = input_frame_do_nothing;
    g_pc_input.frame_y_function[player] = input_frame_do_nothing;

    if ( g_input.machine_type == MACHINE_TYPE_5200 ) {
        if ( ctrl[DIR_LEFT].device[0] == DEV_MOUSE )
            g_pc_input.frame_x_function[player] = input_center_x_axis_frame;
        if ( ctrl[DIR_UP].device[0] == DEV_MOUSE )
            g_pc_input.frame_y_function[player] = input_center_y_axis_frame;

        if ( ctrl[DIR_LEFT].device[0]== DEV_JOYSTICK &&
             ctrl[DIR_LEFT].part_type[0]== PART_TYPE_BALL )
            g_pc_input.frame_x_function[player] = input_center_x_axis_frame;
        if ( ctrl[DIR_UP].device[0]== DEV_JOYSTICK &&
             ctrl[DIR_UP].part_type[0]== PART_TYPE_BALL )
            g_pc_input.frame_y_function[player] = input_center_y_axis_frame;
    }

    if ( g_input.players[player].stick.simulate_analog ) {
        g_pc_input.frame_x_function[player] = input_simulate_analog_x_axis_frame;
        g_pc_input.frame_y_function[player] = input_simulate_analog_y_axis_frame;
    }

    return 0;

} /* end input_setup_frame_functions */

/******************************************************************************
**  Function   :  input_setup_stick_values
**                            
**  Objective  :  This function sets up pot values and axis idle values for
**                the emulated stick
**
**  Parameters :  player - which controller to setup for
**                                                
**  return     :  0
**      
******************************************************************************/
int input_setup_stick_values ( int player ) {

    int j;
    double deadzone, deadzone_ratio;
    int pot_high_value, pot_low_value, pot_center_value;

    if ( g_input.machine_type == MACHINE_TYPE_5200 ) {
        pot_high_value = POT_HIGH_VALUE;
        pot_low_value = POT_LOW_VALUE;
        pot_center_value = POT_CENTER_VALUE;
    }
    else {
        pot_high_value = POT_800_HIGH_VALUE;
        pot_low_value = POT_800_LOW_VALUE;
        pot_center_value = POT_800_CENTER_VALUE;
    }

    /*
     * Use sensitivity to find upper, lower and center values
     */
    g_stick_interface[player].pot_low = pot_low_value + (int)( (double)(pot_high_value-pot_low_value) * (double)(100-g_input.players[player].stick.sensitivity) / 200.0 );
    g_stick_interface[player].pot_high = pot_high_value - (int)( (double)(pot_high_value-pot_low_value) * (double)(100-g_input.players[player].stick.sensitivity) / 200.0 );
    g_stick_interface[player].axis_value[0] = pot_center_value;
    g_stick_interface[player].axis_value[1] = pot_center_value;
    g_stick_interface[player].sa_high_threshold = (g_stick_interface[player].pot_high+pot_center_value)>>1;
    g_stick_interface[player].sa_low_threshold = (g_stick_interface[player].pot_low+pot_center_value)>>1;

    /*
     * Analog axis values
     */
    deadzone = 32768*g_input.players[player].stick.deadzone/100;
    deadzone_ratio = ((double)(32768-deadzone))/((double)(g_stick_interface[player].pot_high-g_stick_interface[player].pot_low)/2.0);

    for ( j = -32768; j < 32768; ++j ) {
        if ( abs(j) < deadzone ) {
            g_stick_interface[player].pot_values[j+32768] = pot_center_value;
        }
        else {
            if ( j > 0 )
                g_stick_interface[player].pot_values[j+32768] = (unsigned char)(((j-deadzone)/deadzone_ratio)+(pot_center_value+1));
            else
                g_stick_interface[player].pot_values[j+32768] = (unsigned char)(((j+deadzone)/deadzone_ratio)+(pot_center_value-1));
        }
    }

    /*
     * Setup both axis idle variables and the pressed variables
     */
    g_stick_interface[player].axis_idle[0] = -1;
    g_stick_interface[player].axis_idle[1] = -1;
    for ( j = DIR_LEFT; j < MAX_DIR; ++j )
        g_stick_interface[player].pressed[j] = 0;

    /*
     * Check for mouse on this player
     */
    if ( (g_input.players[player].stick.direction[0].device[0] == DEV_MOUSE && 
          g_input.machine_type == MACHINE_TYPE_5200) ) {
        g_stick_interface[player].axis_idle[0] = 4;
    }
    return 0;

} /* end input_setup_stick_values */

int input_do_nothing (int key, int dir, int temp) {

    return 0;

} /* input_do_nothing */

int input_frame_do_nothing (int key, int dir) {

    return 0;

} /* input_frame_do_nothing */

int input_center_x_axis_frame (int player, int dir) {

    g_stick_interface[player].axis_idle[0]--;
    if ( !g_stick_interface[player].axis_idle[0] )
        input_stick_h_axis ( player, POT_CENTER_VALUE );

    return 0;

} /* end input_center_x_axis_frame */

int input_center_y_axis_frame (int player, int dir) {

    g_stick_interface[player].axis_idle[1]--;
    if ( !g_stick_interface[player].axis_idle[1] )
        input_stick_v_axis ( player, POT_CENTER_VALUE );

    return 0;

} /* end input_center_y_axis_frame */

int input_simulate_analog_x_axis_frame (int player, int tmp) {

    if ( g_stick_interface[player].pressed[DIR_LEFT] )
        input_stick_h_axis ( player, g_stick_interface[player].axis_value[0]-=3 );
    if ( g_stick_interface[player].pressed[DIR_RIGHT] )
        input_stick_h_axis ( player, g_stick_interface[player].axis_value[0]+=3 );

    if ( g_stick_interface[player].axis_value[0] < 
            g_stick_interface[player].pot_low ) 
        g_stick_interface[player].axis_value[0] = 
            g_stick_interface[player].pot_low;

    if ( g_stick_interface[player].axis_value[0] > 
            g_stick_interface[player].pot_high ) 
        g_stick_interface[player].axis_value[0] = 
            g_stick_interface[player].pot_high;

    return 0;

} /* end input_simulate_analog_x_axis_frame */

int input_simulate_analog_y_axis_frame (int player, int tmp) {

    if ( g_stick_interface[player].pressed[DIR_UP] )
        input_stick_v_axis ( player, g_stick_interface[player].axis_value[1]-=3 );
    if ( g_stick_interface[player].pressed[DIR_DOWN] )
        input_stick_v_axis ( player, g_stick_interface[player].axis_value[1]+=3 );

    if ( g_stick_interface[player].axis_value[1] < 
            g_stick_interface[player].pot_low ) 
        g_stick_interface[player].axis_value[1] = 
            g_stick_interface[player].pot_low;

    if ( g_stick_interface[player].axis_value[1] > 
            g_stick_interface[player].pot_high ) 
        g_stick_interface[player].axis_value[1] = 
            g_stick_interface[player].pot_high;

    return 0;

} /* end input_simulate_analog_y_axis_frame */

int input_key_down_stick (int player, int dir, int key) {

    g_stick_interface[player].pressed[dir]= 1;
    switch ( dir ) {
        case DIR_LEFT: 
            pokey_set_pot_value ( (player<<1), g_stick_interface[player].pot_low );
            break;
        case DIR_RIGHT: 
            pokey_set_pot_value ( (player<<1), g_stick_interface[player].pot_high );
            break;
        case DIR_UP: 
            pokey_set_pot_value ( ((player<<1) | 0x1), g_stick_interface[player].pot_low );
            break;
        case DIR_DOWN: 
            pokey_set_pot_value ( ((player<<1) | 0x1), g_stick_interface[player].pot_high );
            break;
    }

    return 0;

} /* end input_key_down_stick */

int input_key_down_stick_sa (int player, int dir, int key) {

    g_stick_interface[player].pressed[dir]= 1;

    switch ( dir ) {
        case DIR_LEFT: 
            input_stick_h_axis ( player, g_stick_interface[player].axis_value[0]-=3 );
            if ( g_stick_interface[player].axis_value[0] < 
                    g_stick_interface[player].pot_low ) 
                g_stick_interface[player].axis_value[0] = 
                    g_stick_interface[player].pot_low;
            break;
        case DIR_RIGHT: 
            input_stick_h_axis ( player, g_stick_interface[player].axis_value[0]+=3 );
            if ( g_stick_interface[player].axis_value[0] > 
                    g_stick_interface[player].pot_high ) 
                g_stick_interface[player].axis_value[0] = 
                    g_stick_interface[player].pot_high;
            break;
        case DIR_UP: 
            input_stick_v_axis ( player, g_stick_interface[player].axis_value[1]-=3 );
            if ( g_stick_interface[player].axis_value[1] < 
                    g_stick_interface[player].pot_low ) 
                g_stick_interface[player].axis_value[1] = 
                    g_stick_interface[player].pot_low;
            break;
        case DIR_DOWN: 
            input_stick_v_axis ( player, g_stick_interface[player].axis_value[1]+=3 );
            if ( g_stick_interface[player].axis_value[1] > 
                    g_stick_interface[player].pot_high ) 
                g_stick_interface[player].axis_value[1] = 
                    g_stick_interface[player].pot_high;
            break;
    }

    return 0;

} /* end input_key_down_stick_sa */

int input_key_up_stick (int player, int dir, int key) {

    int mask = (dir & 0x2) >> 1;
    g_stick_interface[player].pressed[dir]= 0;
    if ( g_stick_interface[player].pressed[dir ^ 0x1]) {
        if ( dir & 0x1 )
            pokey_set_pot_value ( (player<<1) | mask, g_stick_interface[player].pot_low );
        else
            pokey_set_pot_value ( (player<<1) | mask, g_stick_interface[player].pot_high );
    }
    else
        pokey_set_pot_value ( (player<<1) | mask, POT_CENTER_VALUE );

    return 0;

} /* end input_key_up_stick */

int input_key_up_stick_sa (int player, int dir, int key) {

    g_stick_interface[player].pressed[dir]= 0;

    return 0;

} /* end input_key_up_stick_sa */

int input_key_down_digital_stick (int player, int dir, int key) {

    int value[] = {0x4,0x8,0x1,0x2};

    g_stick_interface[player].digital_pressed[dir]= 1;
    pia_set_input ( player, value[dir], 0 );
    pia_set_input ( player, value[dir^1], value[dir^1] );

    return 0;

} /* end input_key_down_digital_stick */

int input_key_up_digital_stick (int player, int dir, int key) {

    int value[] = {0x4,0x8,0x1,0x2};

    g_stick_interface[player].digital_pressed[dir]= 0;
    pia_set_input ( player, value[dir], value[dir] );

    if ( g_stick_interface[player].digital_pressed[dir^1] )
        pia_set_input ( player, value[dir^1], 0 );

    return 0;

} /* end input_key_up_digital_stick */

int input_joystick_pots_sa ( int player, int dir, int reading ) {

    int value;
    int rate;

    if ( dir & 0x1 ) {
        value = (g_stick_interface[player].pot_values[32768+reading] > g_stick_interface[player].sa_high_threshold);
        rate = 3;
    }
    else {
        value = (g_stick_interface[player].pot_values[32768-reading] < g_stick_interface[player].sa_low_threshold);
        rate = -3;
    }

    if ( value ) {
        g_stick_interface[player].pressed[dir] = 1;
        g_stick_interface[player].pressed[dir ^ 1] = 0;
        if ( dir & 0x2 )
            input_stick_v_axis ( player, g_stick_interface[player].axis_value[0] += rate );
        else
            input_stick_h_axis ( player, g_stick_interface[player].axis_value[0] += rate );
    }
    else
        g_stick_interface[player].pressed[dir] = 0;

    return 0;

} /* input_joystick_pots_sa */

int input_joystick_pots ( int player, int dir, int reading ) {

    int value;

    //if ( dir >> 1 )printf ( "y=%d\n", reading );
    //else printf ( "x=%d\n", reading );
    if ( dir & 0x1 )
        value = 32768+reading;
    else
        value = 32768-reading;

    pokey_set_pot_value ( (player<<1 | (dir>>1)), g_stick_interface[player].pot_values[value] );

    return 0;

} /* end input_joystick_pots */

int input_axis_digital_stick ( int player, int dir, int reading ) {

    int value[] = {0x4,0x8,0x1,0x2};

    if ( reading > 16384 )
        pia_set_input ( player, value[dir], 0 );
    else
        pia_set_input ( player, value[dir], value[dir] );

    return 0;

} /* end input_axis_digital_stick */

int input_keypad_down ( int player, int dir, int kbcode ) {

    pokey_set_keypad_down ( player, kbcode, 1 );

    return 0;

} /* end input_keypad_down */

int input_keypad_up ( int player, int dir, int kbcode ) {

    pokey_set_keypad_down ( player, kbcode, 0 );

    return 0;

} /* end input_keypad_up */

int input_modkey_down ( int player, int dir, int kbcode ) {

    switch ( kbcode ) {
        case 0: pokey_set_shift_key_down ( 1 ); break;
        case 1: pokey_set_ctrl_key_down ( 1 ); break;
        case 2: pokey_set_break_key_down ( 1 ); break;
        case 3: antic_set_reset(); break;
        default: pokey_set_shift_key_down ( 1 ); break;
    }

    return 0;

} /* end input_modkey_down */

int input_modkey_up ( int player, int dir, int kbcode ) {

    switch ( kbcode ) {
        case 0: pokey_set_shift_key_down ( 0 ); break;
        case 1: pokey_set_ctrl_key_down ( 0 ); break;
        case 2: pokey_set_break_key_down ( 0 ); break;
        case 3: break;
        default: pokey_set_shift_key_down ( 0 ); break;
    }

    return 0;

} /* end input_modkey_up */

//int input_joyaxis_modkey ( int player, int reading, int kbcode ) {
int input_joyaxis_modkey ( int player, int kbcode, int reading ) {

    int value;

    if ( reading > 16384 )
        value = 1;
    else
        value = 0;

    switch ( kbcode ) {
        case 0: pokey_set_shift_key_down ( value ); break;
        case 1: pokey_set_ctrl_key_down ( value ); break;
        case 2: pokey_set_break_key_down ( value ); break;
        default: pokey_set_shift_key_down ( value ); break;
    }

    return 0;

} /* end input_joyaxis_modkey */

int input_switch_down ( int player, int dir, int mask ) {

    gtia_set_switch ( mask, 0 );

    return 0;

} /* end input_switch_down */

int input_switch_up ( int player, int dir, int mask ) {

    gtia_set_switch ( mask, 1 );

    return 0;

} /* end input_switch_up */

//int input_joyaxis_switch ( int player, int reading, int mask ) {
int input_joyaxis_switch ( int player, int mask, int reading ) {

    if ( reading > 16384 )
        gtia_set_switch ( mask, 0 );
    else
        gtia_set_switch ( mask, 1 );

    return 0;

} /* end input_joyaxis_switch */

//int input_joyaxis_keypad ( int player, int reading, int kbcode ) {
int input_joyaxis_keypad ( int player, int kbcode, int reading ) {

    if ( reading > 16384 )
        pokey_set_keypad_down( player, kbcode, 1 );
    else
        pokey_set_keypad_down( player, kbcode, 0 );

    return 0;

} /* end input_joyaxis_keypad */

int input_800keypad_down ( int player, int dir, int kbcode ) {

    pia_set_keypad_down ( player, kbcode, 1 );

    return 0;

} /* end input_800keypad_down */

int input_800keypad_up ( int player, int dir, int kbcode ) {

    pia_set_keypad_down ( player, kbcode, 0 );

    return 0;

} /* end input_800keypad_up */

int input_joyaxis_800keypad ( int player, int reading, int kbcode ) {

    if ( reading > 16384 )
        pia_set_keypad_down( player, kbcode, 1 );
    else
        pia_set_keypad_down( player, kbcode, 0 );

    return 0;

} /* end input_joyaxis_800keypad */

int input_top_button_down ( int player, int dir, int temp ) {

    pokey_set_top_button_down ( player, 1 );

    return 0;

} /* end pc_top_button_down */

int input_top_button_up ( int player, int dir, int temp ) {

    pokey_set_top_button_down ( player, 0 );

    return 0;

} /* end pc_top_button_up */

int input_joyaxis_top_button ( int player, int dir, int reading ) {

    if ( reading > 16384 )
        pokey_set_top_button_down ( player, 1 );
    else
        pokey_set_top_button_down ( player, 0 );

    return 0;

} /* end input_joyaxis_top_button */

int input_joyaxis_bottom_button ( int player, int dir, int reading ) {

    if ( reading > 16384 )
        gtia_bottom_button_down ( player, 0, 0 );
    else
        gtia_bottom_button_up ( player, 0, 0 );

    return 0;

} /* end input_joyaxis_bottom_button */

int input_stick_h_axis ( int player, int reading ) {

    if ( reading < g_stick_interface[player].pot_low ) 
        reading = g_stick_interface[player].pot_low;
    if ( reading > g_stick_interface[player].pot_high ) 
        reading = g_stick_interface[player].pot_high;
    pokey_set_pot_value ( player<<1, reading );

    return 0;

} /* end input_stick_h_axis */

int input_stick_v_axis ( int player, int reading ) {

    if ( reading < g_stick_interface[player].pot_low ) 
        reading = g_stick_interface[player].pot_low;
    if ( reading > g_stick_interface[player].pot_high ) 
        reading = g_stick_interface[player].pot_high;
    pokey_set_pot_value ( ((player<<1) | 0x1), reading );

    return 0;

} /* end input_stick_v_axis */

int input_mouse_axis (int player, int dir, int value ) {

    switch ( dir ) {
        case DIR_LEFT: 
            input_stick_h_axis ( player, g_stick_interface[player].axis_value[0]-=value );
            break;
        case DIR_RIGHT: 
            input_stick_h_axis ( player, g_stick_interface[player].axis_value[0]+=value );
            break;
        case DIR_UP: 
            input_stick_v_axis ( player, g_stick_interface[player].axis_value[1]-=value );
            break;
        case DIR_DOWN: 
            input_stick_v_axis ( player, g_stick_interface[player].axis_value[1]+=value );
            break;
    }

    return 0;

} /* end input_mouse_axis */

int input_mouse_vel_axis (int player, int dir, int value ) {

    switch ( dir ) {
        case DIR_LEFT: 
            g_stick_interface[player].axis_idle[0] = 4;
            input_stick_h_axis ( player, -value + POT_CENTER_VALUE );
            break;
        case DIR_RIGHT: 
            g_stick_interface[player].axis_idle[0] = 4;
            input_stick_h_axis ( player, value + POT_CENTER_VALUE );
            break;
        case DIR_UP: 
            g_stick_interface[player].axis_idle[1] = 4;
            input_stick_v_axis ( player, -value + POT_CENTER_VALUE );
            break;
        case DIR_DOWN: 
            g_stick_interface[player].axis_idle[1] = 4;
            input_stick_v_axis ( player, value + POT_CENTER_VALUE );
            break;
    }

    return 0;

} /* end input_mouse_vel_axis */

int input_trackball_x_axis (int player, int dir, int value) {

    g_stick_interface[player].axis_idle[0] = 4;
    input_stick_h_axis ( player, value );

    return 0;

} /* end input_trackball_x_axis */

int input_trackball_y_axis (int player, int dir, int value) {

    g_stick_interface[player].axis_idle[1] = 4;
    input_stick_v_axis ( player, value );

    return 0;

} /* end input_trackball_y_axis */

#define COMBO_TURN_ON(search) \
    while ( tmp ) { \
        if ( search ) { \
            *tmp->function_addr = tmp->combo_on_function;\
            for ( i = 0; i < 3; ++i )\
                *(tmp->param_addr[i]) = tmp->param_on[i];\
        }\
        tmp = tmp->next;\
    }

#define COMBO_TURN_OFF(search) \
    while ( tmp ) { \
        if ( search ) { \
            *tmp->function_addr = tmp->combo_off_function;\
            for ( i = 0; i < 3; ++i )\
                *(tmp->param_addr[i]) = tmp->param_off[i];\
        }\
        tmp = tmp->next;\
    }

int input_keyboard_combo_engage (int device_num, int part_type, int part_num) {

    int i;
    t_pc_combo_function *tmp = g_pc_input.combo_list;

    COMBO_TURN_ON((tmp->ctrl->device_num[1] == device_num && 
                   tmp->ctrl->part_num[1] == part_num))

    return 0;
}

int input_keyboard_combo_disengage (int device_num, int part_type, int part_num) {

    int i;
    t_pc_combo_function *tmp = g_pc_input.combo_list;

    COMBO_TURN_OFF((tmp->ctrl->device_num[1] == device_num && 
                   tmp->ctrl->part_num[1] == part_num))
    return 0;
}

int input_joyaxis_combo (int device_num, int part_num, int value) {

    int i;
    t_pc_combo_function *tmp = g_pc_input.combo_list;

    if ( value < 16000 ) {
        COMBO_TURN_OFF((tmp->ctrl->device_num[1] == device_num && 
                   tmp->ctrl->part_num[1] == part_num))
    }
    else {
        COMBO_TURN_ON((tmp->ctrl->device_num[1] == device_num && 
                   tmp->ctrl->part_num[1] == part_num))
    }

    return 0;
}

int input_joyhat_combo_engage (int device_num, int direction, int part_num) {

    int i;
    t_pc_combo_function *tmp = g_pc_input.combo_list;

    COMBO_TURN_ON((tmp->ctrl->device_num[1] == device_num && 
                   tmp->ctrl->direction[1] == direction &&
                   tmp->ctrl->part_num[1] == part_num))
    return 0;
}

int input_joyhat_combo_disengage (int device_num, int direction, int part_num) {

    int i;
    t_pc_combo_function *tmp = g_pc_input.combo_list;

    COMBO_TURN_OFF((tmp->ctrl->device_num[1] == device_num && 
                   tmp->ctrl->direction[1] == direction &&
                   tmp->ctrl->part_num[1] == part_num))
    return 0;
}
