/******************************************************************************
*
* Copyright (C) 2017  Brian Berlin
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
*
******************************************************************************
*
* FILENAME: shader.c
*
* DESCRIPTION:  This contains CRT shader handling
*
* AUTHORS:  bberlin
*
* CHANGE LOG
*
*  VER     DATE     CHANGED BY   DESCRIPTION OF CHANGE
* ------  --------  ----------   ------------------------------------
* 0.8.0   03/26/17  bberlin      Creation
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "shader.h"
#include "kconfig.h"
#include "logger.h"
#include "db_if.h"
#include "util.h"
#include "logger.h"
#ifdef __APPLE__
#include "SDL2_image/SDL_image.h"
#else
#include <SDL_image.h>
#endif
#include "../kat5200.h"

#define LOG_SRC ""

int shader_save_uniform ( char *master, char *shader, t_param *p_param );
int shader_open_shader ( char *file, char *master, t_shader *shader );
int shader_get_location ( GLhandleARB prog, char *name );
int shader_compile_shader_program ( t_shader *p_shader );
int shader_create_lut_textures ( t_master_shader *p_master);
int shader_set_texture_uniforms ( t_texture_locs *p_locs, t_texture *p_tex );
int shader_gen_tex_coord_buffer ( t_texture *p_tex );
int shader_set_shader_sizes ( t_shader *p_shader, t_texture *input_texture,
                              SDL_Rect viewport );

int shader_gen_texture ( t_master_shader *p_master, t_texture *p_tex, 
              void *frame, int wrap_mode, int filter, int float_frame );
int shader_gen_fbo ( t_master_shader *p_master, GLuint *fbo, t_texture *fbo_tex,
                                      int wrap_mode, int filter_linear, 
                                      int float_framebuffer );
int shader_set_values ( t_master_shader *p_master, int pass );

void shader_init_shader ( t_shader *p_shader );
void shader_init_texture_locs ( t_texture_locs *p_locs );
void shader_free_context ( );
void shader_free_master ( t_master_shader *p_master );
void shader_free_program ( t_shader *p_shader );
void shader_free_objects ( t_shader *p_shader );
int shader_check_error ( char *prefix );

static t_master_shader g_master;

static PFNGLGETSHADERINFOLOGPROC glGetShaderInfoLogARB;
static PFNGLDETACHSHADERPROC glDetachShaderARB;
static PFNGLDELETESHADERPROC glDeleteShaderARB;
static PFNGLDELETEPROGRAMPROC glDeleteProgramARB;
static PFNGLACTIVETEXTUREPROC glActiveTexture2;
static PFNGLATTACHSHADERPROC glAttachShaderARB;
static PFNGLGETSHADERIVPROC glGetShaderivARB;
static PFNGLUSEPROGRAMPROC glUseProgramARB;
static PFNGLCREATEPROGRAMPROC glCreateProgramARB;
static PFNGLCREATESHADERPROC glCreateShaderARB;

static PFNGLCOMPILESHADERPROC glCompileShaderARB;
static PFNGLGETUNIFORMLOCATIONPROC glGetUniformLocationARB;
static PFNGLGETATTRIBLOCATIONPROC glGetAttribLocationARB;
static PFNGLENABLEVERTEXATTRIBARRAYPROC glEnableVertexAttribArrayARB;
static PFNGLVERTEXATTRIBPOINTERPROC glVertexAttribPointerARB;
static PFNGLLINKPROGRAMPROC glLinkProgramARB;
static PFNGLSHADERSOURCEPROC glShaderSourceARB;
static PFNGLUNIFORM1IPROC glUniform1iARB;
static PFNGLUNIFORM1FPROC glUniform1fARB;
static PFNGLUNIFORM2FVPROC glUniform2fvARB;
static PFNGLUNIFORMMATRIX4FVPROC glUniformMatrix4fvARB;
static PFNGLGENVERTEXARRAYSPROC glGenVertexArraysARB;
static PFNGLBINDVERTEXARRAYPROC glBindVertexArrayARB;
static PFNGLDELETEVERTEXARRAYSPROC glDeleteVertexArraysARB;
static PFNGLGENBUFFERSPROC glGenBuffersARB;
static PFNGLBINDBUFFERPROC glBindBufferARB;
static PFNGLBUFFERDATAPROC glBufferDataARB;
static PFNGLBUFFERSUBDATAPROC glBufferSubDataARB;
static PFNGLDELETEBUFFERSPROC glDeleteBuffersARB;
static PFNGLENABLEVERTEXATTRIBARRAYPROC glEnableVertexAttribArrayARB;
static PFNGLVERTEXATTRIBPOINTERPROC glVertexAttribPointerARB;

static PFNGLGENFRAMEBUFFERSPROC glGenFramebuffers;
static PFNGLBINDFRAMEBUFFERPROC glBindFramebuffer;
static PFNGLDELETEFRAMEBUFFERSPROC glDeleteFramebuffers;
static PFNGLFRAMEBUFFERTEXTUREPROC glFramebufferTexture;
static PFNGLFRAMEBUFFERTEXTURE2DPROC glFramebufferTexture2D;
static PFNGLDRAWBUFFERSPROC glDrawBuffers;
static PFNGLGENERATEMIPMAPPROC glGenerateMipmap;

float color_black[] = { 0.0f, 0.0f, 0.0f, 0.0f };
int uniform_found;
int m_in_draw;
int m_log_errors = 1;

GLfloat g_vcoords[] = {
   -1.0f, -1.0f,
   1.0f, -1.0f,
   -1.0f, 1.0f,
   1.0f, 1.0f
};

GLfloat g_vcoords_backbuffer[] = {
   -1.0, 1.0f,
   1.0f, 1.0f,
   -1.0f, -1.0f,
   1.0f, -1.0f
};
    
GLfloat g_tcoords[] = {
   0.0f, 0.0f,
   1.0f, 0.0f,
   0.0f, 1.0f,
   1.0f, 1.0f
};

char *stock_shader = " \
\n\
#if defined(VERTEX) \n\
\n\
#if __VERSION__ >= 130 \n\
#define COMPAT_VARYING out \n\
#define COMPAT_ATTRIBUTE in \n\
#define COMPAT_TEXTURE texture \n\
#else \n\
#define COMPAT_VARYING varying  \n\
#define COMPAT_ATTRIBUTE attribute  \n\
#define COMPAT_TEXTURE texture2D \n\
#endif \n\
\n\
#ifdef GL_ES \n\
#define COMPAT_PRECISION mediump \n\
#else \n\
#define COMPAT_PRECISION \n\
#endif \n\
\n\
COMPAT_ATTRIBUTE vec4 VertexCoord;\n\
COMPAT_ATTRIBUTE vec4 COLOR;\n\
COMPAT_ATTRIBUTE vec4 TexCoord;\n\
COMPAT_VARYING vec4 COL0;\n\
COMPAT_VARYING vec4 TEX0;\n\
\n\
vec4 _oPosition1; \n\
uniform mat4 MVPMatrix;\n\
\n\
void main()\n\
{\n\
    vec4 _oColor;\n\
    vec2 _otexCoord;\n\
    gl_Position = VertexCoord.x * MVPMatrix[0] + VertexCoord.y * MVPMatrix[1] + VertexCoord.z * MVPMatrix[2] + VertexCoord.w * MVPMatrix[3];\n\
    _oPosition1 = gl_Position;\n\
    _oColor = COLOR;\n\
    _otexCoord = TexCoord.xy;\n\
    COL0 = COLOR;\n\
    TEX0.xy = TexCoord.xy;\n\
}\n\
\n\
#elif defined(FRAGMENT)\n\
\n\
#if __VERSION__ >= 130\n\
#define COMPAT_VARYING in\n\
#define COMPAT_TEXTURE texture\n\
out vec4 FragColor;\n\
#else\n\
#define COMPAT_VARYING varying\n\
#define FragColor gl_FragColor\n\
#define COMPAT_TEXTURE texture2D\n\
#endif\n\
\n\
#ifdef GL_ES\n\
#ifdef GL_FRAGMENT_PRECISION_HIGH\n\
precision highp float;\n\
#else\n\
precision mediump float;\n\
#endif\n\
#define COMPAT_PRECISION mediump\n\
#else\n\
#define COMPAT_PRECISION\n\
#endif\n\
\n\
struct output_dummy {\n\
    vec4 _color;\n\
};\n\
\n\
uniform sampler2D Texture;\n\
COMPAT_VARYING vec4 TEX0;\n\
//standard texture sample looks like this: COMPAT_TEXTURE(Texture, TEX0.xy);\n\
\n\
void main()\n\
{\n\
    output_dummy _OUT;\n\
    _OUT._color = COMPAT_TEXTURE(Texture, TEX0.xy);\n\
    FragColor = _OUT._color;\n\
    return;\n\
} \n\
#endif\n\
";

/******************************************************************************
**  Function   :  shader_set_defaults                                            
**                                                                    
**  Objective  :  This function sets up defaults for shader settings
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success
**                1 if failure
******************************************************************************/ 
int shader_set_defaults ( void ) {

    strcpy ( g_master.name, SHADER_NONE );
    g_master.shaders = 0;
    g_master.profile_dirty = 0;
    g_master.file = 0;
    g_master.num_shaders = 0;

    return 0;

} /* end shader_set_defaults */

t_master_shader * shader_get_master() {
    return &g_master;
}

/************************************************************************
 *
 *  Profile functions
 *
 ************************************************************************/

/******************************************************************************
**  Function   :  shader_set_profile                                            
**                                                                    
**  Objective  :  This function loads up an sound profile from db and puts it
**                in our global sound struct
**                                                   
**  Parameters :  name - name of sound profile to load
**                tex_rect - orig texture rect, check for changing
**                 
**  return     :  1 if using a shader, otherwise 0
**
******************************************************************************/ 
int shader_set_profile ( char *name, SDL_Rect tex_rect ) {

    m_in_draw = 0;
    m_log_errors = 1;

    int i;

    /* figure out if we need to reload and recompile */
    if (strcmp(name, g_master.name) || !g_master.context)
        g_master.profile_dirty = 1;

    if ( g_master.orig.id ) {
        if ( tex_rect.x != g_master.orig.dimen.x ||
             tex_rect.y != g_master.orig.dimen.y ||
             tex_rect.w != g_master.orig.dimen.w ||
             tex_rect.h != g_master.orig.dimen.h )
            g_master.profile_dirty = 1;
    }

    /* Complete Load of Profile including compile */
    if ( g_master.profile_dirty ) {

        shader_free_master ( &g_master );

        for ( i = 0; i < g_master.num_shaders; ++i ) {
            shader_free_program(&g_master.shaders[i]);
            shader_free_objects(&g_master.shaders[i]);
        }

        shader_free_master_contents ( &g_master );

        if ( *name == '\0' )
            shader_set_defaults();
        else {
            if ( !shader_load_profile ( name, &g_master ) ) {
                shader_open_master ( g_master.name,  &g_master );
                g_master.profile_dirty = 1;
                return 1;
            }
            shader_set_defaults();
        }
    }

    /* Else nothing changed, use what we have */
    else {

        return 1;
    }

    return 0;

} /* end shader_set_profile */

/******************************************************************************
**  Function   :  shader_load_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  sound     - pointer to sound struct
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**                 
**  return     :  0 if success
**                -1 if failure
******************************************************************************/ 
int shader_load_callback ( void *shader, int argc, char **argv, char **azColName ) {

    int i;
    t_master_shader *p_shader = shader;

    if ( argc < 2 )
        return -1;

    /*
     * THIS DEPENDS ON MATCHING THE DATABASE!
     */
    i = -1;
    if ( argv[++i] ) strcpy ( p_shader->name, argv[i] );
    if ( argv[++i] ) { 
        p_shader->file = malloc(strlen(argv[i]) + 1);
        strcpy ( p_shader->file, argv[i] );
    }

    return 0;

} /* end shader_load_callback */

/******************************************************************************
**  Function   :  shader_load_profile                                            
**                                                                    
**  Objective  :  This function loads up a shader profile from db
**                                                   
**  Parameters :  name     - name of shader profile
**                p_shader  - shader structure to load with info
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int shader_load_profile ( const char *name, t_master_shader *p_shader ) {
    
    char statement[257];
    int status;
    char *p_tmp;

    /*
     * Create the SQL query statement
     */
    sprintf ( statement, "SELECT * FROM MasterShader WHERE Name='%s'", 
                p_tmp=util_replace_and_copy(name, "'", "''" ) );

    free ( p_tmp );

    status = db_if_exec_sql ( statement, shader_load_callback, p_shader );

    return status;

} /* end shader_load_profile */

/******************************************************************************
**  Function   :  shader_save_profile                                            
**                                                                    
**  Objective  :  This function saves the shader configuration to db
**                                                   
**  Parameters :  name     - name of shader profile to save
**                p_shader - shader structure containing info to save
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int shader_save_profile ( char *name, t_master_shader *p_shader ) {

    char *statement;
    char *p_tmp[2];
    int status = 0;
    int i,j;

    statement = malloc ( strlen(name) + strlen(p_shader->file) + 257 );

    /*
     * Create the INSERT statement and execute
     */
    sprintf ( statement, "INSERT OR REPLACE INTO MasterShader VALUES ( '%s', '%s' )",
             p_tmp[0] = util_replace_and_copy(name, "'", "''" ),
             p_tmp[1] = util_replace_and_copy(p_shader->file, "'", "''" ));

    status = db_if_exec_sql ( statement, 0, 0 );

    for ( i = 0; i < 2; ++i )
        free ( p_tmp[i] );

    free ( statement );

    /*
     * Save shaders
     */
    for ( i = 0; i < p_shader->num_shaders; ++i ) {
        for ( j = 0 ; j < p_shader->shaders[i].num_params; ++j ) {
            shader_save_uniform ( name, p_shader->shaders[i].file, 
                                       &p_shader->shaders[i].params[j] );
        }
    }

    return status;

} /* end shader_save_profile */

/******************************************************************************
**  Function   :  shader_uniform_callback                                            
**                                                                    
**  Objective  :  This function is called by sqlite in response to the query
**                                                   
**  Parameters :  param     - pointer to param struct
**                argc      - number of columns returned
**                argv      - array of strings containing data
**                azColName - array of strings containing column names
**                 
**  return     :  0 if success
**                -1 if failure
******************************************************************************/ 
int shader_uniform_callback ( void *param, int argc, char **argv, char **azColName ) {

    int i;
    t_param *p_param = param;

    if ( argc < 9 )
        return -1;

    /*
     * THIS DEPENDS ON MATCHING THE DATABASE!
     */
    i = 1;
    if ( argv[++i] ) strcpy ( p_param->name, argv[i] );
    if ( argv[++i] ) strcpy ( p_param->description, argv[i] );
    if ( argv[++i] ) p_param->val = atof ( argv[i] );
    if ( argv[++i] ) p_param->def = atof ( argv[i] );
    if ( argv[++i] ) p_param->min = atof ( argv[i] );
    if ( argv[++i] ) p_param->max = atof ( argv[i] );
    if ( argv[++i] ) p_param->step = atof ( argv[i] );

    uniform_found = 1;

    return 0;

} /* end shader_uniform_callback */

/******************************************************************************
**  Function   :  shader_load_uniform                                            
**                                                                    
**  Objective  :  This function loads up a shader profile from db
**                                                   
**  Parameters :  name     - name of shader profile
**                p_shader  - shader structure to load with info
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int shader_load_uniform ( const char *name, const char *master, 
                          const char *shader, t_param *p_param ) {
    
    char statement[257];
    int status;
    char *p_tmp1, *p_tmp2, *p_tmp3;

    uniform_found = 0;

    /*
     * Create the SQL query statement
     */
    sprintf ( statement, "SELECT * FROM ShaderUniforms WHERE Name='%s' AND Master='%s' AND File='%s'", 
                p_tmp1=util_replace_and_copy(name, "'", "''" ),
                   p_tmp2=util_replace_and_copy(master, "'", "''" ),
                p_tmp3=util_replace_and_copy(shader, "'", "''" ) );

    free ( p_tmp1 );
    free ( p_tmp2 );
    free ( p_tmp3 );

    status = db_if_exec_sql ( statement, shader_uniform_callback, p_param );

    if ( !uniform_found ) 
        status = -1;

    return status;

} /* end shader_load_uniform */

/******************************************************************************
**  Function   :  shader_save_uniform                                            
**                                                                    
**  Objective  :  This function saves the uniform
**                                                   
**  Parameters :  name     - name of shader profile to save
**                p_shader - shader structure containing info to save
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int shader_save_uniform ( char *master, char *shader, t_param *p_param ) {

    char *statement;
    char *p_tmp[4];
    int status = 0;
    int i;

    statement = malloc ( strlen(master) + strlen(shader) + 
                                        + 600 );
    /*
     * Create the INSERT statement and execute
     */
    sprintf ( statement, "INSERT OR REPLACE INTO ShaderUniforms VALUES ( '%s', '%s', '%s', '%s', '%.3f', '%.3f', '%.3f', '%.3f', '%.3f' )",
             p_tmp[1] = util_replace_and_copy(shader, "'", "''" ),
             p_tmp[0] = util_replace_and_copy(master, "'", "''" ),
             p_tmp[2] = util_replace_and_copy(p_param->name, "'", "''" ),
             p_tmp[3] = util_replace_and_copy(p_param->description, "'", "''" ),
             p_param->val, p_param->def, p_param->min, p_param->max, 
             p_param->step );

    status = db_if_exec_sql ( statement, 0, 0 );

    for ( i = 0; i < 4; ++i )
        free ( p_tmp[i] );

    free ( statement );

    return status;

} /* end shader_save_uniform */

int shader_master_param ( char *line, t_master_shader *pm_shader ) {
    char *end, *tmp;
    char *i, *j;
    int shader_num;
    char st[257];
    t_shader *tmp_shader;

    if ( line[0] == '#' )
        return -1;

    if( sscanf(line, "%[^0123456789=]%d=", st, &shader_num) !=2 )
        return 1;

    end = strchr ( line, '=' );
    if ( !end )
        return -1;

    if ( shader_num >= 0 && shader_num <= pm_shader->num_shaders ) {

        tmp_shader = &pm_shader->shaders[shader_num];

        i = j = end+1;
        if ( (tmp = strchr(end, '\"')) ) {
            if ( strchr(tmp+1, '\"')) {
                end = tmp;
                *(strchr(end+1,'\"')) = '\0';
            }
        }
        else {
            do { *i = *j; if(!isspace(*i)) i++; } while(*j++ != 0);
        }

        if ( strstr ( line, "shader" ) ) {
            tmp_shader->file = malloc(strlen(end+1) + 1);
            strcpy ( tmp_shader->file, end+1 );
            return 0;
        }

        i = end+1; for ( ; *i; ++i) *i = tolower(*i);

        if ( strstr ( line, "filter_linear" ) ) {
            tmp_shader->filter_linear = 
                strstr ( end+1, "true" ) || strchr ( end+1, '1' ) ? 1 : 0;
        }
        else if ( strstr ( line, "float_framebuffer" ) ) {
            tmp_shader->float_framebuffer = 
                strstr ( end+1, "true" ) || strchr ( end+1, '1' ) ? 1 : 0;
        }
        else if ( strstr ( line, "mipmap_input" ) ) {
            tmp_shader->mipmap_input = 
                strstr ( end+1, "true" ) || strchr ( end+1, '1' ) ? 1 : 0;
        }
        else if ( strstr ( line, "wrap_mode" ) ) {
            if ( strstr ( end+1, "clamp_to_edge" ) )
                tmp_shader->wrap_mode = GL_CLAMP_TO_EDGE;
            else if ( strstr ( end+1, "mirrored_repeat" ) )
                tmp_shader->wrap_mode = GL_MIRRORED_REPEAT;
            else if ( strstr ( end+1, "repeat" ) )
                tmp_shader->wrap_mode = GL_REPEAT;
            else
                tmp_shader->wrap_mode = GL_CLAMP_TO_BORDER;
        }
        else if ( strstr ( line, "srgb_framebuffer" ) ) {
            tmp_shader->srgb_framebuffer = 
                strstr ( end+1, "true" ) || strchr ( end+1, '1' ) ? 1 : 0;
        }
        else if ( strstr ( line, "frame_count_mod" ) ) {
            tmp_shader->frame_count_mod = atoi ( end+1 ); 
        }
        else if ( strstr ( line, "scale_type_x" ) ) {
            strcpy ( tmp_shader->scale_type_x, end+1 );
        }
        else if ( strstr ( line, "scale_type_y" ) ) {
            strcpy ( tmp_shader->scale_type_y, end+1 );
        }
        else if ( strstr ( line, "scale_type" ) ) {
            strcpy ( tmp_shader->scale_type_x, end+1 );
            strcpy ( tmp_shader->scale_type_y, end+1 );
        }
        else if ( strstr ( line, "scale_x" ) ) {
            strcpy ( tmp_shader->scale_x, end+1 );
        }
        else if ( strstr ( line, "scale_y" ) ) {
            strcpy ( tmp_shader->scale_y, end+1 );
        }
        else if ( strstr ( line, "scale" ) ) {
            strcpy ( tmp_shader->scale_x, end+1 );
            strcpy ( tmp_shader->scale_y, end+1 );
        }
        else if ( strstr ( line, "alias" ) ) {
            strcpy ( tmp_shader->alias, end+1 );
        }
    }

    return 0;

} /* end shader_master_param */

int shader_texture_param ( char *line, t_master_shader *pm_shader ) {
    char tmp[257], *end, *tmp1;
    char *q, *r;
    t_lut_texture *tmp_texture;
    int j;

    end = strchr ( line, '=' );
    if ( !end )
        return -1;

    q = r = end+1;

    if ( (tmp1 = strchr(end, '\"')) ) {
        if ( strchr(tmp1+1, '\"')) {
            end = tmp1;
            *(strchr(end+1,'\"')) = '\0';
        }
    }
    else {
        do { *q = *r; if(!isspace(*q)) q++; } while(*r++ != 0);
    }
    for ( j = 0; j < pm_shader->num_lut_textures; ++j ) {
        tmp_texture = &pm_shader->lut_textures[j];
        strcpy ( tmp, tmp_texture->name );
        strcat ( tmp, "_" );
        if ( strstr ( line, tmp ) ) {
            q = end+1; for ( ; *q; ++q) *q = tolower(*q);
            if ( strstr ( line, "_linear" ) ) {
                tmp_texture->linear = 
                    strstr ( end+1, "true" ) || strchr ( end+1, '1' ) ? 1 : 0;
            }
            if ( strstr ( line, "_mipmap" ) ) {
                tmp_texture->mipmap = 
                    strstr ( end+1, "true" ) || strchr ( end+1, '1' ) ? 1 : 0;
            }
            if ( strstr ( line, "_repeat" ) ) {
                tmp_texture->wrap_mode = 
                    strstr ( end+1, "true" ) || strchr ( end+1, '1' ) ? 
                                              GL_REPEAT : GL_CLAMP_TO_BORDER;
            }
            if ( strstr ( line, "_wrap_mode" ) ) {
                if ( strstr ( end+1, "clamp_to_edge" ) )
                    tmp_texture->wrap_mode = GL_CLAMP_TO_EDGE;
                else if ( strstr ( end+1, "mirrored_repeat" ) )
                    tmp_texture->wrap_mode = GL_MIRRORED_REPEAT;
                else if ( strstr ( end+1, "repeat" ) )
                    tmp_texture->wrap_mode = GL_REPEAT;
                else
                    tmp_texture->wrap_mode = GL_CLAMP_TO_BORDER;
            }
        }
        else if ( strstr ( line, tmp_texture->name ) ) {
            tmp_texture->file = malloc(strlen(end+1) + 1);
            strcpy ( tmp_texture->file, end+1);
        }
    }

    return 0;

} /* end shader_texture_param */

int shader_uni_param ( char *line, t_master_shader *pm_shader ) {
    char *end, *tmp1;
    char *q, *r;
    t_param *tmp_param;
    int j;

    end = strchr ( line, '=' );
    if ( !end )
        return -1;

    q = r = end+1;

    if ( (tmp1 = strchr(end, '\"')) ) {
        if ( strchr(tmp1+1, '\"')) {
            end = tmp1;
            *(strchr(end+1,'\"')) = '\0';
        }
    }
    else {
        do { *q = *r; if(!isspace(*q)) q++; } while(*r++ != 0);
    }
    for ( j = 0; j < pm_shader->num_params; ++j ) {
        tmp_param = &pm_shader->params[j];
        if ( strstr ( line, tmp_param->name ) ) {
            q = end+1; for ( ; *q; ++q) *q = tolower(*q);
            tmp_param->val = atof ( end+1 );
        }
    }

    return 0;

} /* end shader_uni_param */

/******************************************************************************
**  Function   :  shader_open_master
**                            
**  Objective  :  This function reads the glslp shader code from a file
**
**  Parameters :  name       - name of master shader from database 
**                pm_shader  - master struct to fill in
**                                                
**  return     :  Error code
**      
******************************************************************************/
int shader_open_master ( char *name, t_master_shader *pm_shader ) {

    char msg[257];
    char new_name[257];
    char line[257];
    char *p_ptr, *token;
    FILE *fp;
    int i;

    /*
     * Open the shader file and reset our data
     */
    if ( !strcmp(name, SHADER_NONE) )
        return -1;

    if ((fp = fopen(pm_shader->file,"r")) == NULL) {
        sprintf ( msg, "Could not open master shader file: %s", name );
        logger_log_message ( LOG_ERROR, msg, LOG_SRC );
        return -1;
    }

    if (pm_shader->shaders)
        shader_free_master_contents ( pm_shader );

    /*
     * Read each line from file and fill the struct
     */
    while (fgets(line, sizeof(line), fp)) {
        util_search_and_replace ( line, " ", "" );
        util_search_and_replace ( line, "\"", "" );
        util_search_and_replace ( line, "\n", "" );

        /* Ignore comments */
        if ( line[0] == '#' )
            continue;

        if ( strstr ( line, "shaders=" ) ) {
            p_ptr = strchr ( line, '=' );
            if ( p_ptr ) {
                p_ptr++;
                pm_shader->num_shaders = atoi ( p_ptr );
                pm_shader->shaders = malloc(pm_shader->num_shaders * 
                                          sizeof(*(pm_shader->shaders)));
                for ( i = 0; i < pm_shader->num_shaders; ++i ) {
                    shader_init_shader ( &(pm_shader->shaders[i]) );
                }
            }
        }
        else if ( strstr ( line, "textures=" ) ) {
            int num_textures = 0;
            p_ptr = strchr ( line, '=' ) + 1;
            for (i=0; p_ptr[i]; p_ptr[i]==';' ? i++ : *p_ptr++);
            pm_shader->num_lut_textures = i+1;
            pm_shader->lut_textures = malloc((i+1) * 
                                          sizeof(*(pm_shader->lut_textures)));
            p_ptr = strchr ( line, '=' ) + 1;
            token = strtok ( p_ptr, ";" ); 
            while (token) {
                strcpy ( pm_shader->lut_textures[num_textures].name, token );
                pm_shader->lut_textures[num_textures].file = 0;
                pm_shader->lut_textures[num_textures].linear = 1;
                pm_shader->lut_textures[num_textures].wrap_mode = GL_CLAMP_TO_BORDER;
                pm_shader->lut_textures[num_textures].mipmap = 0;
                num_textures++; 
                token = strtok ( NULL, ";" ); 
            }
        }
        else if ( strstr ( line, "parameters=" ) ) {
            int num_params = 0;
            p_ptr = strchr ( line, '=' ) + 1;
            for (i=0; p_ptr[i]; p_ptr[i]==';' ? i++ : *p_ptr++);
            pm_shader->num_params = i+1;
            pm_shader->params = malloc((i+1) * 
                                          sizeof(*(pm_shader->params)));
            p_ptr = strchr ( line, '=' ) + 1;
            token = strtok ( p_ptr, ";" ); 
            while (token) {
                strcpy ( pm_shader->params[num_params].name, token );
                pm_shader->params[num_params].val = 0;
                pm_shader->params[num_params].def= 0;
                pm_shader->params[num_params].min= 0;
                pm_shader->params[num_params].max= 0;
                pm_shader->params[num_params].step= 0;
                num_params++; 
                token = strtok ( NULL, ";" ); 
            }
        }
        else if ( strstr ( line, "feedback_pass=" ) ) {
            p_ptr = strchr ( line, '=' );
            if ( p_ptr ) {
                p_ptr++;
                pm_shader->prev.feedback_pass = atoi ( p_ptr );
            }
        }
    }

    /*
     * Now that we have shaders allocated, read again for data
     */
    fseek (fp, 0, SEEK_SET);

    while (fgets(line, sizeof(line), fp)) {
        shader_master_param ( line, pm_shader );
    }

    /*
     * Read through again for the textures and common uniforms
     */
    fseek (fp, 0, SEEK_SET);

    while (fgets(line, sizeof(line), fp)) {
        shader_texture_param ( line, pm_shader );
        shader_uni_param ( line, pm_shader );
    }

    fclose ( fp );

    /*
     * Now save to the database
     */
    char *tmp_ptr = 0;
    if ( !strlen(name) || util_all_whitespace(name) ) {
        if ( !(tmp_ptr = strrchr(name, '\\')) ) {
            if ( !(tmp_ptr = strrchr(name, '/')) )
                tmp_ptr = name;
        }
        strcpy ( new_name, tmp_ptr );
        if ( (tmp_ptr = strchr(new_name, '.')) )
            *tmp_ptr = '\0';
    }
    else {
        if ( strlen(name) > 256) name[257] = '\0';
        strcpy ( new_name, name);
    }

    if ( shader_save_profile ( new_name, pm_shader ) )
        return -1;

    /*
     * Now open our shaders and store uniforms if needed
     */
    char *full_file;
    for ( i = 0; i < pm_shader->num_shaders; i++ ) {

        /* Create absolute filename */
        full_file = malloc ( strlen(pm_shader->file) + 
                             strlen(pm_shader->shaders[i].file) );

        strcpy ( full_file, pm_shader->file );
        util_strip_file_to_directory ( full_file );
        sprintf ( full_file, "%s%c%s", full_file, 
                      util_get_dir_seperator_character(), 
                      pm_shader->shaders[i].file );

        shader_open_shader ( full_file, name, &pm_shader->shaders[i] );

        free ( full_file );
    }

    return 0;

} /* end shader_open_master */

/******************************************************************************
**  Function   :  shader_open_shader
**                            
**  Objective  :  This function reads the glsl shader code from a file
**
**  Parameters :  file  - name of glsl file containing shader code
**                master - name of shader
**                                                
**  return     :  Error code
**      
******************************************************************************/
int shader_open_shader ( char *file, char *master, t_shader *shader ) {

    char msg[257];
    int length, length_read;
    FILE *fp;

    /*
     * Open the shader file and get the code
     */
    if ((fp = fopen(file,"r")) == NULL) {
        sprintf ( msg, "Could not open shader file: %s", file );
        logger_log_message ( LOG_ERROR, msg, LOG_SRC );
        return -1;
    }

    //shader->file = malloc(strlen(file) + 1);
    //strcpy ( shader->file, file );

    fseek (fp, 0, SEEK_END);
    length = ftell (fp);
    fseek (fp, 0, SEEK_SET);
    shader->source = malloc (length + 1);
    length_read = fread (shader->source, 1, length, fp);
    shader->source[length] = '\0';

    fclose ( fp );

    if ( !shader->source || (length_read != length) ) {
        sprintf ( msg, "Error reading from shader file: %s", file );
        logger_log_message ( LOG_ERROR, msg, LOG_SRC );
        if (shader->source) free(shader->source);
        if (shader->file) free(shader->file);
        shader->source = 0;
        shader->file = 0;
        return -1;
    }

    /*
     * Go through our pragmas and update the uniforms
     */
    char *end_line;
    char *line;
    char *token;
    char *start_line = shader->source;
    char tmp_line[257];
    char search[10];
    int param_num = 0;;
    int col_num = 0;;
    int in_string = 0;
    int status;

    /* Go through once just to count our uniforms and alloc memory */
    while ( (end_line = strchr( start_line, '\n' )) ) {
        strncpy ( tmp_line, start_line, end_line-start_line);
        tmp_line[end_line-start_line+1] = '\0';
        if ( strstr(tmp_line, "#pragma") && strstr(tmp_line, "parameter") ) {
            param_num++;
        }
        start_line = end_line + 1;
    }

    if ( param_num ) {
        shader->params = malloc ( sizeof(*(shader->params)) * param_num );
        shader->num_params = param_num;
    }
    else {
        shader->num_params = 0;
        return 0;
    }

    param_num = -1;
    start_line = shader->source;

    /* Go through again and grab the data */
    while ( (end_line = strchr( start_line, '\n' )) ) {
        *end_line = '\0';
        if ( strstr(start_line, "#pragma") && strstr(start_line, "parameter") ) {
            param_num++;
            strcpy ( shader->params[param_num].description, "" );
            col_num = 0;
            line = strdup ( start_line );
            strcpy ( search, " " );
            token = strtok ( line, search ); 

            while (token) {
                if (strstr(token, "#pragma") || strstr(token, "parameter")) {
                    token = strtok ( NULL, search ); 
                    continue;
                }

                if (!in_string) {
                    switch (col_num) {
                        case 0:
                            strcpy ( shader->params[param_num].name, token );
                            status =  shader_load_uniform ( shader->params[param_num].name, master, shader->file, &shader->params[param_num] );
                            strcpy ( search, "\"" );
                            break;
                        case 1:
                            strcpy ( shader->params[param_num].description, token );
                            strcpy ( search, " " );
                            break;
                        case 2:
                            shader->params[param_num].def = atof ( token );
                            break;
                        case 3:
                            shader->params[param_num].min = atof ( token );
                            break;
                        case 4:
                            shader->params[param_num].max = atof ( token );
                            break;
                        case 5:
                            shader->params[param_num].step = atof ( token );
                            if ( status ) {
                                shader->params[param_num].val = shader->params[param_num].def;
                                shader_save_uniform ( master, shader->file, &shader->params[param_num] );
                            }
                            break;
                    }
                    col_num++;
                }

                token = strtok ( NULL, search ); 

            } /* end while token */

            free ( line );

        } /* end if contains pragma parameter */

        *end_line = '\n';
        start_line = end_line + 1;

    } /* end while reading lines */

    return 0;

} /* end shader_open_shader */

/******************************************************************************
**  Function   :  shader_init
**                                                                    
**  Objective  :  This function checks for shader support and sets defaults
**                                                   
**  Parameters :  window to setup a context for
**                 
**  return     :  0 if success, otherwise failure
**
******************************************************************************/ 
int shader_init( SDL_Window *window ) {

    char msg[512];

    /* Check for shader support */
    shader_init_master ( &g_master );
    g_master.supported = 0;

    if ( shader_setup_context ( window ) ) {
        g_master.supported = 0;
        return -1;
    }

    glActiveTexture2 = (PFNGLACTIVETEXTUREPROC) SDL_GL_GetProcAddress("glActiveTexture");
    glAttachShaderARB = (PFNGLATTACHSHADERPROC) SDL_GL_GetProcAddress("glAttachShader");
    glCompileShaderARB = (PFNGLCOMPILESHADERPROC) SDL_GL_GetProcAddress("glCompileShader");
    glCreateProgramARB = (PFNGLCREATEPROGRAMPROC) SDL_GL_GetProcAddress("glCreateProgram");
    glCreateShaderARB = (PFNGLCREATESHADERPROC) SDL_GL_GetProcAddress("glCreateShader");
    glDetachShaderARB = (PFNGLDETACHSHADERPROC) SDL_GL_GetProcAddress("glDetachShader");
    glDeleteShaderARB = (PFNGLDELETESHADERPROC) SDL_GL_GetProcAddress("glDeleteShader");
    glDeleteProgramARB = (PFNGLDELETEPROGRAMPROC) SDL_GL_GetProcAddress("glDeleteProgram");
    glGetShaderInfoLogARB = (PFNGLGETSHADERINFOLOGPROC) SDL_GL_GetProcAddress("glGetShaderInfoLog");
    glGetShaderivARB = (PFNGLGETSHADERIVPROC) SDL_GL_GetProcAddress("glGetShaderiv");
    glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONPROC) SDL_GL_GetProcAddress("glGetUniformLocation");
    glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONPROC) SDL_GL_GetProcAddress("glGetAttribLocation");
    glEnableVertexAttribArrayARB = (PFNGLENABLEVERTEXATTRIBARRAYPROC ) SDL_GL_GetProcAddress("glEnableVertexAttribArray");
    glVertexAttribPointerARB= (PFNGLVERTEXATTRIBPOINTERPROC ) SDL_GL_GetProcAddress("glVertexAttribPointer");
    glLinkProgramARB = (PFNGLLINKPROGRAMPROC) SDL_GL_GetProcAddress("glLinkProgram");
    glShaderSourceARB = (PFNGLSHADERSOURCEPROC) SDL_GL_GetProcAddress("glShaderSource");
    glUniform1iARB = (PFNGLUNIFORM1IPROC) SDL_GL_GetProcAddress("glUniform1i");
    glUniform1fARB = (PFNGLUNIFORM1FPROC) SDL_GL_GetProcAddress("glUniform1f");
    glUniform2fvARB = (PFNGLUNIFORM2FVPROC) SDL_GL_GetProcAddress("glUniform2fv");
    glUniformMatrix4fvARB = (PFNGLUNIFORMMATRIX4FVPROC) SDL_GL_GetProcAddress("glUniformMatrix4fv");
    glUseProgramARB = (PFNGLUSEPROGRAMPROC) SDL_GL_GetProcAddress("glUseProgram");
    glGenVertexArraysARB = (PFNGLGENVERTEXARRAYSPROC) SDL_GL_GetProcAddress("glGenVertexArrays");
    glBindVertexArrayARB = (PFNGLBINDVERTEXARRAYPROC) SDL_GL_GetProcAddress("glBindVertexArray");
    glDeleteVertexArraysARB = (PFNGLDELETEVERTEXARRAYSPROC) SDL_GL_GetProcAddress("glDeleteVertexArrays");
    glGenBuffersARB = (PFNGLGENBUFFERSPROC) SDL_GL_GetProcAddress("glGenBuffers");
    glBindBufferARB = (PFNGLBINDBUFFERPROC) SDL_GL_GetProcAddress("glBindBuffer");
    glBufferDataARB = (PFNGLBUFFERDATAPROC) SDL_GL_GetProcAddress("glBufferData");
    glBufferSubDataARB = (PFNGLBUFFERSUBDATAPROC) SDL_GL_GetProcAddress("glBufferSubData");
    glDeleteBuffersARB = (PFNGLDELETEBUFFERSPROC) SDL_GL_GetProcAddress("glDeleteBuffers");
    glGenFramebuffers = (PFNGLGENFRAMEBUFFERSPROC) SDL_GL_GetProcAddress("glGenFramebuffers");
    glDeleteFramebuffers = (PFNGLDELETEFRAMEBUFFERSPROC) SDL_GL_GetProcAddress("glDeleteFramebuffers");
    glBindFramebuffer = (PFNGLBINDFRAMEBUFFERPROC) SDL_GL_GetProcAddress("glBindFramebuffer");
    glFramebufferTexture = (PFNGLFRAMEBUFFERTEXTUREPROC) SDL_GL_GetProcAddress("glFramebufferTexture");
    glFramebufferTexture2D = (PFNGLFRAMEBUFFERTEXTURE2DPROC) SDL_GL_GetProcAddress("glFramebufferTexture2D");
    glDrawBuffers = (PFNGLDRAWBUFFERSPROC) SDL_GL_GetProcAddress("glDrawBuffers");
    glGenerateMipmap = (PFNGLGENERATEMIPMAPPROC) SDL_GL_GetProcAddress("glGenerateMipmap");

    if (glAttachShaderARB &&
        glActiveTexture2 &&
        glCompileShaderARB &&
        glCreateProgramARB &&
        glCreateShaderARB &&
        glDetachShaderARB &&
        glDeleteShaderARB &&
        glDeleteProgramARB &&
        glGetShaderInfoLogARB &&
        glGetShaderivARB &&
        glGetUniformLocationARB &&
        glGetAttribLocationARB &&
        glVertexAttribPointerARB &&
        glEnableVertexAttribArrayARB &&
        glLinkProgramARB &&
        glShaderSourceARB &&
        glGenVertexArraysARB &&
        glBindVertexArrayARB &&
        glDeleteVertexArraysARB &&
        glGenBuffersARB &&
        glDeleteBuffersARB &&
        glBindBufferARB &&
        glBufferDataARB &&
        glBufferSubDataARB &&
        glUniform1iARB &&
        glUniform1fARB &&
        glUniform2fvARB &&
        glUniformMatrix4fvARB &&
        glGenFramebuffers &&
        glDeleteFramebuffers &&
        glBindFramebuffer &&
        glFramebufferTexture &&
        glFramebufferTexture2D &&
        glDrawBuffers &&
        glGenerateMipmap &&
        glUseProgramARB) {
        g_master.supported = 1;
    }

    if (!g_master.supported) {
        logger_log_message ( LOG_ERROR, "Shaders not supported on this system", 
                             LOG_SRC );

        if ( g_master.context )
            SDL_GL_DeleteContext(g_master.context);

        g_master.context = 0;
        return -1;
    }

    /*
     * Log actual OpenGL version and vendor
     */
    sprintf ( msg, "OpenGL Version - %s", glGetString(GL_VERSION));
    logger_log_message ( LOG_INFO, msg, LOG_SRC );
    sprintf ( msg, "       Vendor  - %s", glGetString(GL_VENDOR));
    logger_log_message ( LOG_INFO, msg, LOG_SRC );
    sprintf ( msg, "       Renderer  - %s", glGetString(GL_RENDERER));
    logger_log_message ( LOG_INFO, msg, LOG_SRC );

    shader_set_defaults ();

    /* We have shader support */

    return 0;

} /* end shader_init */

/******************************************************************************
**  Function   :  shader_setup_context
**                                                                    
**  Objective  :  This function does sets up the context for the SDL window.
**                Should be called whenever the window changes.
**                                                   
**  Parameters :  window - pointer to SDL window to tie this to.
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_setup_context ( SDL_Window *window ) {

    if ( !(g_master.context = SDL_GL_CreateContext(window))) {
        logger_log_message (LOG_ERROR, "Failed to create OpenGL context", LOG_SRC);
        return -1;
    }

    return 0;

} /* end shader_setup_context */

/******************************************************************************
**  Function   :  shader_set_viewport_values 
**                                                                    
**  Objective  :  This function sets the width and height of the viewport for 
**                the window.  We do it in a seperate function since the user
**                could resize window and we can set this on resize.
**                                                   
**  Parameters :  win_w - window width in pixels
**                win_h - window height in pixels
**                tex_w - texture width in pixels
**                tex_h - texture height in pixels
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_set_viewport_values ( int win_w, int win_h, int out_w, int out_h ) {

    double aspect;
    double win_aspect;
    int x,y, width, height;

    /* We need to set this based on the aspect ratio of the texture */
    aspect = (double)out_w / (double)out_h;
    win_aspect = (double)win_w / (double)win_h;

    if (aspect > win_aspect) {
        width = win_w;
        height = width / aspect;
    }
    else {
        height = win_h;
        width = height * aspect;
    }

    x = (win_w - width) / 2;
    y = (win_h - height) / 2;

    g_master.viewport.x = x;
    g_master.viewport.y = y;
    g_master.viewport.w = width;
    g_master.viewport.h = height;

    return 0;

} /* end shader_set_viewport_values */

/******************************************************************************
**  Function   :  shader_update_viewport_values 
**                                                                    
**  Objective  :  This function updates the viewport when the user changes
**                window size.  We need to go through the shaders looking for
**                where output depends on viewport.  At that point we need
**                to recreate fbo texture and update tex coordinates.
**                                                   
**  Parameters :  win_w - window width in pixels
**                win_h - window height in pixels
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_update_viewport_values ( int win_w, int win_h ) {

    int i,j;
    int start_pass = -1;
    int filter_linear = 0;
    t_texture *input_texture;
    t_shader *p_shader;

    m_in_draw = 0;
    m_log_errors = 1;

    /* go ahead and set the viewport before changing shaders */
    shader_set_viewport_values ( win_w, win_h, g_master.orig.dimen.w, 
                                               g_master.orig.dimen.h );

    /* Set Viewport to full screen to force OpenGL to change backbuffer  */
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
    glViewport(0, 0, win_w, win_h);

    /* Detect if any shaders use viewport, we change all after */
    for ( i = 0; i < g_master.num_shaders; ++i ) {
        if ( strstr(g_master.shaders[i].scale_type_x, "viewport") || 
             strstr(g_master.shaders[i].scale_type_y, "viewport") ) {
            start_pass = i;
            break;
        }
    }

    /* If no shader uses viewport, exit now */
    if ( start_pass < 0 )
        return 0;

    /* Cycle through shaders and update */
    for ( i = start_pass; i < g_master.num_shaders; ++i ) {

		glUseProgramARB(g_master.shaders[i].program);

        if ( i == 0 )
            input_texture = &g_master.orig;
        else
            input_texture = &g_master.shaders[i-1].fbo_tex;

        p_shader = &g_master.shaders[i];

        /* Update Tex coords after first viewport use */
        if ( i > start_pass ) {

            glBindBufferARB(GL_ARRAY_BUFFER, g_master.shaders[i].vbo_tex);
            glBufferSubDataARB(GL_ARRAY_BUFFER, 0, 8 * sizeof(GLfloat), 
                    input_texture->coord);
        }

        /* Update input/output/texture size */
        shader_set_shader_sizes ( &g_master.shaders[i], input_texture,
                                      g_master.viewport );

        /* Re-create output FBO/Texture */
        if ( g_master.shaders[i].fbo ) {

            if (i+1 == g_master.num_shaders)
                filter_linear = 0;
            else
                filter_linear = g_master.shaders[i+1].filter_linear;

            p_shader->fbo_tex.dimen.x = input_texture->dimen.x;
            p_shader->fbo_tex.dimen.y = input_texture->dimen.y;
            p_shader->fbo_tex.dimen.w = p_shader->vals.output_size[0];
            p_shader->fbo_tex.dimen.h = p_shader->vals.output_size[1];

            shader_gen_texture ( &g_master, &p_shader->fbo_tex, 
              0, p_shader->wrap_mode, filter_linear, p_shader->float_framebuffer );

            if (i+1 == g_master.num_shaders && !g_master.framebuffer_out ) {
                glBindBufferARB(GL_ARRAY_BUFFER, g_master.screen_vbo_tex);
                glBufferSubDataARB(GL_ARRAY_BUFFER, 0, 8 * sizeof(GLfloat), 
                                    p_shader->fbo_tex.coord);
            }
        }

        /* Re-recreate feedback FBO */
        if ( i == g_master.prev.feedback_pass ) {

            g_master.prev.feedback_tex.dimen.x = 0;
            g_master.prev.feedback_tex.dimen.y = 0;
            g_master.prev.feedback_tex.dimen.w = 
                g_master.shaders[i].fbo_tex.texture_size[0];
            g_master.prev.feedback_tex.dimen.h = 
                g_master.shaders[i].fbo_tex.texture_size[1];

            shader_gen_texture ( &g_master, &g_master.prev.feedback_tex, 
              0, GL_CLAMP_TO_BORDER, 0, p_shader->float_framebuffer );

            // Update uniform in each shader
            for ( j = 0; j < g_master.num_shaders; ++j ) {
                glUseProgramARB(g_master.shaders[j].program);
                shader_set_texture_uniforms ( &g_master.shaders[j].locs.feedback, 
                                              &g_master.prev.feedback_tex );
            }

		    glUseProgramARB(g_master.shaders[i].program);
        }

        /* Re-setup previous pass sizes */
        for ( j = 1; j <= p_shader->num_passes; ++j ) {
            shader_set_texture_uniforms ( &g_master.shaders[i].locs.pass[j-1], 
                                      &g_master.shaders[j-1].fbo_tex );
        }

    }

    /* Set viewport based on last shader output */
    shader_set_viewport_values ( win_w, win_h, 
            g_master.shaders[g_master.num_shaders-1].vals.output_size[0], 
            g_master.shaders[g_master.num_shaders-1].vals.output_size[1] ); 

    return 0;

} /* end shader_update_viewport_values */

/******************************************************************************
**  Function   :  shader_setup
**                                                                    
**  Objective  :  This function does initial setup for use of the shader with
**                the renderer
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_setup ( int vsync, SDL_Window *window, SDL_Rect tex_rect ) {

    int i, j;

    m_in_draw = 0;
    m_log_errors = 1;

    if (!g_master.supported || !g_master.shaders)
        return -1;

    SDL_GL_MakeCurrent(window, g_master.context);

    if ( SDL_GL_SetSwapInterval(vsync) < 0 ) {
        logger_log_message (LOG_ERROR, "Failed to set VSYNC", LOG_SRC);
    }

    int w,h;
    SDL_GetWindowSize(window, &w, &h);

    glClearColor(0.0f, 0.0f, 0.0f, 0.0f); 
    glDisable(GL_DEPTH_TEST);        
    glActiveTexture2( GL_TEXTURE0 );

    /* SDL screws us, so set this before creating textures */
    /* why oh why SDL */
    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    //glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

    shader_check_error ( "Shader Init" );

    /*
     * Setup Orig Texture
     */
    g_master.tex_count = -1;
    g_master.orig.dimen.x = tex_rect.x;
    g_master.orig.dimen.y = tex_rect.y;
    g_master.orig.dimen.w = tex_rect.w;
    g_master.orig.dimen.h = tex_rect.h;

    shader_gen_texture ( &g_master, &g_master.orig, 0, GL_CLAMP_TO_BORDER, 
                         g_master.shaders[0].filter_linear ? GL_LINEAR : 
                         GL_NEAREST, 0 );
                            
    glActiveTexture2( GL_TEXTURE0 );
    glBindBufferARB(GL_ARRAY_BUFFER, 0);
    glBindVertexArrayARB(0);

    /*
     * We want viewport set per last shader, but since some shaders need
     *   viewport for scaling we start off based on orig texture
     */
    shader_set_viewport_values ( w, h, tex_rect.w, tex_rect.h );

    /*
     * Compile Shaders and kill existing ones
     */
    if ( g_master.profile_dirty ) {
        shader_create_lut_textures ( &g_master);
        for ( i = 0; i < g_master.num_shaders; ++i ) {

            if ( shader_compile_shader_program(&g_master.shaders[i]) )
                break;

            shader_set_values( &g_master, i );
        }

        /* if we failed, free previous resources */
        if ( i != g_master.num_shaders ) {

            for ( j = 0; j < i; ++j ) {
                shader_free_program( &g_master.shaders[i] );
                shader_free_objects( &g_master.shaders[i] );
            }

            shader_free_master_contents ( &g_master );
            glActiveTexture2( GL_TEXTURE0 );
            glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
            glBindVertexArrayARB(0);
            glUseProgramARB(0);
            return -1;
        }
    }

    /* Update vertex values of last shader, y-flipped for back buffer */
    //g_master.num_shaders = 1;
    //g_master.shaders[g_master.num_shaders-1].fbo = 0;
    glBindBufferARB(GL_ARRAY_BUFFER, 
            g_master.shaders[g_master.num_shaders-1].vbo_vertex);

    if ( !g_master.shaders[g_master.num_shaders-1].fbo ) {
        glBufferSubDataARB(GL_ARRAY_BUFFER, 0, 8 * sizeof(GLfloat), 
                g_vcoords_backbuffer);
    }

    /* Set viewport based on last shader output */
    shader_set_viewport_values ( w, h, 
            g_master.shaders[g_master.num_shaders-1].vals.output_size[0], 
            g_master.shaders[g_master.num_shaders-1].vals.output_size[1] ); 

    g_master.profile_dirty = 0;

    return 0;

} /* end shader_setup */

/******************************************************************************
**  Function   :  shader_draw
**                                                                    
**  Objective  :  This function draws the texture to screen with shader
**                                                   
**  Parameters :  name     - name of shader profile to save
**                p_shader - shader structure containing info to save
**                 
**  return     :  0 if error, otherwise success
**
******************************************************************************/ 
int shader_draw ( SDL_Window *window, void *pixels, int width ) {

    int i, j, count;

    m_in_draw = 1;
    glActiveTexture2( GL_TEXTURE0 );
    glPixelStorei(GL_UNPACK_ROW_LENGTH, width);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, g_master.orig.dimen.w, 
        g_master.orig.dimen.h, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, pixels);

    /* Run through the shaders passes */
    for ( i = 0; i < g_master.num_shaders; ++i ) {

        glUseProgramARB(g_master.shaders[i].program);

        /* Handle frame count if shader needs it */
        if ( g_master.shaders[i].locs.frame_count >= 0 ) {

            count = g_master.shaders[i].vals.frame_count++;
            if ( g_master.shaders[i].frame_count_mod )
                count %= g_master.shaders[i].frame_count_mod;

            glUniform1iARB(g_master.shaders[i].locs.frame_count, count);
        }

        /* Update Previous Frame uniforms if needed */
        for ( j = 0; j < MAX_PREV_FRAME; ++j ) {
            if ( g_master.shaders[i].locs.prev[j].tex_unit >= 0 ) {
                shader_set_texture_uniforms ( &g_master.shaders[i].locs.prev[j], 
                        &g_master.prev.tex[(j+g_master.prev.frame_ptr) % MAX_PREV_FRAME] );
            }
        }

        /* During first pass copy orig tex into previous frame buffers */
        if ( i == 0 ) {

            /* Copy contents of the original texture into the backup texture */
            if (g_master.prev.tex[g_master.prev.frame_ptr].id) {

                g_master.prev.frame_ptr--;
                if ( g_master.prev.frame_ptr < 0 )
                    g_master.prev.frame_ptr = 6;

                glUseProgramARB(g_master.prev.shader.program);

                glUniform1iARB(g_master.prev.shader.locs.texture, 0);
                glBindVertexArrayARB(g_master.prev.shader.vao);

                glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 
                                  g_master.prev.fbo[g_master.prev.frame_ptr]);
                glViewport(0, 0, 
                      g_master.prev.tex[g_master.prev.frame_ptr].texture_size[0], 
                      g_master.prev.tex[g_master.prev.frame_ptr].texture_size[1]);
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
                glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

                glUseProgramARB(g_master.shaders[i].program);
            }
        }

        /*
         * Now we draw for this shader pass
         */
        glBindVertexArrayARB(g_master.shaders[i].vao);

        glBindFramebuffer(GL_DRAW_FRAMEBUFFER, g_master.shaders[i].fbo);

        /* If drawing to FBO use tex viewport, else use screen viewport */
        if ( g_master.shaders[i].fbo ) {
            
            glViewport(g_master.shaders[i].fbo_tex.dimen.x,
                       g_master.shaders[i].fbo_tex.dimen.y, 
                       g_master.shaders[i].vals.output_size[0], 
                       g_master.shaders[i].vals.output_size[1]);
        }
        else {
            glViewport(g_master.viewport.x, g_master.viewport.y, 
                       g_master.viewport.w, g_master.viewport.h);
        }

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);


        /* If last shader and we drew to FBO, draw to screen */
        if ( i == (g_master.num_shaders-1) && g_master.shaders[i].fbo && 
                                              !g_master.framebuffer_out ) {

            glUseProgramARB(g_master.prev.shader.program);

            glUniform1iARB(g_master.prev.shader.locs.texture, 
                    g_master.shaders[i].fbo_tex.tex_unit);

            glBindVertexArrayARB(g_master.screen_vao);
            glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
            glViewport(g_master.viewport.x, g_master.viewport.y, 
                       g_master.viewport.w, g_master.viewport.h);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
        }

        /* If this is the feedback pass, copy the output to FBO */
        if ( i == g_master.prev.feedback_pass ) {

            glUseProgramARB(g_master.prev.shader.program);

            glUniform1iARB(g_master.prev.shader.locs.texture, 
                    g_master.shaders[i].fbo_tex.tex_unit);
            glBindVertexArrayARB(g_master.prev.shader.vao);

            glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 
                              g_master.prev.feedback_fbo);
            glViewport(0, 0, 
                  g_master.prev.feedback_tex.texture_size[0], 
                  g_master.prev.feedback_tex.texture_size[1]);

            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

            glUseProgramARB(g_master.shaders[i].program);
        }

    }

    SDL_GL_SwapWindow( window );

    /* Always make sure original texture is active if we go to gui */
    glActiveTexture2( GL_TEXTURE0 );
    glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
    glBindVertexArrayARB(0);

    return 0;

} /* end shader_draw */

/******************************************************************************
**  Function   :  shader_compile_shader
**                                                                    
**  Objective  :  This function compiles the shader program
**                                                   
**  Parameters :  name     - name of shader profile to save
**                p_shader - shader structure containing info to save
**                 
**  return     :  0 if error, otherwise success
**
******************************************************************************/ 
int shader_compile_shader ( GLhandleARB shader, const char *source ) {

    GLint status;

    glShaderSourceARB(shader, 1, &source, NULL);
    glCompileShaderARB(shader);
    glGetShaderivARB(shader, GL_COMPILE_STATUS, &status);
    if (status == 0) {
        GLint length;
        char *info;

        glGetShaderivARB(shader, GL_INFO_LOG_LENGTH, &length);
        info = SDL_stack_alloc(char, length+1);
        glGetShaderInfoLogARB(shader, length, NULL, info);
        length -= 1;
        while (length > 0 && info[length-1] == '\n') {
            info[length-1] = '\0';
            length = strlen(info);
        }
        logger_log_message (LOG_ERROR, "Shader compile error(s) encountered:", LOG_SRC);
        logger_log_message (LOG_ERROR, info, LOG_SRC);
        SDL_stack_free(info);

        return 0;
    } else  {
        return 1;
    }

} /* end shader_compile_shader */

/******************************************************************************
**  Function   :  shader_compile_shader_program
**                                                                    
**  Objective  :  This function compiles the shader program
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_compile_shader_program ( t_shader *p_shader ) {

    char fragment[] = "#define FRAGMENT\n#define PARAMETER_UNIFORM\n";
    char vertex[] = "#define VERTEX\n#define PARAMETER_UNIFORM\n";
    char *source;
    char *src_start;
    char version[257];

    version[0] = '\0';
    src_start = p_shader->source;

    glGetError();

    /* Create one program object to rule them all */
    p_shader->program = glCreateProgramARB();

    /* Check for version string */
    char *end = strchr(p_shader->source, '\n');
    if (end && (end-p_shader->source) < 256) {
        strncpy(version, p_shader->source, end - p_shader->source+1);
        version[end-p_shader->source+1] = '\0';
        if (strstr(version, "version")) {
            src_start = end;
        }
        else { 
            version[0] = '\0';
        }
        strcpy ( version, "#version 150\n");
    }

    /* Create the vertex shader */
    source = malloc ( strlen(version) + strlen(vertex) + 
                      strlen(src_start) + 1 ); 
    sprintf ( source, "%s%s%s", version, vertex, src_start);
    p_shader->vertex_shader = glCreateShaderARB(GL_VERTEX_SHADER);
    if (!shader_compile_shader(p_shader->vertex_shader, source)) {
        logger_log_message (LOG_ERROR, "Failed to compile vertex shader", LOG_SRC);
        free ( source );
        shader_free_program(p_shader);
        return -1;
    }
    free ( source );

    /* Create the fragment shader */
    source = malloc ( strlen(version) + strlen(fragment) + 
                      strlen(src_start) + 1 ); 
    sprintf ( source, "%s%s%s", version, fragment, src_start);
    p_shader->fragment_shader = glCreateShaderARB(GL_FRAGMENT_SHADER);
    if (!shader_compile_shader(p_shader->fragment_shader, source)) {
        logger_log_message (LOG_ERROR, "Failed to compile fragment shader", LOG_SRC);
        free ( source );
        shader_free_program(p_shader);
        return -1;
    }
    free ( source );

    /* ... and in the darkness bind them */
    glAttachShaderARB(p_shader->program, p_shader->vertex_shader);
    glAttachShaderARB(p_shader->program, p_shader->fragment_shader);
    glLinkProgramARB(p_shader->program);

    glDetachShaderARB(p_shader->program, p_shader->vertex_shader);
    glDetachShaderARB(p_shader->program, p_shader->fragment_shader);

    glDeleteShaderARB(p_shader->vertex_shader);
    glDeleteShaderARB(p_shader->fragment_shader);

    glUseProgramARB(0);

    return shader_check_error ( "Shader Compile" );

} /* end shader_compile_shader_program */

/******************************************************************************
**  Function   :  shader_texture_from_sdl_surface
**                                                                    
**  Objective  :  This function copies an sdl surface to a texture and frees
**                the surface.
**                                                   
**  Parameters :  program - compiled shader program
**                name    - parameter name
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_texture_from_sdl_surface ( SDL_Surface *surface ) {

    SDL_Surface *new_surface;
    Uint32 rmask,gmask,bmask,amask;

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x000000ff;
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = 0xff000000;
#endif

    new_surface = SDL_CreateRGBSurface(0, surface->w, surface->h, 32,
                                   rmask, gmask, bmask, amask);
    SDL_BlitSurface(surface, NULL, new_surface, NULL);

    glTexImage2D ( GL_TEXTURE_2D, 0, GL_RGBA, surface->w, surface->h, 0,
                   GL_RGBA, GL_UNSIGNED_INT_8_8_8_8_REV, new_surface->pixels);

    SDL_FreeSurface ( surface );
    SDL_FreeSurface ( new_surface );

    return shader_check_error ( "Shader Texture from Image" );

} /* end shader_texture_from_sdl_surface */

/******************************************************************************
**  Function   :  shader_gen_tex_coord_buffer
**                                                                    
**  Objective  :  This function generates the buffer to hold the texture 
**                coordinate buffer and copies the values to it
**                                                   
**  Parameters :  p_tex - texture to generate buffer for
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_gen_tex_coord_buffer ( t_texture *p_tex ) {

    /* create buffer of coordinates to use later */
    if ( p_tex->vbo_coord ) {
        glBindBufferARB(GL_ARRAY_BUFFER, p_tex->vbo_coord);
        glBufferSubDataARB(GL_ARRAY_BUFFER, 0, 8 * sizeof(GLfloat), 
                p_tex->coord);
    }
    else {
        glGenBuffersARB(1, &p_tex->vbo_coord);
        glBindBufferARB(GL_ARRAY_BUFFER, p_tex->vbo_coord);
        glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                p_tex->coord, GL_STATIC_DRAW);
    }
    glBindBufferARB(GL_ARRAY_BUFFER, 0);

    return shader_check_error ( "Shader Generate TexCoord Buffer" );

} /* end shader_gen_tex_coord_buffer */

/******************************************************************************
**  Function   :  shader_gen_texture
**                                                                    
**  Objective  :  This function generates a new texture for input tex struct.
**                The caller needs to bind the right FBO before coming here.
**                Scaling for width and height is done before this
**                                                   
**  Parameters :  program - compiled shader program
**                name    - parameter name
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_gen_texture ( t_master_shader *p_master, t_texture *p_tex, 
              void *frame, int wrap_mode, int filter, int float_frame ) {

    int status;
    int new = 0;

    /* create texture and bind to a unit, unit is the uniform */
	if ( !p_tex->id ) {
		glActiveTexture2(GL_TEXTURE0 + (++p_master->tex_count));
		glGenTextures(1, &p_tex->id);
		glBindTexture(GL_TEXTURE_2D, p_tex->id);
		p_tex->tex_unit = p_master->tex_count;
        new = 1;
	}

    /* else we are just changing sizes (like on viewport change) */
    else {
		glActiveTexture2(GL_TEXTURE0 + p_tex->tex_unit);
		glBindTexture(GL_TEXTURE_2D, p_tex->id);
    }

    /* width and height are input size, power of 2 for texture size */
    p_tex->texture_size[0] = util_next_power_of_two ( p_tex->dimen.w + p_tex->dimen.x );
    p_tex->texture_size[1] = util_next_power_of_two ( p_tex->dimen.h + p_tex->dimen.y);

    GLfloat x_start = (GLfloat)p_tex->dimen.x / p_tex->texture_size[0];
    GLfloat x_stop = ((GLfloat)p_tex->dimen.x + (GLfloat)p_tex->dimen.w) / 
                     p_tex->texture_size[0];
    GLfloat y_start = (GLfloat)p_tex->dimen.y / p_tex->texture_size[1];
    GLfloat y_stop = ((GLfloat)p_tex->dimen.y + (GLfloat)p_tex->dimen.h) / 
                     p_tex->texture_size[1];

    p_tex->coord[0] = x_start;
    p_tex->coord[1] = y_start;
    p_tex->coord[2] = x_stop;
    p_tex->coord[3] = y_start;
    p_tex->coord[4] = x_start;
    p_tex->coord[5] = y_stop;
    p_tex->coord[6] = x_stop;
    p_tex->coord[7] = y_stop;

    /* set texture parameters */
    glTexImage2D(GL_TEXTURE_2D, 0, float_frame ? GL_RGBA32F : GL_RGBA, 
            p_tex->texture_size[0], p_tex->texture_size[1], 0, GL_RGBA, 
            float_frame ? GL_FLOAT : GL_UNSIGNED_BYTE, 0);
    if ( new ) {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter );
        glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, color_black);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_mode);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_mode);
    }

    status = shader_check_error ( "Shader Generate Texture" );

    status += shader_gen_tex_coord_buffer ( p_tex );

    glActiveTexture2(GL_TEXTURE0);

    return status;

} /* end shader_gen_texture */

/******************************************************************************
**  Function   :  shader_gen_fbo
**                                                                    
**  Objective  :  This function creates an fbo for the shader to draw to if
**                not drawing to the screen.  We also call this to recreate
**                the texture if the screen size changes and tex is based
**                on viewport size.
**                                                   
**  Parameters :  p_shader      - the master shader
**                pass          - current shader pass that fbo is ouput for
**                in_tex        - input texture to the shader
**                filter_linear - apply linear filtering to texture
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_gen_fbo ( t_master_shader *p_master, GLuint *fbo, t_texture *fbo_tex,
                                      int wrap_mode, int filter_linear, 
                                      int float_framebuffer ) {
    if ( !(*fbo) )
        glGenFramebuffers(1, fbo);

    glBindFramebuffer(GL_FRAMEBUFFER, *fbo);

    shader_check_error ( "Shader Generate FBO" );

    if ( fbo_tex->id )
        glDeleteTextures(1, &fbo_tex->id);

    shader_gen_texture ( p_master, fbo_tex, 0, 
        wrap_mode, filter_linear ?  GL_LINEAR : GL_NEAREST, float_framebuffer );

    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, 
                              GL_TEXTURE_2D, fbo_tex->id, 0);
    GLenum DrawBuffers[1] = {GL_COLOR_ATTACHMENT0};
    glDrawBuffers(1, DrawBuffers); 

    glBindFramebuffer(GL_FRAMEBUFFER, 0);

    return shader_check_error ( "Shader FBO Texture" );

} /* end shader_gen_fbo */

/******************************************************************************
**  Function   :  shader_set_texture_uniforms
**                                                                    
**  Objective  :  This function sets uniforms for the input texture
**                                                   
**  Parameters :  p_locs - locations of texture uniforms
**                p_tex  - input texture information
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_set_texture_uniforms ( t_texture_locs *p_locs, t_texture *p_tex ) {

    GLfloat input_size[] = {p_tex->dimen.w, p_tex->dimen.h};

    if ( p_locs->tex_unit >= 0 ) 
        glUniform1iARB(p_locs->tex_unit, p_tex->tex_unit);

    if ( p_locs->input_size >= 0 )
        glUniform2fvARB(p_locs->input_size, 1, input_size);

    if ( p_locs->texture_size >= 0 )
        glUniform2fvARB(p_locs->texture_size, 1, p_tex->texture_size);

    shader_check_error ( "Shader Set Texture Uniforms" );

    if ( p_locs->coord >= 0 ) {
        glBindBufferARB(GL_ARRAY_BUFFER, p_tex->vbo_coord);
        glEnableVertexAttribArrayARB(p_locs->coord);
        glVertexAttribPointerARB(p_locs->coord, 2, GL_FLOAT, GL_FALSE, 
                0, 0);
        glBindBufferARB(GL_ARRAY_BUFFER, 0);
    }

    return shader_check_error ( "Shader Set Texture Coord Buffer" );

} /* end shader_set_texture_uniforms */

/******************************************************************************
**  Function   :  shader_get_location
**                                                                    
**  Objective  :  This function get uniform or attribute location, if not 
**                found loc will be -1
**                                                   
**  Parameters :  program - compiled shader program
**                name    - parameter name
**                 
**  return     :  integer location, negative if not found
**
******************************************************************************/ 
int shader_get_location ( GLhandleARB prog, char *name ) {

    int loc;
    loc = glGetUniformLocationARB(prog, name);

    if (loc >= 0)
        return loc;

    return glGetAttribLocationARB(prog, name);

} /* end shader_get_location */

/******************************************************************************
**  Function   :  shader_get_texture_location
**                                                                    
**  Objective  :  This function get the needed texture parameter locations. It
**                only looks for locations that have not been searched for yet
**                This is due to Prev / PrevPass etc being able to represent 
**                same actual location
**                                                   
**  Parameters :  program - compiled shader program
**                name    - Texture name
**                locs    - locations to fill
**                 
**  return     :  how many of the params were found
**
******************************************************************************/ 
int shader_get_texture_location ( GLhandleARB prog, char *name, 
                                                    t_texture_locs *locs ) {

    int num_found = 0;
    char tmp_var[4][32];

    sprintf ( tmp_var[0], "%sTexture", name );
    sprintf ( tmp_var[1], "%sInputSize", name );
    sprintf ( tmp_var[2], "%sTextureSize", name );
    sprintf ( tmp_var[3], "%sTexCoord", name );

    if ( locs->tex_unit < 0 )
        locs->tex_unit = shader_get_location(prog, tmp_var[0]);
    if ( locs->input_size < 0 )
        locs->input_size = shader_get_location(prog, tmp_var[1]);
    if ( locs->texture_size < 0 )
        locs->texture_size = shader_get_location(prog, tmp_var[2]);
    if ( locs->coord < 0 )
        locs->coord = shader_get_location(prog, tmp_var[3]);

    if ( locs->tex_unit >= 0 )
        num_found++;
    if ( locs->input_size >= 0 )
        num_found++;
    if ( locs->texture_size >= 0 )
        num_found++;
    if ( locs->coord >= 0 )
        num_found++;

    return num_found;

} /* end shader_get_texture_location */

/******************************************************************************
**  Function   :  shader_create_lut_textures
**                                                                    
**  Objective  :  This function creates the lut textures
**                                                   
**  Parameters :  p_master - master shader struct with texture file info
**                 
**  return     :  0 success, otherwise error(s) were found
**
******************************************************************************/ 
int shader_create_lut_textures ( t_master_shader *p_master) {

    int i, status = 0;
    char *full_file;
    SDL_Surface *sfc;
    t_lut_texture *lut_tex;
    GLenum filter_min;
    GLenum filter_mag;

    for ( i = 0; i < p_master->num_lut_textures; ++i ) {

        lut_tex = &(p_master->lut_textures[i]);

        /* Create absolute filename */
        full_file = malloc ( strlen(p_master->file) + strlen(lut_tex->file) );

        strcpy ( full_file, p_master->file );
        util_strip_file_to_directory ( full_file );
        sprintf ( full_file, "%s%c%s", full_file, 
                      util_get_dir_seperator_character(), lut_tex->file );

        sfc = IMG_Load ( full_file );
        if ( sfc ) {
            free(full_file);
            glActiveTexture2(GL_TEXTURE0 + (++p_master->tex_count));
            glGenTextures(1, &lut_tex->tex_id);
            glBindTexture( GL_TEXTURE_2D, lut_tex->tex_id );

            shader_texture_from_sdl_surface ( sfc );
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0);
            if ( p_master->lut_textures[i].mipmap ) {
                glGenerateMipmap(GL_TEXTURE_2D);
                shader_check_error ( "Gen Mipmap LUT Texture" );
                filter_min = lut_tex->linear ? 
                            GL_LINEAR_MIPMAP_LINEAR : GL_NEAREST_MIPMAP_NEAREST;
                filter_mag = lut_tex->linear ? 
                            GL_LINEAR: GL_NEAREST;
            }
            else {
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0);
                filter_min = filter_mag = lut_tex->linear ? GL_LINEAR : GL_NEAREST;
            }
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter_min );
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter_mag );
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, lut_tex->wrap_mode);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, lut_tex->wrap_mode);
            glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, color_black);
            lut_tex->tex_unit = p_master->tex_count;
            status += shader_check_error ( "LUT Texture" );
        }
        else {
            free(full_file);
            full_file = malloc ( strlen(lut_tex->file) + 50 );
            sprintf ( full_file, "Shader LUT Image Load Error: %s", lut_tex->file );
            logger_log_message (LOG_ERROR, full_file, LOG_SRC);
		    const char *msg = IMG_GetError();
            free(full_file);
            full_file = malloc ( strlen(msg) + 50 );
            sprintf ( full_file, "                : %s", msg );
            logger_log_message (LOG_ERROR, full_file, LOG_SRC);
            free(full_file);
        }
    }

    glActiveTexture2(GL_TEXTURE0);

    return status;

} /* end shader_create_lut_textures */

/******************************************************************************
**  Function   :  shader_set_shader_sizes
**                                                                    
**  Objective  :  This function sets input, output, and texture sizes based
**                on the input texture and viewports as required
**                                                   
**  Parameters :  p_shader      - shader pass to set values for
**                input_texture - texture to set sizes based on
**                viewport      - used for viewport scaling
**                 
**  return     :  0 success, otherwise error(s) were found
**
******************************************************************************/ 
int shader_set_shader_sizes ( t_shader *p_shader, t_texture *input_texture,
                              SDL_Rect viewport ) {

    /* Input and Texture sizes based on input texture */
    p_shader->vals.texture_size[0] = input_texture->texture_size[0];
    p_shader->vals.texture_size[1] = input_texture->texture_size[1];
    p_shader->vals.input_size[0] = input_texture->dimen.w;
    p_shader->vals.input_size[1] = input_texture->dimen.h;

    /* Scale Output */
        /* FIXME: Do we change viewports based on source type */ 
    if ( strstr(p_shader->scale_type_x, "viewport") )
        p_shader->vals.output_size[0] = (GLfloat)viewport.w *  
                                                 atof ( p_shader->scale_x );
    else if ( strstr(p_shader->scale_type_x, "absolute") )
        p_shader->vals.output_size[0] = atoi ( p_shader->scale_x );
    else
        p_shader->vals.output_size[0] = (GLfloat) p_shader->vals.input_size[0] * 
                                                  atof ( p_shader->scale_x );

    if ( strstr(p_shader->scale_type_y, "viewport") )
        p_shader->vals.output_size[1] = (GLfloat)viewport.h *  
                                                 atof ( p_shader->scale_y );
    else if ( strstr(p_shader->scale_type_y, "absolute") )
        p_shader->vals.output_size[1] = atoi ( p_shader->scale_y );
    else
        p_shader->vals.output_size[1] = (GLfloat) p_shader->vals.input_size[1] * 
                                                  atof ( p_shader->scale_y );
    /* Set Uniforms */
    if ( p_shader->locs.input_size >= 0 )
        glUniform2fvARB(p_shader->locs.input_size, 1, p_shader->vals.input_size);
    shader_check_error ( "Shader In Sizes" );
    if ( p_shader->locs.output_size >= 0 )
        glUniform2fvARB(p_shader->locs.output_size, 1, p_shader->vals.output_size);
    shader_check_error ( "Shader Output Sizes" );
    if ( p_shader->locs.texture_size >= 0 )
        glUniform2fvARB(p_shader->locs.texture_size, 1, p_shader->vals.texture_size);
    shader_check_error ( "Shader Tex Sizes" );

    return shader_check_error ( "Shader In/Out/Tex Sizes" );

} /* end shader_set_shader_sizes */

/******************************************************************************
**  Function   :  shader_set_values
**                                                                    
**  Objective  :  This function sets uniform or attribute values to be used
**                                                   
**  Parameters :  p_master - master shader 
**                pass - which shader pass we are setting values for
**                 
**  return     :  error, 0 for success
**
******************************************************************************/ 
int shader_set_values ( t_master_shader *p_master, int pass ) {

    int i,j;
    static int create_prev_frame;
    static int create_feedback;
    t_uniform_locations *locs = &(p_master->shaders[pass].locs);
    t_shader *p_shader = &(p_master->shaders[pass]);
    locs->pass = 0;

    /* Attach this shader program */
    glUseProgramARB(p_shader->program);

    if ( pass == 0 ) {
        create_prev_frame = 0;
        create_feedback = 0;
    }

    /*
     * Get Locations of Uniforms
     */
    locs->texture = shader_get_location ( p_shader->program, "Texture" );
    locs->mvp_matrix = shader_get_location(p_shader->program, "MVPMatrix");
    locs->tex_coord = shader_get_location(p_shader->program, "TexCoord");
    locs->vertex_coord = shader_get_location(p_shader->program, "VertexCoord");
    locs->color = shader_get_location(p_shader->program, "Color");
    locs->lut_tex_coord = shader_get_location(p_shader->program, "LUTTexCoord");
    locs->input_size = shader_get_location(p_shader->program, "InputSize");
    locs->output_size = shader_get_location(p_shader->program, "OutputSize");
    locs->texture_size = shader_get_location(p_shader->program, "TextureSize");
    locs->frame_count = shader_get_location(p_shader->program, "FrameCount");
    locs->frame_direction = shader_get_location(p_shader->program, "FrameDirection");

    shader_get_texture_location ( p_shader->program, "Orig", &locs->orig ); 

    create_feedback += shader_get_texture_location ( p_shader->program, 
                                                  "Feedback", &locs->feedback );

    shader_check_error ( "Shader Set Values Locations" );

    /* Location Previous Pass Textures */
    int high_pass = 0;
    char tmp_var[24];
    t_texture_locs tmp_loc;

    if ( pass > 1 ) {
        sprintf ( tmp_var, "PassPrev%d", pass );
        shader_get_texture_location ( p_shader->program, tmp_var, &locs->orig ); 
    }

    // Determine passes to allocate
    for ( i = 1; i < pass; ++i ) {
        shader_init_texture_locs ( &tmp_loc );
        sprintf ( tmp_var, "Pass%d", i );
        if ( shader_get_texture_location ( p_shader->program, tmp_var, &tmp_loc ) ) 
            high_pass = i;

        shader_init_texture_locs ( &tmp_loc );
        sprintf ( tmp_var, "PassPrev%d", pass-i );
        if ( shader_get_texture_location ( p_shader->program, tmp_var, &tmp_loc ) )
            high_pass = i;

        if ( strlen(p_master->shaders[i-1].alias) ) {
            shader_init_texture_locs ( &tmp_loc );
            if ( shader_get_texture_location ( p_shader->program, 
                    p_master->shaders[i-1].alias, &tmp_loc ) )
                high_pass = i;
        }
    }

    // Once found, allocate and find locs
    if ( high_pass ) {
        p_shader->num_passes = high_pass;
        locs->pass = malloc ( sizeof(*(locs->pass)) * high_pass );
        for ( i = 1; i <= high_pass; ++i ) {
            shader_init_texture_locs ( &locs->pass[i-1] );
            sprintf ( tmp_var, "Pass%d", i );
            shader_get_texture_location ( p_shader->program, tmp_var, &locs->pass[i-1] );
            sprintf ( tmp_var, "PassPrev%d", pass-i );
            shader_get_texture_location ( p_shader->program, tmp_var, &locs->pass[i-1] );
            if ( i < p_master->num_shaders ) {
                if ( strlen(p_master->shaders[i-1].alias) ) {
                    shader_get_texture_location ( p_shader->program, 
                               p_master->shaders[i-1].alias, &locs->pass[i-1] );
                }
            }
        }
    }

    /* Previous Frame Textures Uniform Locations */
    create_prev_frame += shader_get_texture_location ( p_shader->program, 
                                                       "Prev", &locs->prev[0] );
    for ( i = 1; i < 7; ++i ) {
        sprintf ( tmp_var, "Prev%d", i );
        create_prev_frame += shader_get_texture_location ( p_shader->program, 
                                                            tmp_var, &locs->prev[i] );
    }

    /*
     * Now set the values
     */

    /* Texture (from previous output) */
    if ( locs->texture >= 0 ) {
        if ( pass == 0 )
            glUniform1iARB(locs->texture, 0);
        else
            glUniform1iARB(locs->texture, p_master->shaders[pass-1].fbo_tex.tex_unit);
    }

    /* MVP Matrix */
    p_shader->vals.mvp_matrix[0] =  1.0f;
    p_shader->vals.mvp_matrix[1] =  0.0f;
    p_shader->vals.mvp_matrix[2] =  0.0f;
    p_shader->vals.mvp_matrix[3] =  0.0f;
    p_shader->vals.mvp_matrix[4] =  0.0f;
    p_shader->vals.mvp_matrix[5] =  1.0f;
    p_shader->vals.mvp_matrix[6] =  0.0f;
    p_shader->vals.mvp_matrix[7] =  0.0f;
    p_shader->vals.mvp_matrix[8] =  0.0f;
    p_shader->vals.mvp_matrix[9] =  0.0f;
    p_shader->vals.mvp_matrix[10] = 1.0f;
    p_shader->vals.mvp_matrix[11] = 0.0f;
    p_shader->vals.mvp_matrix[12] =  0.0f;
    p_shader->vals.mvp_matrix[13] =  0.0f;
    p_shader->vals.mvp_matrix[14] =  0.0f;
    p_shader->vals.mvp_matrix[15] =  1.0f;

    if ( locs->mvp_matrix >= 0 )
        glUniformMatrix4fvARB(locs->mvp_matrix, 1, GL_FALSE, 
                                              p_shader->vals.mvp_matrix);

    /* 
     * Tex/Vertex Coords 
     *   tex coords originate at top-left while vertex
     *   is at middle.
     */
    t_texture *tmp_texture;

    if ( pass == 0 )
        tmp_texture = &p_master->orig;
    else
        tmp_texture = &p_master->shaders[pass-1].fbo_tex;
    
    // Start a VAO for our coordinates
    glGenVertexArraysARB(1, &p_shader->vao);
    glGenBuffersARB(1, &p_shader->vbo_tex);
    glGenBuffersARB(1, &p_shader->vbo_vertex);

    glBindVertexArrayARB(p_shader->vao);

    /* Texture Coordinates */
    glBindBufferARB(GL_ARRAY_BUFFER, p_shader->vbo_tex);
    glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                    tmp_texture->coord, GL_STATIC_DRAW);

    if ( p_shader->locs.tex_coord >= 0 ) {
         glEnableVertexAttribArrayARB(p_shader->locs.tex_coord);
         glVertexAttribPointerARB(p_shader->locs.tex_coord, 2, GL_FLOAT, GL_FALSE, 
                 0, 0);
    }
   
    /* Vertex Coordinates */
    glBindBufferARB(GL_ARRAY_BUFFER, p_shader->vbo_vertex);
    glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                               g_vcoords, GL_STATIC_DRAW);

    if ( p_shader->locs.vertex_coord >= 0 ) {
         glEnableVertexAttribArrayARB(p_shader->locs.vertex_coord);
         glVertexAttribPointerARB(p_shader->locs.vertex_coord, 2, GL_FLOAT, GL_FALSE, 
                 0, 0);
    }
   
    /* LUT Tex Coordinates */
    if ( p_shader->locs.lut_tex_coord >= 0 ) {
        p_shader->vals.lut_tex_coord[0] = 0.0f;
        p_shader->vals.lut_tex_coord[1] = 0.0f;
        p_shader->vals.lut_tex_coord[2] = 1.0f;
        p_shader->vals.lut_tex_coord[3] = 0.0f;
        p_shader->vals.lut_tex_coord[4] = 0.0f;
        p_shader->vals.lut_tex_coord[5] = 1.0f;
        p_shader->vals.lut_tex_coord[6] = 1.0f;
        p_shader->vals.lut_tex_coord[7] = 1.0f;
        glGenBuffersARB(1, &p_shader->vbo_lut_tex);
        glBindBufferARB(GL_ARRAY_BUFFER, p_shader->vbo_lut_tex);
        glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                        p_shader->vals.lut_tex_coord, GL_STATIC_DRAW);

        glEnableVertexAttribArrayARB(p_shader->locs.lut_tex_coord);
        glVertexAttribPointerARB(p_shader->locs.lut_tex_coord, 2, 
                GL_FLOAT, GL_FALSE, 0, 0);
    }

    glBindBufferARB(GL_ARRAY_BUFFER, 0);
    glBindVertexArrayARB(0);

    /* Find Input/Output/Texture sizes and set uniforms */
    shader_set_shader_sizes ( p_shader, tmp_texture, p_master->viewport );

    /* Frame Stuff */
    p_shader->vals.frame_direction = 1.0;
    p_shader->vals.frame_count = 0.0;
    if ( p_shader->locs.frame_count >= 0 )
        glUniform1iARB(p_shader->locs.frame_count, p_shader->vals.frame_count);
    if ( p_shader->locs.frame_direction >= 0 )
        glUniform1iARB(p_shader->locs.frame_direction, p_shader->vals.frame_direction);


    shader_check_error ( "Shader Set Values Common" );

    /* Mipmaps - if needed, set for the input texture from last pass */
    if (p_shader->mipmap_input) {
        glBindTexture( GL_TEXTURE_2D, tmp_texture->id );
        glGenerateMipmap(GL_TEXTURE_2D);

        shader_check_error ( "Shader Set Values Mipmaps" );
    }

    /* Setup Parameter Uniforms in shader */
    int num_params, loc;
    t_param *tmp_params;

    if ( p_master->num_params > 0 ) {
        for ( i = 0; i < p_master->num_params; ++i ) {
            loc = shader_get_location(p_shader->program, p_master->params[i].name);
            glUniform1fARB(loc, p_master->params[i].val);
        } 
    }

    if ( p_shader->num_params ) {
        num_params = p_shader->num_params;
        tmp_params = p_shader->params;
    }
    else {
        num_params = p_master->shaders[p_master->num_shaders-1].num_params;
        tmp_params = p_master->shaders[p_master->num_shaders-1].params;
    }

    if ( num_params > 0 ) {
        for ( i = 0; i < num_params; ++i ) {
            loc = shader_get_location(p_shader->program, tmp_params[i].name);
            glUniform1fARB(loc, tmp_params[i].val);
        } 
    }

    shader_check_error ( "Shader Set Values Uniforms" );

    /* Textures */
    shader_set_texture_uniforms ( &p_shader->locs.orig, &p_master->orig );

    /*
     * If not the last shader run, alloacate an FBO and texture for output
     *   if it is the last shader, we may need an FBO and texture if we 
     *   have a feedback pass or scaling is used, also if previewing in 
     *   the GUI
     */
    int create_fbo = 0;
    int filter_linear = 0;
    if ( pass == (p_master->num_shaders-1) ) {
        if ( p_master->framebuffer_out )
            create_fbo = 1;
        if ( p_master->prev.feedback_pass == pass )
            create_fbo = 1;
        if ( fabs(p_shader->vals.output_size[0] - p_shader->vals.input_size[0]) > 0.5 )
            create_fbo = 1;
        if ( fabs(p_shader->vals.output_size[1] - p_shader->vals.input_size[1]) > 0.5 )
            create_fbo = 1;
    }
    else {
        create_fbo = 1;
        filter_linear = p_master->shaders[pass+1].filter_linear;
    }

    if ( create_fbo ) {

        p_shader->fbo_tex.dimen.x = tmp_texture->dimen.x;
        p_shader->fbo_tex.dimen.y = tmp_texture->dimen.y;
        p_shader->fbo_tex.dimen.w = p_shader->vals.output_size[0];
        p_shader->fbo_tex.dimen.h = p_shader->vals.output_size[1];

        shader_gen_fbo ( p_master, &p_shader->fbo, &p_shader->fbo_tex, 
                p_shader->wrap_mode, filter_linear, p_shader->float_framebuffer );
    }
    else {
        p_shader->fbo = 0;
    }

    /* if the last shader run, allocate previous frame/feedback tex if needed */
    if ( pass == (p_master->num_shaders-1) ) {

        glUseProgramARB(0);

        //create_prev_frame = 0;

        if ( (create_prev_frame || create_feedback || 
                    (create_fbo && !p_master->framebuffer_out)) && 
                    !shader_compile_shader_program ( &p_master->prev.shader ) ) {

            glUseProgramARB(p_master->prev.shader.program);

            // Uniform Locations
            locs = &p_master->prev.shader.locs;
            locs->tex_coord = shader_get_location(p_master->prev.shader.program, "TexCoord");
            locs->vertex_coord = shader_get_location(p_master->prev.shader.program, "VertexCoord");

            locs->texture = shader_get_location ( p_master->prev.shader.program, "Texture" );
            locs->mvp_matrix = shader_get_location(p_master->prev.shader.program, "MVPMatrix");
            // Uniform Values
            if ( locs->texture >= 0 ) 
                glUniform1iARB(locs->texture, 0);
            if ( locs->mvp_matrix >= 0 ) 
                glUniformMatrix4fvARB(locs->mvp_matrix, 1, GL_FALSE, 
                                                      p_shader->vals.mvp_matrix);
            glGenVertexArraysARB(1, &p_master->prev.shader.vao);
            glGenBuffersARB(1, &p_master->prev.shader.vbo_vertex);
            glGenBuffersARB(1, &p_master->prev.shader.vbo_tex);

            glBindVertexArrayARB(p_master->prev.shader.vao);

            glBindBufferARB(GL_ARRAY_BUFFER, p_master->prev.shader.vbo_vertex);
            glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                            g_vcoords, GL_STATIC_DRAW);

            if ( p_master->prev.shader.locs.vertex_coord >= 0 ) {
                 glEnableVertexAttribArrayARB(locs->vertex_coord);
                 glVertexAttribPointerARB(locs->vertex_coord, 2, 
                                                 GL_FLOAT, GL_FALSE, 0, 0);
            }

            glBindBufferARB(GL_ARRAY_BUFFER, p_master->prev.shader.vbo_tex);
            glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                                g_tcoords, GL_STATIC_DRAW);

            if ( p_master->prev.shader.locs.tex_coord >= 0 ) {
                 glEnableVertexAttribArrayARB(locs->tex_coord);
                 glVertexAttribPointerARB(locs->tex_coord, 2, 
                                                 GL_FLOAT, GL_FALSE, 0, 0);
            }

            glBindBufferARB(GL_ARRAY_BUFFER, 0);
            glBindVertexArrayARB(0);

            // Create previous frame textures and tie them to FBOs
            for ( i = 0; i < MAX_PREV_FRAME && create_prev_frame; ++i ) {
                p_master->prev.tex[i].dimen.x = 0;
                p_master->prev.tex[i].dimen.y = 0;
                p_master->prev.tex[i].dimen.w = p_master->orig.texture_size[0];
                p_master->prev.tex[i].dimen.h = p_master->orig.texture_size[1];

                shader_gen_fbo ( p_master, &p_master->prev.fbo[i], 
                                           &p_master->prev.tex[i],
                                           GL_CLAMP_TO_BORDER, 
                                           0, 0 );


                // Update uniform in each shader
                for ( j = 0; j < p_master->num_shaders; ++j ) {
                    glUseProgramARB(p_master->shaders[j].program);
                    shader_set_texture_uniforms ( &p_master->shaders[j].locs.prev[i], 
                                                  &p_master->prev.tex[i] );
                }

                shader_check_error ( "Shader Set Values Previous" );
            }

            /* Create Feedback FBO and texture if needed */
            if ( create_feedback && p_master->prev.feedback_pass > -1 ) {

                int pass = p_master->prev.feedback_pass;

                p_master->prev.feedback_tex.dimen.x = 0;
                p_master->prev.feedback_tex.dimen.y = 0;
                p_master->prev.feedback_tex.dimen.w = 
                    p_master->shaders[pass].fbo_tex.texture_size[0];
                p_master->prev.feedback_tex.dimen.h = 
                    p_master->shaders[pass].fbo_tex.texture_size[1];

                shader_gen_fbo ( p_master, &p_master->prev.feedback_fbo, 
                                           &p_master->prev.feedback_tex,
                                           GL_CLAMP_TO_BORDER, 
                                           0, p_shader->float_framebuffer );

                // Update uniform in each shader
                for ( j = 0; j < p_master->num_shaders; ++j ) {
                    glUseProgramARB(p_master->shaders[j].program);
                    shader_set_texture_uniforms ( &p_master->shaders[j].locs.feedback, 
                                                  &p_master->prev.feedback_tex );
                }

                shader_check_error ( "Shader Set Values Feedback" );
            }

            /* 
             * if we are drawing to fbo and need to draw to screen, create a vao
             * to use with stock shader 
             */
            if (create_fbo && !p_master->framebuffer_out) {

                glGenVertexArraysARB(1, &p_master->screen_vao);

                glGenBuffersARB(1, &p_master->screen_vbo_vertex);
                glGenBuffersARB(1, &p_master->screen_vbo_tex);

                glBindVertexArrayARB(p_master->screen_vao);

                glBindBufferARB(GL_ARRAY_BUFFER, p_master->screen_vbo_vertex);
                glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                                g_vcoords_backbuffer, GL_STATIC_DRAW);

                if ( p_master->prev.shader.locs.vertex_coord >= 0 ) {
                     glEnableVertexAttribArrayARB(locs->vertex_coord);
                     glVertexAttribPointerARB(locs->vertex_coord, 2, 
                                                     GL_FLOAT, GL_FALSE, 0, 0);
                }

                glBindBufferARB(GL_ARRAY_BUFFER, p_master->screen_vbo_tex);
                glBufferDataARB(GL_ARRAY_BUFFER, 8 * sizeof(GLfloat), 
                                    p_shader->fbo_tex.coord, GL_STATIC_DRAW);

                if ( p_master->prev.shader.locs.tex_coord >= 0 ) {
                     glEnableVertexAttribArrayARB(locs->tex_coord);
                     glVertexAttribPointerARB(locs->tex_coord, 2, 
                                                     GL_FLOAT, GL_FALSE, 0, 0);
                }

                glBindBufferARB(GL_ARRAY_BUFFER, 0);
                glBindVertexArrayARB(0);

                shader_check_error ( "Shader Set Values Screen Out" );
            }
        }

        glUseProgramARB(p_shader->program);
    }

    /* Set pass texture uniforms */
    for ( i = 1; i <= high_pass; ++i ) {
        shader_set_texture_uniforms ( &p_master->shaders[pass].locs.pass[i-1], 
                                      &p_master->shaders[i-1].fbo_tex );
    }

    /* Lookup Texture locations */
    t_lut_texture *lut_tex;
    int tex_unit_loc;
    for ( i = 0; i < p_master->num_lut_textures; ++i ) {
        lut_tex = &(p_master->lut_textures[i]);
        tex_unit_loc = shader_get_location(p_shader->program, 
                                                          lut_tex->name);
        if ( tex_unit_loc >= 0 ) 
            glUniform1iARB(tex_unit_loc, lut_tex->tex_unit);
    }

    return shader_check_error ( "Shader Set Values" );

} /* end shader_set_values */

void shader_init_master ( t_master_shader *p_master ) {

    int i;

    if (!p_master)
        return;

    p_master->file = 0;
    p_master->shaders = 0;
    p_master->profile_dirty = 0;
    p_master->params_dirty = 0;
    p_master->num_shaders = 0;
    p_master->tex_count = 0;
    p_master->orig.id = 0;
    p_master->orig.vbo_coord = 0;
    p_master->orig.tex_unit = 0;
    p_master->lut_textures = 0;
    p_master->num_lut_textures = 0;
    p_master->params = 0;
    p_master->num_params = 0;
    p_master->framebuffer_out = 0;
    p_master->screen_vao = 0;
    p_master->screen_vbo_tex = 0;
    p_master->screen_vbo_vertex = 0;

    shader_init_shader( &p_master->prev.shader );
    p_master->prev.shader.file = malloc(12);
    strcpy ( p_master->prev.shader.file, "stock" );
    p_master->prev.shader.source = malloc(strlen(stock_shader)+1);
    strcpy ( p_master->prev.shader.source, stock_shader );
    p_master->prev.frame_ptr = 0;
    p_master->prev.feedback_pass = -1;
    for ( i = 0; i < MAX_PREV_FRAME; ++i ) {
        p_master->prev.tex[i].id = 0;
        p_master->prev.tex[i].vbo_coord = 0;
        p_master->prev.fbo[i] = 0;
    }
    p_master->prev.feedback_tex.id = 0;
    p_master->prev.feedback_tex.vbo_coord = 0;
    p_master->prev.feedback_fbo = 0;

}

void shader_init_shader ( t_shader *p_shader ) {

    int i;

    p_shader->file = 0;
    p_shader->source = 0;
    p_shader->params = 0;
    p_shader->vertex_shader = 0;
    p_shader->filter_linear = 0;
    p_shader->fragment_shader = 0;
    p_shader->float_framebuffer = 0;
    p_shader->srgb_framebuffer = 0;
    p_shader->wrap_mode = GL_CLAMP_TO_BORDER;
    p_shader->mipmap_input = 0;
    p_shader->program = 0;
    p_shader->vbo_tex = 0;
    p_shader->vao = 0;
    p_shader->vbo_lut_tex = 0;
    p_shader->vbo_vertex = 0;
    p_shader->frame_count_mod = 0;
    p_shader->num_passes = 0;
    p_shader->num_params = 0;
    p_shader->fbo = 0;
    p_shader->locs.pass = 0;
    strcpy ( p_shader->scale_x, "1.0" );
    strcpy ( p_shader->scale_y, "1.0" );
    strcpy ( p_shader->alias, "" );

    shader_init_texture_locs ( &p_shader->locs.orig );
    shader_init_texture_locs ( &p_shader->locs.feedback );
    for ( i = 0 ; i < MAX_PREV_FRAME; ++i ) {
        shader_init_texture_locs ( &p_shader->locs.prev[i] );
    }

    p_shader->fbo_tex.id = 0;
    p_shader->fbo_tex.vbo_coord = 0;
}

void shader_init_texture_locs ( t_texture_locs *p_locs ) {

    p_locs->tex_unit = -1;
    p_locs->input_size = -1;
    p_locs->texture_size = -1;
    p_locs->coord = -1;
}

/******************************************************************************
**  Function   :  shader_free_program
**                                                                    
**  Objective  :  This function frees program from current shader
**                                                   
**  Parameters :  p_shader - shader to free program for
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
void shader_free_program ( t_shader *p_shader ) {

    if (p_shader->program)
        glDeleteProgramARB(p_shader->program);

    p_shader->vertex_shader = 0;
    p_shader->fragment_shader = 0;
    p_shader->program = 0;

    shader_check_error ( "Shader Program Delete" );

} /* end shader_free_program */

/******************************************************************************
**  Function   :  shader_free_objects
**                                                                    
**  Objective  :  This function frees objects in the currently loaded shader
**                                                   
**  Parameters :  p_shader - shader pass to free objects for
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
void shader_free_objects ( t_shader *p_shader ) {

    if (p_shader->vbo_tex)
        glDeleteBuffersARB(1, &p_shader->vbo_tex);
    if (p_shader->vbo_lut_tex)
        glDeleteBuffersARB(1, &p_shader->vbo_lut_tex);
    if (p_shader->vbo_vertex)
        glDeleteBuffersARB(1, &p_shader->vbo_vertex);

    if (p_shader->vao)
        glDeleteVertexArraysARB(1, &p_shader->vao);

    /* Free textures */
    if ( p_shader->fbo_tex.id )
        glDeleteTextures(1, &p_shader->fbo_tex.id);

    if ( p_shader->fbo_tex.vbo_coord )
        glDeleteBuffersARB(1, &p_shader->fbo_tex.vbo_coord);

    if ( p_shader->fbo )
        glDeleteFramebuffers(1, &p_shader->fbo);

    p_shader->vbo_tex = 0;
    p_shader->vbo_lut_tex = 0;
    p_shader->vbo_vertex = 0;
    p_shader->vao = 0;
    p_shader->fbo = 0;
    p_shader->fbo_tex.id = 0;
    p_shader->fbo_tex.vbo_coord = 0;

    shader_check_error ( "Shader Object Delete" );

} /* end shader_free_objects */

/******************************************************************************
**  Function   :  shader_free_master
**                                                                    
**  Objective  :  This function frees OpenGL resources from the master
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
void shader_free_master ( t_master_shader *p_master ) {

    int i;

    if (!p_master->context)
        return;

    glUseProgramARB(0);
    glBindBufferARB(GL_ARRAY_BUFFER, 0);
    glBindVertexArrayARB(0);

    shader_free_program ( &p_master->prev.shader );
    shader_free_objects ( &p_master->prev.shader );
    if ( p_master->orig.id )
        glDeleteTextures(1, &p_master->orig.id);
    for ( i = 0; i < p_master->num_lut_textures; ++i ) {
        if (p_master->lut_textures[i].tex_id)
            glDeleteTextures(1, &p_master->lut_textures[i].tex_id);
    }

    for ( i = 0; i < MAX_PREV_FRAME; ++i ) {
        if ( p_master->prev.tex[i].id )
            glDeleteTextures(1, &p_master->prev.tex[i].id);
        if ( p_master->prev.tex[i].vbo_coord )
            glDeleteBuffersARB(1, &p_master->prev.tex[i].vbo_coord);
        if ( p_master->prev.fbo[i] )
            glDeleteFramebuffers(1, &p_master->prev.fbo[i]);
    }
    if ( p_master->prev.feedback_tex.id )
        glDeleteTextures(1, &p_master->prev.feedback_tex.id);
    if ( p_master->prev.feedback_tex.vbo_coord )
        glDeleteBuffersARB(1, &p_master->prev.feedback_tex.vbo_coord);
    if ( p_master->prev.feedback_fbo )
        glDeleteFramebuffers(1, &p_master->prev.feedback_fbo);

    if ( p_master->prev.shader.vbo_tex )
        glDeleteBuffersARB(1, &p_master->prev.shader.vbo_tex);
    if ( p_master->prev.shader.vbo_vertex )
        glDeleteBuffersARB(1, &p_master->prev.shader.vbo_vertex);
    if ( p_master->prev.shader.vao )
        glDeleteVertexArraysARB(1, &p_master->prev.shader.vao);

    if ( p_master->screen_vbo_tex )
        glDeleteBuffersARB(1, &p_master->screen_vbo_tex);
    if ( p_master->screen_vbo_vertex )
        glDeleteBuffersARB(1, &p_master->screen_vbo_vertex);
    if ( p_master->screen_vao )
        glDeleteVertexArraysARB(1, &p_master->screen_vao);

} /* end shader_free_master */

/******************************************************************************
**  Function   :  shader_free_master_contents
**                                                                    
**  Objective  :  This function frees master shader contents, includes all
**                shaders and textures
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
void shader_free_master_contents ( t_master_shader *pm_shader ) {

    int i;

    for ( i = 0; i < pm_shader->num_shaders; ++i ) {
        free ( pm_shader->shaders[i].file );
        free ( pm_shader->shaders[i].source );
        free ( pm_shader->shaders[i].params );
    }
    for ( i = 0; i < pm_shader->num_lut_textures; ++i ) {
        free ( pm_shader->lut_textures[i].file );
    }
    free ( pm_shader->lut_textures );
    free ( pm_shader->params );

    free ( pm_shader->shaders );
    free ( pm_shader->file );
    shader_init_master ( pm_shader );

} /* end shader_free_master_contents */

/******************************************************************************
**  Function   :  shader_free_context
**                                                                    
**  Objective  :  This function frees currently loaded context
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
void shader_free_context ( ) {

    if (g_master.context) {
        SDL_GL_DeleteContext(g_master.context);
    }

    g_master.context = 0;

} /* end shader_free_context */

/******************************************************************************
**  Function   :  shader_exit
**                                                                    
**  Objective  :  This function is called at exit to ensure resources are freed
**                                                   
**  Parameters :  NONE
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
void shader_exit ( ) {

    int i;

    shader_free_master ( &g_master );

    for ( i = 0; i < g_master.num_shaders; ++i ) {
        shader_free_program(&g_master.shaders[i]);
        shader_free_objects(&g_master.shaders[i]);
    }

    shader_free_context();

    shader_free_master_contents ( &g_master );

} /* end shader_exit */

/******************************************************************************
**  Function   :  shader_check_error
**                                                                    
**  Objective  :  This function checks for a GL error and prints to log
**                                                   
**  Parameters :  prefix - string to precede error in log
**                 
**  return     :  0 if success, otherwise error
**
******************************************************************************/ 
int shader_check_error ( char *prefix ) {

    GLenum error;
    char *text;

    if ((error = glGetError())) {

        if ( !m_log_errors )
            return error ? -1 : 0;

        text = malloc ( strlen(prefix) + 40 );
        strcpy ( text, prefix );
        switch ( error ) {
            case GL_INVALID_ENUM:
                strcat ( text, ": GL Invalid Enum" );
                break;
            case GL_INVALID_VALUE:
                strcat ( text, ": GL Invalid Value" );
                break;
            case GL_INVALID_OPERATION:
                strcat ( text, ": GL Invalid Operation" );
                break;
            case GL_INVALID_FRAMEBUFFER_OPERATION:
                strcat ( text, ": GL Invalid Framebuffer Operation" );
                break;
            case GL_OUT_OF_MEMORY:
                strcat ( text, ": GL Out of Memory" );
                break;
            default:
                strcat ( text, ": GL Unknown Error" );
                break;
        }

        logger_log_message (LOG_ERROR, text, LOG_SRC);

        free(text);

        if (m_in_draw)
            m_log_errors = 0;
    }

    return error ? -1 : 0;

} /* end shader_check_error */
