/******************************************************************************
*
* Copyright (C) 2017  Brian Berlin
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>
*
******************************************************************************
*
* FILENAME: shader.h
*
* DESCRIPTION:  This contains function declarations for using OpenGL shaders,
*                these would mostly be used to emulate CRT behavior
*
* AUTHORS:  bberlin
*
* CHANGE LOG
*
*  VER     DATE     CHANGED BY   DESCRIPTION OF CHANGE
* ------  --------  ----------   ------------------------------------
* 0.8.0   03/26/17  bberlin      Creation
******************************************************************************/
#ifndef shader_h
#define shader_h

#ifdef WIN32
    #include <windows.h>
#endif

#ifdef __APPLE__
    #include <SDL2/SDL.h>
    #include <SDL2/SDL_opengl.h>
#else
    #include <SDL.h>
    #include <SDL_opengl.h>
#endif

#define MAX_PREV_FRAME 7
#define SHADER_NONE "None"

typedef struct {
    GLuint id;
    int tex_unit;
    SDL_Rect dimen;
    GLfloat texture_size[2];
    GLfloat coord[8];
    GLuint vbo_coord;
} t_texture;

typedef struct {
    int tex_unit;
    int input_size;
    int texture_size;
    int coord;
} t_texture_locs;

typedef struct {
    int texture;
    int mvp_matrix;
    int tex_coord;
    int vertex_coord;
    int color;
    int lut_tex_coord;
    int input_size;
    int output_size;
    int texture_size;
    int frame_count;
    int frame_direction;
    t_texture_locs orig;
    t_texture_locs feedback;
    t_texture_locs prev[MAX_PREV_FRAME];
    t_texture_locs *pass;
} t_uniform_locations;

typedef struct {
    GLfloat mvp_matrix[16];
    GLfloat tex_coord[8];
    GLfloat lut_tex_coord[8];
    GLfloat vertex_coord[8];
    int color;
    GLfloat input_size[2];
    GLfloat output_size[2];
    GLfloat texture_size[2];
    int frame_count;
    int frame_direction;
} t_uniform_values;

typedef struct {
    char name[257];
    char description[257];
    GLfloat val;
    GLfloat def;
    GLfloat min;
    GLfloat max;
    GLfloat step;
} t_param;

typedef struct {
    char name[257];
    char *file;
    int linear;
    int wrap_mode;
    int mipmap;
    GLuint tex_id;
    int tex_unit;
} t_lut_texture;

typedef struct {
    char *file;
    char *source;
    int filter_linear;
    int float_framebuffer;
    int srgb_framebuffer;
    int wrap_mode;
    int mipmap_input;
    int frame_count_mod;
    char scale_type_x[48];
    char scale_type_y[48];
    char scale_x[48];
    char scale_y[48];
    char alias[64];
    GLuint fbo;
    GLuint vao;
    t_texture fbo_tex;
    GLhandleARB program;
    GLhandleARB vertex_shader;
    GLhandleARB fragment_shader;
    GLuint vbo_lut_tex;
    GLuint vbo_tex;
    GLuint vbo_vertex;
    int num_passes;
    int num_params;
    t_param *params;
    t_uniform_locations locs;
    t_uniform_values vals;
} t_shader;

typedef struct {
    t_shader shader;
    t_texture tex[7];
    GLuint fbo[7];
    int frame_ptr;
    int feedback_pass;
    t_texture feedback_tex;
    GLuint feedback_fbo;
} t_prev_frames;

typedef struct {
    char name[257];
    char *file;
    int profile_dirty;
    int params_dirty;
    int num_shaders;
    int framebuffer_out;
    t_shader *shaders;
    int num_lut_textures;
    t_lut_texture *lut_textures;
    int num_params;
    t_param *params;
    int supported;
    SDL_GLContext context;
    SDL_Rect viewport;
    int tex_count;
    t_texture orig;
    t_prev_frames prev;
    GLuint screen_vao;
    GLuint screen_vbo_vertex;
    GLuint screen_vbo_tex;
} t_master_shader;

/*
 * Function Declarations
 */
int shader_set_defaults ( void );
t_master_shader * shader_get_master();
int shader_set_profile ( char *name, SDL_Rect tex_rect );
int shader_load_profile ( const char *name, t_master_shader *p_shader );
int shader_save_profile ( char *name, t_master_shader *p_shader );
int shader_open_master ( char *name, t_master_shader *pm_shader );
int shader_setup_context ( SDL_Window *window );
int shader_init ( SDL_Window *window );
int shader_setup ( int vsync, SDL_Window *window, SDL_Rect tex_rect );
int shader_draw ( SDL_Window *window, void *pixels, int width );
int shader_update_viewport_values ( int win_w, int win_h );

void shader_init_master ( t_master_shader *p_master );
void shader_free_master_contents ( t_master_shader *pm_shader );
void shader_exit ( void );

#endif
